page    60, 132

;******************************************************************************
        title   KEYDISP.ASM - keyboard scan code display
;******************************************************************************
;
;   Title:    KEYDISP.EXE - keyboard scan code display
;
;   Version:  1.00
;
;   Date:     October 27, 1993
;
;   Author:   Bryan A. Woodruff
;
;******************************************************************************
;
;   Change log:
;
;   DATE      REVISION  DESCRIPTION                                 AUTHOR
;   --------  --------  ------------------------------------------  -----------
;   10/27/93  1.00      Wrote it.                                   BryanW
;
;******************************************************************************
;
;   Functional Description:
;
;      Hooks INT 9h and displays scan codes
;
;******************************************************************************
        
        .286


;
; keyboard interface
;

KYBD_PORT_A             equ     60h
KYBD_PORT_B             equ     61h

_TEXT   segment byte public 'CODE'

        assume cs:_TEXT, ds:_TEXT

OldKybdInt      dd      0
ScanBuffer      db      16 dup (?)
ScanStart       dw      0
ScanEnd         dw      0
fDone           db      0
                            
ScanCode        db      '00 ', '$'

;---------------------------------------------------------------------------
;
;   Keyboard_Handler
;
;   DESCRIPTION:
;       This hooks int 09h and displays the stuffs the scan codes
;       in the display buffer.
;
;---------------------------------------------------------------------------

Keyboard_Handler        proc    near

        pusha
        push    ds

        push    cs
        pop     ds

        in      al, KYBD_PORT_A
        cmp     al, 1                           ; Q: Escape?
        jnz     SHORT KH_StuffIt
        mov     fDone, 1
        jmp     SHORT KH_EatIt

KH_StuffIt:
        mov     di, ScanEnd
        mov     byte ptr ScanBuffer[di], al
        inc     di
        cmp     di, 10h
        jb      KH_NoWrap
        xor     di, di

KH_NoWrap:
        mov     ScanEnd, di

KH_EatIt:
        in      al, KYBD_PORT_B
        or      al, 80h
        jmp     $+2
        out     KYBD_PORT_B, al
        and     al, 7Fh
        jmp     $+2
        out     KYBD_PORT_B, al

        ;
        ; EOI the PIC
        ;

        mov     al, 20h
        out     20h, al

        pop     ds
        popa

        iret

Keyboard_Handler        endp

Keyboard_Display        proc    near

        pusha
        push    ds

        push    cs
        pop     ds

        ;
        ; Install the keyboard handler
        ;

        pushf
        cli

        mov     ax, 3509h
        int     21h
        mov     word ptr cs:OldKybdInt, bx
        mov     word ptr cs:OldKybdInt + 2, es

        push    ds
        mov     dx, offset Keyboard_Handler
        push    cs
        pop     ds
        mov     ax, 2509h
        int     21h
        pop     ds

        popf

KD_Loop:
        cmp     fDone, 1
        je      SHORT KD_Exit

KD_Display:
        mov     di, ScanStart
        cmp     di, ScanEnd
        je      SHORT KD_Loop

        mov     al, ScanBuffer[di]
        inc     di
        cmp     di, 10h
        jb      SHORT KD_NoWrap
        xor     di, di

KD_NoWrap:
        mov     ScanStart, di

        ;
        ; convert to hex in a cheap sorta way
        ;

        mov     ah, al
        and     al, 0Fh
        shr     ah, 4
        add     ax, 3030h
        cmp     ah, 3Ah
        jb      SHORT @F
        add     ah, 'A' - '0' - 10

@@:
        cmp     al, 3Ah
        jb      SHORT @F
        add     al, 'A' - '0' - 10

@@:
        ; display the string

        xchg    ah, al
        mov     word ptr ScanCode, ax

        mov     dx, offset ScanCode
        mov     ah, 09h
        int     21h
        jmp     SHORT KD_Display

KD_Exit:

        ;
        ; restore keyboard interrupt
        ;

        pushf
        cli

        push    ds
        mov     dx, word ptr cs:OldKybdInt
        mov     ds, word ptr cs:OldKybdInt + 2
        mov     ax, 2509h
        int     21h
        pop     ds

        popf

        pop     ds
        popa

        ; exit

        mov     ax, 4C00h
        int     21h

Keyboard_Display        endp

_TEXT   ends

end Keyboard_Display


