        page    60, 132

;******************************************************************************
	title	DOSXFER - A simple virtual device driver example
;******************************************************************************
;
;   (C) Copyright Bryan A. Woodruff, 1993
;
;   Title:    DOSXFER.386 - Sample virtual device driver
;
;   Module:   DOSXFER.ASM - Core code
;
;   Version:  1.00
;
;   Date:     June 5, 1993
;
;   Author:   Bryan A. Woodruff
;
;******************************************************************************
;
;   Change log:
;
;   DATE      REVISION  DESCRIPTION                                 AUTHOR
;   --------  --------	------------------------------------------  -----------
;    6/ 5/93  1.00      Wrote it.                                   BryanW
;
;******************************************************************************
;
;   Functional Description:
;
;      Demonstrates the use of a semaphore with DOS VMs.
;
;******************************************************************************

        .386p

;***************************************************************************
;                             I N C L U D E S
;***************************************************************************

        .XLIST
        INCLUDE VMM.Inc
        INCLUDE Debug.Inc
        .LIST

        INCLUDE DOSXFER.Inc

;***************************************************************************
;              V I R T U A L  D E V I C E  D E C L A R A T I O N
;***************************************************************************

MAJOR_VERSION   equ     01h
MINOR_VERSION   equ     00h

Declare_Virtual_Device DOSXFER, MAJOR_VERSION, MINOR_VERSION,\
                       DOSXFER_Control, DOSXFER_Device_ID,\
                       Undefined_Init_Order,\
                       DOSXFER_V86_API_Proc, DOSXFER_PM_API_Proc

;***************************************************************************
;                    D E V I C E  F I X E D  D A T A
;***************************************************************************

VxD_LOCKED_DATA_SEG

lpPostMessage           dd      0               ; PostMessage() ptr
fCallBack_Enabled       db      False           ; callbacks enabled?

hNotifyWnd              dw      ?               ; notification window
hNotify_List            dd      ?               ; notification list
hCurrentNode            dd      ?               ; current node
hSemaphore              dd      ?               ; semaphore handle

VxD_LOCKED_DATA_ENDS

;***************************************************************************
;                          D E V I C E  D A T A
;***************************************************************************

VxD_DATA_SEG

DOSXFER_PM_Call_Table LABEL DWORD
        dd      OFFSET32        DOSXFER_Get_Version
        dd      OFFSET32        DOSXFER_PM_Enable_CallBacks
        dd      OFFSET32        DOSXFER_PM_Copy_Data

Max_DOSXFER_PM_Service   equ    ($ - DOSXFER_PM_Call_Table) / 4

DOSXFER_V86_Call_Table LABEL DWORD
        dd      OFFSET32        DOSXFER_Get_Version
        dd      OFFSET32        DOSXFER_V86_Post_Message

Max_DOSXFER_V86_Service  equ    ($ - DOSXFER_V86_Call_Table) / 4

VxD_DATA_ENDS

;***************************************************************************
;                 I N I T I A L I Z A T I O N  C O D E
;***************************************************************************

VxD_ICODE_SEG

;===========================================================================
;
;   DOSXFER_Device_Init:
;
;      Allocates a notification list.
;
;===========================================================================

BeginProc DOSXFER_Device_Init, PUBLIC

        Trace_Out "In DOSXFER_Device_Init"

        pushad

        ;
        ; Create semaphore
        ;

        mov     ecx, 1                                  ; initial token of 1
        VMMCall Create_Semaphore                        ; create the semaphore
        mov     hSemaphore, eax                         ; store handle

IFDEF DEBUG
        jnc     SHORT VDI_Sem_Create_Success
        Debug_Out "Create_Semaphore failed!"
VDI_Sem_Create_Success:
ENDIF

        ;
        ; Allocate notification list
        ;

        mov     eax, LF_Use_Heap + LF_Alloc_Error       ; alloc from heap
        mov     ecx, SIZE DOSXFER_Notify_Struc          ; size of node
        VMMCall List_Create                             ; create list
        jnc     SHORT VDI_NoError                       ; error?

        Debug_Out "DOSXFER: Error creating notification list"
        mov     esi, 0                                  ; clear for error

VDI_NoError:
        mov     hNotify_List, esi                       ; store list handle

VDI_Exit:
        popad
        clc
        ret

EndProc DOSXFER_Device_Init

VxD_ICODE_ENDS


VxD_CODE_SEG

;***************************************************************************
;                         E X P O R T E D  A P I 
;***************************************************************************

;===========================================================================
;
;   DOSXFER_PM_API_Proc
;
;===========================================================================

BeginProc DOSXFER_PM_API_Proc, PUBLIC

        Trace_Out "In DOSXFER_PM_API_Proc"

        VMMCall Test_Sys_VM_Handle
IFDEF DEBUG
        jz      SHORT @f
        Debug_Out "DOSXFER_PM_API_Proc not from SYS VM"
@@:
ENDIF
        jnz     SHORT DOSXFER_PM_Call_BadToo
        movzx   eax, [ebp.Client_DX]                    ; function in DX
        cmp     eax, Max_DOSXFER_PM_Service
        jae     SHORT DOSXFER_PM_Call_Bad
        and     [ebp.Client_EFLAGS], NOT CF_Mask        ; clear carry
        call    DOSXFER_PM_Call_Table[ eax * 4 ]        ; call service
        ret

DOSXFER_PM_Call_Bad:
IFDEF DEBUG
        Debug_Out "DOSXFER ERROR: Invalid function #EAX on DOSXFER_PM_API_Proc"
ENDIF
DOSXFER_PM_Call_BadToo:
        mov     [ebp.Client_EAX], 0FFFFFFFFh

DOSXFER_PM_API_Failed:
        or      [ebp.Client_EFLAGS], CF_Mask            ; set carry
        ret

EndProc DOSXFER_PM_API_Proc

;===========================================================================
;
;   DOSXFER_V86_API_Proc
;
;===========================================================================

BeginProc DOSXFER_V86_API_Proc, PUBLIC

        Trace_Out "In DOSXFER_V86_API_Proc"

        Assert_VM_Handle ebx
        movzx   eax, [ebp.Client_DX]                    ; function in DX
        cmp     eax, Max_DOSXFER_V86_Service
        jae     SHORT DOSXFER_V86_Call_Bad
        and     [ebp.Client_EFLAGS], NOT CF_Mask        ; clear carry
        call    DOSXFER_V86_Call_Table[ eax * 4 ]       ; call service
        ret

DOSXFER_V86_Call_Bad:
IFDEF DEBUG
        Debug_Out "DOSXFER ERROR: Invalid function #EAX on DOSXFER_V86_API_Proc"
ENDIF
        mov     [ebp.Client_EAX], 0FFFFFFFFh

DOSXFER_V86_API_Failed:
        or      [ebp.Client_EFLAGS], CF_Mask            ; set carry
        ret

EndProc DOSXFER_V86_API_Proc

;===========================================================================
;
;   DOSXFER_Get_Version
;
;      Version call for PM API entry point
;
;   ENTRY:      Client_DX = 0
;
;   EXIT:       Client_AX = Version
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc DOSXFER_Get_Version

        Trace_Out "In DOSXFER_Get_Version"

        mov  eax, MAJOR_VERSION SHL 8 or MINOR_VERSION
        mov  [ebp.Client_AX], ax
        ret

EndProc DOSXFER_Get_Version

;===========================================================================
;
;   DOSXFER_PM_Enable_CallBacks
;
;      Enables callbacks to application.
;
;   ENTRY:      Client_DX =     1
;               Client_AX =     Disable/Enable flag
;                               0 = Disabled
;                               ELSE = Enabled
;               Client_CX =     hWnd
;               Client_ES:BX =  CS:IP of callback to PostMessage
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc DOSXFER_PM_Enable_CallBacks
        
        Trace_Out "In DOSXFER_PM_Enable_CallBacks"

        mov     ax, [ebp.Client_AX]
        mov     fCallBack_Enabled, al
        cmp     al, 0
        jz      SHORT @F

        ;
        ; Store PostMessage() address if callbacks are enabled
        ;

        mov     ax, [ebp.Client_BX]
        mov     word ptr lpPostMessage, ax
        mov     ax, [ebp.Client_ES]
        mov     word ptr lpPostMessage + 2, ax
        mov     ax, [ebp.Client_CX]
        mov     hNotifyWnd, ax

@@:
        ret

EndProc DOSXFER_PM_Enable_CallBacks

;===========================================================================
;
;   DOSXFER_PM_Copy_Data
;
;      Copies data to given buffer from given node.  Deletes node
;      from list.
;
;   ENTRY:      Client_DX =     2
;               Client_DS:DI =  dest buffer
;               Client_BX:SI =  source node to copy from
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc DOSXFER_PM_Copy_Data

        Trace_Out "In DOSXFER_PM_Copy_Data"

        pushad

        ;
        ; Unblock DOS VM
        ;

        mov     eax, hSemaphore                         ; get current handle
        VMMCall Signal_Semaphore                        ; signal semaphore

        ;
        ; Copy data to buffer provided by caller
        ;

        Debug_Out "Mapping client pointer"

        Client_Ptr_Flat edi, DS, DI                     ; map dest to flat
        movzx   esi, [ebp.Client_BX]                    ; hiword of node
        shl     esi, 16
        mov     si, [ebp.Client_SI]                     ; loword of node
        mov     ecx, SIZE DOSXFER_Notify_Struc          ; size in bytes
        shr     ecx, 1                                  ; convert to words
        push    ds
        pop     es
        rep     movsw

        ;
        ; Remove node from list
        ;

        mov     ebx, esi
        mov     esi, hNotify_List                       ; handle to list
        VMMCall List_Get_First                          ; get first node

VPCD_Loop:
        cmp     eax, 0                                  ; null handle?
        je      SHORT VPCD_Exit

        cmp     eax, ebx                                ; is this the node?
        jne     SHORT VPCD_Next_Node                    ; nope, keep looking

        VMMCall List_Remove                             ; remove the node
        VMMCall List_DeAllocate                         ; delete node
        jmp     SHORT VPCD_Exit

VPCD_Next_Node:
        VMMCall List_Get_Next                           ; get next node
        jmp     SHORT VPCD_Loop                         ; loop

VPCD_Exit:
        popad
        ret

EndProc DOSXFER_PM_Copy_Data

;===========================================================================
;
;   DOSXFER_V86_Post_Message
;
;      "Notifies" the VxD from the DOS VM.  Schedule a priority
;      VM event for the System VM if notifications have been enabled.
;
;   ENTRY:      Client_DX = 1
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc DOSXFER_V86_Post_Message

        Trace_Out "In DOSXFER_V86_Notify_VxD"

        cmp     fCallback_Enabled, 0                    ; if we're not enabled
        je      DEBFAR VPM_Exit                         ; then get out
        cmp     lpPostMessage, 0                        ; simple NULL check
        je      DEBFAR VPM_Exit                         ; if NULL, get out

        pushad

        ;
        ; Allocate a list node, fill it and schedule an event
        ;

        mov     esi, hNotify_List
        cmp     esi, 0                                  ; do we have a list?
        je      DEBFAR VPM_Exit                         ; no? can't allocate
        VMMCall List_Allocate                           ; allocate a node
        jc      DEBFAR VPM_Exit                         ; fatal error

        mov     hCurrentNode, eax
        mov     edi, eax
        push    eax
        VMMCall Get_System_Time
        mov     [edi.NS_Time], eax
        pop     eax

        VMMCall List_Attach_Tail                        ; add to list

VPM_Call_Event:

        ;
        ; call PostMessage_Event
        ;

        VMMCall Get_Sys_VM_Handle                       ; get sys VM handle
        mov     eax, Low_Pri_Device_Boost               
        mov     ecx, PEF_Wait_For_STI or PEF_Wait_Not_Crit or PEF_Always_Sched
        mov     esi, OFFSET32 DOSXFER_PostMessage_Event
        mov     edx, hCurrentNode                       ; data for event
        VMMCall Call_Priority_VM_Event

        ;
        ; Block current VM on semaphore
        ;

        Debug_Out "Waiting on Semaphore"
        mov     eax, hSemaphore
        mov     ecx, Block_Svc_Ints                     ; services ints
        VMMCall Wait_Semaphore                          ; wait on semaphore
        Debug_Out "Wait complete."
        jmp     SHORT VPM_Exit

VPM_Sem_Create_Failed:
        Debug_Out "Semaphore creation failed!"

VPM_Exit:
        popad
        ret

EndProc DOSXFER_V86_Post_Message

;***************************************************************************
;                         E V E N T  H A N D L E R 
;***************************************************************************

;===========================================================================
;
;   DOSXFER_PostMessage_Event
;
;      This procedure is called when the priority VM event occurs -
;      it ensures that the system VM is active such that we can
;      properly call down to the PostMessage() function.
;
;      Entry:
;         EDX contains reference data (list node handle)
;
;         EBP contains pointer to client register structure
;
;===========================================================================

BeginProc DOSXFER_PostMessage_Event

        Debug_Out "Post Message event received."

        Push_Client_State
        VMMCall Begin_Nest_Exec

        mov     ax, hNotifyWnd                          ; hWnd
        VMMCall Simulate_Push
        mov     ax, WM_NOTIFICATION                     ; uMsg
        VMMCall Simulate_Push
        xor     ax, ax
        VMMCall Simulate_Push                           ; wParam is NULL
        mov     eax, edx
        shr     eax, 16
        VMMCall Simulate_Push                           ; LOWORD( node ptr )
        mov     eax, edx
        VMMCall Simulate_Push                           ; HIWORD( node ptr )

        movzx   edx, WORD PTR lpPostMessage             ; IP (EIP)
        mov     cx, WORD PTR lpPostMessage + 2          ; CS

        VMMCall Simulate_Far_Call
        VMMCall Resume_Exec

        VMMCall End_Nest_Exec
        Pop_Client_State

        ret

EndProc DOSXFER_PostMessage_Event

;***************************************************************************
;               S Y S T E M / V M  S T A T E  C O N T R O L
;***************************************************************************

;===========================================================================
;
;   DOSXFER_System_Exit
;
;      Ensures that the IRQ has been masked and that the STAT!
;      board has been RESET.
;
;===========================================================================

BeginProc DOSXFER_System_Exit

        Trace_Out "In DOSXFER_System_Exit"

        push    edx
        push    eax


        ;
        ; Destroy semaphore
        ;

        mov     eax, hSemaphore
        VMMCall Destroy_Semaphore

        ;
        ; De-allocate notify list
        ;

        mov     esi, hNotify_List
        cmp     esi, 0                                  ; do we have a list?
        je      SHORT @F                                ; no? can't delete it
        VMMCall List_Destroy

@@:
        pop     eax
        pop     edx

        clc
        ret

EndProc DOSXFER_System_Exit

VxD_CODE_ENDS

VxD_LOCKED_CODE_SEG

;***************************************************************************
;                    C O N T R O L / D I S P A T C H 
;***************************************************************************

;===========================================================================
;
;   DOSXFER_Control:
;
;      Control/Dispatch function for DOSXFER.
;
;===========================================================================

BeginProc DOSXFER_Control

        Control_Dispatch Device_Init, DOSXFER_Device_Init
        Control_Dispatch System_Exit, DOSXFER_System_Exit

        clc
        ret

EndProc DOSXFER_Control

VxD_LOCKED_CODE_ENDS

END

;---------------------------------------------------------------------------
;  End of File: dosxfer.asm
;---------------------------------------------------------------------------

