        page    60, 132

;******************************************************************************
	title	TSR - A sample TSR for use with VCALLTSR
;******************************************************************************
;
;   (C) Copyright Bryan A. Woodruff, 1993
;
;   Title:    TSR.EXE - TSR supplement to VCALLTSR
;
;   Module:   TSR.ASM - Core code
;
;   Version:  1.00
;
;   Date:     July 5, 1993
;
;   Author:   Bryan A. Woodruff
;
;******************************************************************************
;
;   Change log:
;
;   DATE      REVISION  DESCRIPTION                                 AUTHOR
;   --------  --------	------------------------------------------  -----------
;    7/ 5/93  1.00      Wrote it.                                   BryanW
;
;******************************************************************************
;
;   Functional Description:
;
;      Demonstrates calling TSR code at ring 0 directly from a VxD.
;
;******************************************************************************

        .286

;==============================================================================
;                     I N C L U D E S   &   E Q U A T E S
;==============================================================================

IntVect         equ     60h

cr              equ     0Dh
lf              equ     0Ah

_TEXT   segment para public 'CODE'

;--------------------------  TSR Data Area  -----------------------------

OldInt_Hook     dd      ?
wCount          dw      0

;----------------------------  TSR Code  --------------------------------

        assume  cs:_TEXT

Hook_Handler proc

        cmp     ax, 4257h                       ; Q: procedure addr query?
        jne     short HH_Chain                  ;    N: attempt to chain

        push    cs
        pop     es
        mov     bx, offset TSR_Proc

HH_Exit:
        iret

HH_Chain:
        push    ax
        mov     ax, word ptr OldInt_Hook
        or      ax, word ptr OldInt_Hook + 2
        pop     ax
        jz      SHORT HH_Exit
        jmp     dword ptr cs:OldInt_Hook    

Hook_Handler endp

;------------------------------------------------------------------------
;
; TSR_Proc
;
; Description:
;
;    This procedure is called by the VxD directly at Ring 0.  It
;    assumes that CS and DS point to the same base address.
;
;------------------------------------------------------------------------

        assume  cs:_TEXT, ds:_TEXT

TSR_Proc proc far

        inc     wCount
        mov     ax, wCount

        retf

TSR_Proc endp

        assume cs:_TEXT, ds:_DATA

        ALIGN   16
InitFence:

;-----------------------  Initialization Code  --------------------------

TSR     proc    far

        mov     ax, _DATA
        mov     ds, ax

        mov     ax, IntVect             ; Print message
        mov     bx, offset HookMsgx
        call    HexAsc

        mov     dx, offset HookMsg
        mov     ah, 09h
        int     21h

        mov     ah, 35h         ; Get old interrupt vector
        mov     al, IntVect
        int     21h
        mov     WORD PTR cs:OldInt_Hook, bx
        mov     WORD PTR cs:OldInt_Hook+2, es

        push    ds              ; Install new handler
        mov     dx, offset Hook_Handler
        push    cs
        pop     ds
        mov     ah, 25h
        mov     al, IntVect
        int     21h
        pop     ds

        mov     ax, 3100h       ; Terminate and Stay Resident
        mov     dx, offset InitFence
        add     dx, 15
        shr     dx, 1
        shr     dx, 1
        shr     dx, 1
        shr     dx, 1
        add     dx, 16
        int     21h

TSR     endp

HexAsc  proc    near    ; converts word to hex ASCII
                        ; call with AX = value,
                        ; DS:BX = address for string
                        ; returns AX, BX destroyed

        push    cx      ; save registers
        push    dx

        mov     dx, 4   ; initialize character counter

HexAsc1:
        mov     cx, 4   ;  isolate next 4 bits
        rol     ax, cl
        mov     cx, ax
        and     cx, 0Fh
        add     cx, '0'
        cmp     cx, '9'
        jbe     HexAsc2
        add     cx, 'A'-'9'-1

HexAsc2:
        mov     [bx], cl
        inc     bx      ; bump string pointer
        dec     dx      ; count characters converted
        jnz     HexAsc1 ; loop, not four yet

        pop     dx      ; restore registers
        pop     cx
        ret             ; back to caller

HexAsc  endp

_TEXT   ends


;-----------------------  Initialization Data  --------------------------

_DATA   segment word public 'DATA'

HookMsg         db      'TSR for use with VCALLTSR.386 example', cr, lf
                db      'Hooking Interrupt: '
HookMsgx        dd      ?
                db      'h', cr, lf, '$'

_DATA   ends

        end     TSR
