        page    60, 132

;******************************************************************************
	title	VCALLTSR - A sample virtual device driver
;******************************************************************************
;
;   (C) Copyright Bryan A. Woodruff, 1993
;
;   Title:    VCALLTSR.386 - Demonstrates calling TSR code at ring 0
;
;   Module:   VCALLTSR.ASM - Core code
;
;   Version:  1.00
;
;   Date:     July 5, 1993
;
;   Author:   Bryan A. Woodruff
;
;******************************************************************************
;
;   Change log:
;
;   DATE      REVISION  DESCRIPTION                                 AUTHOR
;   --------  --------	------------------------------------------  -----------
;    7/ 5/93  1.00      Wrote it.                                   BryanW
;
;******************************************************************************
;
;   Functional Description:
;
;      Demonstrates calling TSR code at ring 0 directly from a VxD.
;
;      This code is based on the design concept by Keith Jin at
;      Microsoft Developer Support.
;
;******************************************************************************

        .386p

;==============================================================================
;                     I N C L U D E S   &   E Q U A T E S
;==============================================================================

        .XLIST
        INCLUDE VMM.Inc
        INCLUDE Debug.Inc
        .LIST

VCALLTSR_Major_Ver      equ     01h
VCALLTSR_Minor_Ver      equ     00h
VCALLTSR_Device_ID      equ     Undefined_Device_ID

;==============================================================================
;           V I R T U A L   D E V I C E   D E C L A R A T I O N
;==============================================================================

Declare_Virtual_Device  VCALLTSR, VCALLTSR_Major_Ver, VCALLTSR_Minor_Ver,\
                        VCALLTSR_Control_Proc, VCALLTSR_Device_ID,\
                        Undefined_Init_Order,,

;==============================================================================
;                  N O N - P A G E A B L E    D A T A
;==============================================================================

VxD_LOCKED_DATA_SEG

dwTSR_Ring0_CS          dd      0
dwTSR_Ring0_EIP         dd      0
dwTSR_Ring0_DS          dd      0
hTimeOut                dd      0

dwTSR_RETF_From_16      label   dword
wTSR_Switch_To_Flat_IP  dw      0
wTSR_Switch_To_Flat_CS  dw      0

VxD_LOCKED_DATA_ENDS

;==============================================================================
;                       P A G E A B L E   C O D E
;==============================================================================

VxD_CODE_SEG

;------------------------------------------------------------------------------
;
;   VCALLTSR_TimeOut
;
;   DESCRIPTION:
;       Event handler for global timeout.  Calls TSR code directly
;       from ring 0.
;
;   ENTRY:
;       EBX = Current VM handle
;       ECX = additional ms since timeout
;       EDX = reference data
;       EBP = -> CRS
;
;   EXIT:
;       Reschedules time-out.
;
;   USES:
;       All registers.
;
;------------------------------------------------------------------------------

BeginProc VCALLTSR_TimeOut

        pushad

        mov     hTimeOut, 0                             ; clear handle

        Trace_Out "Setting up stack frames and preparing to call TSR."


        ;
        ; This stack frame is so we can get back to flat model.
        ;

        push    cs                                      ; save CS
        mov     eax, OFFSET32 VCALLTSR_Back_To_Flat
        push    eax                                     ; save EIP

        ;
        ; This stack frame will get us back to 32-bit code in
        ; the VxD and is addressable via 16:16 for the TSR.
        ;

        push    ds                                      ; save off DS
        push    dwTSR_RETF_From_16

        ;
        ; This is the stack frame used to get us to the TSR
        ; code.  Additionally, DS is setup with a R/W pointer
        ; to the same base address.
        ;

        mov     eax, dwTSR_Ring0_DS
        mov     ds, ax
        push    cs:dwTSR_Ring0_CS
        push    cs:dwTSR_Ring0_EIP
        retf                                            ; go to the TSR

VCALLTSR_Switch_To_Flat:
        pop     ds                                      ; restore DS
        retf                                            ; return to flat

VCALLTSR_STF_Size       equ ($ - VCALLTSR_Switch_To_Flat) - 1

VCALLTSR_Back_To_Flat:

        Trace_Out "Back in flat model. Return from TSR = #AX"

        ;
        ; Reschedule time out event
        ;

        mov     eax, 500                                ; 500 ms timeout
        xor     edx, edx                                ; no data
        mov     esi, OFFSET32 VCALLTSR_TimeOut
        VMMCall Set_Global_Time_Out
        mov     hTimeOut, esi

        popad
        ret

EndProc VCALLTSR_TimeOut

VxD_CODE_ENDS

;==============================================================================
;                             I C O D E
;==============================================================================

VxD_REAL_INIT_SEG

;------------------------------------------------------------------------------
;
;   VCALLTSR_Real_Mode_Init
;
;   DESCRIPTION:
;       Determines if the TSR is loaded and queries for the TSR
;       procedure to call.  If the Int 60h is not hooked, this
;       VxD will fail to load.
;
;   EXIT:
;       EDX contains real-mode TSR entry point or 0 if failure.
;
;   USES:
;       All registers.
;
;------------------------------------------------------------------------------

BeginProc VCALLTSR_Real_Mode_Init

        push    ds
        mov     ax, 0
        mov     ds, ax
        mov     ax, ds:[ 60h * 4 ]
        or      ax, ds:[ 60h * 4 + 2 ]
        pop     ds
        jz      SHORT RI_Abort

        mov     ax, 4257h
        int     60h
        mov     ax, es
        movzx   edx, ax
        shl     edx, 16
        mov     dx, bx                          ; reference data is seg:off
                                                ; of TSR_Proc
        xor     bx, bx
        xor     si, si
        mov     ax, Device_Load_OK
        ret

RI_Abort:
        mov     ah, 9
        mov     dx, OFFSET RI_NotFoundMsg
        int     21h
        xor     bx, bx
        xor     si, si
        xor     edx, edx
        mov     ax, Abort_Device_Load + No_Fail_Message
        ret

RI_NotFoundMsg:
        db      'Error!  VCALLTSR will not load.  TSR.EXE is required.'
        db      0dh, 0ah, '$'

EndProc VCALLTSR_Real_Mode_Init

VxD_REAL_INIT_ENDS

VxD_ICODE_SEG

;------------------------------------------------------------------------------
;
;   VCALLTSR_Sys_Critical_Init
;
;   DESCRIPTION:
;       Allocates necessary GDT selectors.
;
;   ENTRY:
;       EBX = handle to Sys VM
;       EDX = reference data from real-mode init
;
;   EXIT:
;       Carry clear if no error, otherwise set if failure.
;
;   USES:
;       Flags
;
;------------------------------------------------------------------------------

BeginProc VCALLTSR_Sys_Critical_Init
   
        Trace_Out "VCALLTSR: Sys_Critical_Init"

        pushad

        ;
        ; Note: An assumption is made that CS:0 is the base of the TSR.
        ;       Since we don't have a segment size, we'll assume 1 page,
        ;       but this could be handled by using a pointer to a structure
        ;       within the TSR obtained from Exec_Int instead of using
        ;       Real_Mode_Init to gather the information.
        ; 

        mov     eax, edx
        movzx   edx, ax
        mov     dwTSR_Ring0_EIP, edx
        shr     eax, 16
        shl     eax, 4

        push    eax                                     ; save address

        VMMCall _BuildDescriptorDWORDS, < eax, <P_SIZE - 1>, Code_Type,\
                                          <D_DEF16 + D_DPL0>, BDDExplicitDPL >
        VMMCall _Allocate_GDT_Selector, < edx, eax, 0 >
        or      eax, eax
        jnz     SHORT SCI_GotCSSel
        pop     eax
        jmp     SHORT SCI_Failure

SCI_GotCSSel:
        mov     dwTSR_Ring0_CS, eax

        pop     eax                                     ; restore address

        VMMCall _BuildDescriptorDWORDS, < eax, <P_SIZE - 1>, RW_Data_Type,\
                                          <D_DEF16 + D_DPL0>, BDDExplicitDPL >
        VMMCall _Allocate_GDT_Selector, < edx, eax, 0 >
        or      eax, eax
        jz      SHORT SCI_Failure
        mov     dwTSR_Ring0_DS, eax

        VMMCall _BuildDescriptorDWORDS, < <OFFSET32 VCALLTSR_Switch_To_Flat>,\
                                          VCALLTSR_STF_Size, Code_Type,\
                                          <D_DEF32 + D_DPL0>,\
                                          BDDExplicitDPL >
        VMMCall _Allocate_GDT_Selector, < edx, eax, 0 >
        or      eax, eax
        jz      SHORT SCI_Failure
        mov     wTSR_Switch_To_Flat_CS, ax

        mov     eax, 500                                ; 500 ms timeout
        xor     edx, edx                                ; no data
        mov     esi, OFFSET32 VCALLTSR_TimeOut
        VMMCall Set_Global_Time_Out
        mov     hTimeOut, esi

        popad
        clc
        ret

SCI_Failure:

        ;
        ; Free any allocated selectors and exit
        ;

        mov     eax, dwTSR_Ring0_CS
        or      eax, eax
        jz      SHORT SCI_Failure_TryDS
        VMMCall _Free_GDT_Selector, <eax, 0>

SCI_Failure_TryDS:
        mov     eax, dwTSR_Ring0_DS
        or      eax, eax
        jz      SHORT SCI_Failure_TryFlat
        VMMCall _Free_GDT_Selector, <eax, 0>

SCI_Failure_TryFlat:
        movzx   eax, wTSR_Switch_To_Flat_CS
        or      eax, eax
        jz      SHORT SCI_Failure_Exit
        VMMCall _Free_GDT_Selector, <eax, 0>

SCI_Failure_Exit:
        popad
        stc
        ret

EndProc VCALLTSR_Sys_Critical_Init

VxD_ICODE_ENDS

VxD_LOCKED_CODE_SEG

;==============================================================================
;                      N O N P A G E A B L E   C O D E
;==============================================================================

;------------------------------------------------------------------------------
;
;   VCALLTSR_Control_Proc
;
;   DESCRIPTION:
;       Dispatches VMM control messages to the appropriate handlers.
;
;   ENTRY:
;       EAX = Message
;       EBX = VM associated with message
;
;   EXIT:
;       Carry clear if no error (or if not handled by the VxD)
;       or set to indicate failure if the message can be failed.
;
;   USES:
;       All registers.
;
;------------------------------------------------------------------------------

BeginProc VCALLTSR_Control_Proc

        Control_Dispatch Sys_Critical_Init, VCALLTSR_Sys_Critical_Init

        clc
        ret

EndProc VCALLTSR_Control_Proc

VxD_LOCKED_CODE_ENDS

END     VCALLTSR_Real_Mode_Init

;---------------------------------------------------------------------------
;  End of File: vcalltsr.asm
;---------------------------------------------------------------------------

