;**************************************
; Specific video adaptor conditional assembly map to generic conditionals
;

;***************************************
; Default VGA
RESERVEROMC6C7	EQU	-1		; Turn on detection and reserve of ROM

IF1
IFDEF	RESERVEROMC6C7			; Support for ROM w/o signature
%OUT	---- C6, C7 w/o sig support ----
ENDIF
%OUT
ENDIF	;IF1

;----------------------------------------------------------------------------
;                      VDD global flag definitions
;----------------------------------------------------------------------------

; There is a global copy kept in [VT_Flags] and local copy in each CB,
; for efficiency.  The initial value of manyu of these flags are passed
; from real mode initialization code that detects Video adapter types.

; Flags valued from 0-7 cannot be passed from real-mode init.  They
; are initialized to zero.  The size of the video mem is passed in place.

fVT_NoSaveResSysVM      equ     1 SHL 0   ; Save/restore SYS VM enabled
bVT_NoSaveResSysVM      equ     0
fVT_NoSaveResOthers     equ     1 SHL 1   ; Save/restore other VMs enabled
bVT_NoSaveResOthers     equ     1
fVT_Mono                equ     1 SHL 2   ; VDD supports Mono VGA memory (B0-B7)
bVT_Mono                equ     2
fVT_MonoPorts           equ     1 SHL 3   ; VDD supports Mono VGA ports
bVT_MonoPorts           equ     3
fVT_Notify_at_PF        equ     1 SHL 4   ; bkgnd INT 2F notification done
                                          ;    at pfault time
bVT_Notify_at_PF        equ     4
fVT_ResvB0              equ     1 SHL 5   ; Reserve memory from B0 to B7
bVT_ResvB0              equ     5
fVT_SuprsMsg            equ     1 SHL 6   ; Do not message NoFGrnd and
                                          ;    NoAvailPages
bVT_SuprsMsg            equ     6
fVT_NoSuspend           equ     1 SHL 7   ; Do not suspend application, 
                                          ;    map null memory
bVT_NoSuspend           equ     7
fVT_SysVMnot1stBank     equ     1 SHL 8   ; Set if Sys VM not running in
                                          ;    1st bank of video memory
bVT_SysVMnot1stBank     equ     8

fVT_SysVMin2ndBank      equ     1 SHL 9   ; Set if Sys VM is running in 2nd
                                          ;    bank of video memory 
                                          ;    (VT_SysVMnot1stBank must be
                                          ;     set also)
bVT_SysVMin2ndBank      equ     9
		
fVT_RsrvC6C7            equ     1 SHL 31  ; Reserve C6 & C7 as ROM
                                          ;    without a signature
bVT_RsrvC6C7            equ     31


;***************************************************************************
;***************************************************************************
;
;  Video display register equates
;	...Colr are for CGA and EGA color modes
;	...Mono are for EGA/VGA monochrome modes
;
;NAMING CONVENTIONS
; pXXXXX is a I/O port definition
; fXXXXX is a bit mask
; bXXXXX is a bit number
; mXXXXX is a bit mask (multiple fXXXX)
; vXXXXX is a value to output/input to/from a port
; xXXXXX is a specific index

;***************
pCRTCIndxColr	EQU	3D4h		; CRTC index register
pCRTCIndxMono	EQU	3B4h

pCRTCDataColr	EQU	3D5h		; CRTC data register
pCRTCDataMono	EQU	3B5h

;***************
pStatColr	EQU	3DAh		; Status register
pStatMono	EQU	3BAh
fStatEna	EQU	00000001b	;   Video RAM access OK for processor
fStatLPTr	EQU	00000010b	;   Light pen triggered
fStatLPSw	EQU	00000100b	;   Light pen switch is "on"
fStatVRTC	EQU	00001000b	;   Vertical retrace in progress
mStat_IVal	EQU	fStatEna+fStatVRTC  ; Initial value for pseudo status

;***************
pFeatColr	EQU	pStatColr	; Feature control
pFeatMono	EQU	pStatMono
pFeatVGAIn	EQU	3CAh		; VGA feature control read
fFeatFC0	EQU	00000001b	;   FC0
fFeatFC1	EQU	00000010b	;   FC1

;***************
pLPen1Colr	EQU	3DBh		; Light pen latch clear
pLPen1Mono	EQU	3BBh
pLPen2Colr	EQU	3DCh		; Light pen latch set
pLPen2Mono	EQU	3BCh

;***************
pMisc		EQU	3C2h		; Miscellaneous output
IFNDEF CTVGA
pMiscIn 	EQU	3CCh		; VGA read Misc. output
ENDIF
fMiscPNum	EQU	00000001b	;   3Dx port numbers(vs. 3Bx)
fMiscREna	EQU	00000010b	;   Ram enable
fMiscDot	EQU	00001100b	;   Dot clock select
fMiscOSrc	EQU	00010000b	;   Output source
fMiscPage	EQU	00100000b	;   Page bit for odd/even mode
fMiscHPol	EQU	01000000b	;   Horizontal retrace polarity
fMiscVPol	EQU	10000000b	;   Vertical retrace polarity
mMiscCRTC	EQU	fMiscREna+fMiscPage ; Mask for CRTC bits
mMiscMemC	EQU	0FFh - mMiscCRTC    ; Mask for MemC bits

;***************
pStt0EGA	EQU	pMisc		; Miscellaneous status
fStt0Swit	EQU	00010000b	;   Switch sense(addr'd by dot clk sel)
fStt0FCI0	EQU	00100000b	;   Feature card input 0
fStt0FCI1	EQU	01000000b	;   Feature card input 1
fStt0VRTC	EQU	10000000b	;   VRTC

;***************
pAttr		EQU	3C0h		; Attribute(palette) address/data
fVAI_Indx	EQU	10000000b	; sign bit of Attribute index is
bVAI_Indx	EQU	7		;   flag indicating port is index
fVAI_ScOn	EQU	00100000b	; Indicates screen is on
bVAI_ScOn	EQU	5

;***************
pSeqIndx	EQU	3C4h		; Sequencer address
pSeqData	EQU	3C5h		; Sequencer data
fSeq0ARst	EQU	00000001b	;   0:Async reset
fSeq0SRst	EQU	00000010b	;   0:Sync reset
fSeq1DPCh	EQU	00000001b	;   1:Dots per character
fSeq1BWid	EQU	00000010b	;   1:Band Width
fSeq1ShLd	EQU	00000100b	;   1:Shift load(0=every char, 1=skip)
fSeq1DClk	EQU	00001000b	;   1:Dot clock(=1 if halved)
mSeq2WMsk	EQU	00001111b	;   2:Write mask
mSeq3ChrB	EQU	00000011b	;   3:Char map B select(attr bit 3 = 0)
mSeq3ChrA	EQU	00001100b	;   3:Char map A select(attr bit 3 = 1)
fSeq4Alph	EQU	00000001b	;   4:Alpha mode(char gen enabled)
fSeq4ExtM	EQU	00000010b	;   4:Extended memory installed
fSeq4SqAd	EQU	00000100b	;   4:Seq'l mem addr'ing(vs. odd/even)
fSeq4Chain4	EQU	00001000b	;   4:Chain 4 (double odd/even) for
					;     256 color (mode 13)
;***************
pGrp1Pos	EQU	3CCh		; Graphics posn 1(=0)
pGrp2Pos	EQU	3CAh		; Graphics posn 2(=1)
pGrpIndx	EQU	3CEh		; Graphics controller address
pGrpData	EQU	3CFh		; Graphics controller data
fGrp0StV0	EQU	00000001b	;   0:Value for plane 0 if enabled
fGrp0StV1	EQU	00000010b	;   0:Value for plane 1 if enabled
fGrp0StV2	EQU	00000100b	;   0:Value for plane 2 if enabled
fGrp0StV3	EQU	00001000b	;   0:Value for plane 3 if enabled
fGrp1Ena0	EQU	00000001b	;   1:Enable for plane 0
fGrp1Ena1	EQU	00000010b	;   1:Enable for plane 1
fGrp1Ena2	EQU	00000100b	;   1:Enable for plane 2
fGrp1Ena3	EQU	00001000b	;   1:Enable for plane 3
mGrp2Colr	EQU	00001111b	;   2:Color compare value
mGrp3RCnt	EQU	00000111b	;   3:Rotate left count for mode 0
mGrp3Func	EQU	00011000b	;   3:Function for modes 0 and 2
mGrp4RMsk	EQU	00000111b	;   4:Read map select
mGrp5WMod	EQU	00000011b	;   5:Write mode
fGrp5Test	EQU	00000100b	;   5:Test condition
fGrp5RMod	EQU	00001000b	;   5:Read mode(0=plane,1=compare)
fGrp5SqAd	EQU	00010000b	;   5:Seq'l mem addr'ing(vs. odd/even)
fGrp5CGA	EQU	00100000b	;   5:CGA 4 color mode addressing
fGrp6Char	EQU	00000001b	;   6:Char or graphics
fGrp6Chain	EQU	00000010b	;   6:Chain odd maps after even
mGrp6Addr	EQU	00001100b	;   6:Processor view of VRAM start addr
					;	00=A000 for 128kb
					;	01=A000 for 64kb
					;	10=B000 for 32kb
					;	11=B800 for 32kb
bGrp6Addr	EQU	     2		;     First of 2 bits for start addr
mGrp7Colr	EQU	00001111b	;   7:Color don't care for read compare
					;   8:Mask: use latch value vs. data

;****** VGA subsystem enable
pVGAEna 	EQU	3C3h		; VGA enable register
fVGAEna 	EQU	00000001b	;   Enable access to VGA

;****** VGA DAC
pDACWindx	EQU	3C8h		; DAC Write index(R/W)
pDACRindx	EQU	3C7h		; DAC Read index(RO)
pDACState	EQU	pDACRindx	; DAC state
pDACData	EQU	3C9h		; DAC data(3 successive accesses)
pDACMask	EQU	3CAh		; DAC mask

;****** VGA read access to write only registers
pMiscRead	EQU	3CCh		; Port to read Misc output register
pFeatRead	EQU	3CAh		; Port to read Feature output register

;******
; values returned from reads of port pDACState:
;
vDAC_Read_Mode	EQU	0
vDAC_Write_Mode EQU	3


;***************
pCMode		EQU	3C6h		; Control mode(Compaq specific)

