//---------------------------------------------------------------------------
//
//  Module: vddinit.c
//
//  Purpose:
//     Init code and data for VDD.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              3/25/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vddvga.h"
#include "vddphmem.h"
#include "portio.h"

#pragma data_seg( "_IDATA", "ICODE" )

//=============================================================================
//                        I N I T   D A T A
//=============================================================================

// Define instance data for instance data manager

INSTDATASTRUC  Inst_1 = { NULL, NULL, EGA_I_Addr1,
                          EGA_I_Len1, ALWAYS_Field },
               Inst_2 = { NULL, NULL, EGA_I_Addr2,
                          EGA_I_Len2, ALWAYS_Field },
               Inst_3 = { NULL, NULL, EGA_I_Addr3,
                          EGA_I_Len3, ALWAYS_Field },
               Inst_4 = { NULL, NULL, EGA_I_Addr4,
                          EGA_I_Len4, ALWAYS_Field } ;

// vddfonts.c

BOOL   fROMFontFlag = 0 ;
DWORD  dwVidFontLen ;
DWORD  dwVidMsgFontHandle, dwVidMsgFontAddr ;

// vddtio.c
                        
VXDIOTABLE  Vid_IO_Table =
   {
      0x00,                    // MUST be set according to size before
                               // calling vmmInstallMultIOHandlers()

      { { 0x3C0, VIOT_Disp },
        { 0x3C1, VIOT_Disp },  // Attribute controller registers
        { 0x3C2, VIOT_Disp },  // Miscellaneous register
        { 0x3C3, VIOT_Disp },
        { 0x3C4, VIOT_Disp },
        { 0x3C5, VIOT_Disp },  // Sequencer registers
        { 0x3C6, VIOT_Disp },  // COMPAQ special, ignore if bkgnd
        { 0x3C7, VIOT_Disp },  // VGA DAC registers
        { 0x3C8, VIOT_Disp },
        { 0x3C9, VIOT_Disp },
        { 0x3CA, VIOT_Disp },
        { 0x3CB, VIOT_Disp },
        { 0x3CC, VIOT_Disp },  // Grx ctrl pos reg,
                               //    Misc O/P Read Reg
        { 0x3CD, VIOT_Disp },
        { 0x3CE, VIOT_Disp },
        { 0x3CF, VIOT_Disp },  // Graphics controller registers
        { 0x3D0, VIOT_Disp },
        { 0x3D1, VIOT_Disp },
        { 0x3D2, VIOT_Disp },
        { 0x3D3, VIOT_Disp },
        { 0x3D4, VIOT_Disp },
        { 0x3D5, VIOT_Disp },  // CRT controller registers
        { 0x3D6, VIOT_Disp },
        { 0x3D7, VIOT_Disp },
        { 0x3D8, VIOT_Disp },
        { 0x3D9, VIOT_Disp },
        { 0x3DA, VIOT_Disp },  // Feature controller/Status
        { 0x3DB, VIOT_Disp },  // Light pen reset
        { 0x3DC, VIOT_Disp },  // Light pen set
        { 0x3DD, VIOT_Disp },
        { 0x3DE, VIOT_Disp },
        { 0x3DF, VIOT_Disp } }
   } ;

// vddvmmem.c
                        
DWORD aVMPagesBuf[ 9 ] ;         // Buffer for reserving pages

//=============================================================================
//                             I C O D E
//=============================================================================

//------------------------------------------------------------------------
//  BOOL VDD_Sys_Critical_Init
//
//  Description:
//     On entry, interrupts are disabled.  We must be ready
//     to handle message service calls after this is complete.
//     Initialize minimum amount of global data structures.
//     Note that the control block is cleared to 0 when allocated,
//     so any routines that access the control block must handle
//     0 values until it is further initialized in Device_Init
//     and Sys_VM_Init.
//
//     This procedure is called when the VxD_Control_Proc
//     dispatches the Sys_Critical_Init notification from VMM.
//
//     We can notify VMM of failure by returning FALSE,
//     or TRUE will suggest success.
//
//  Parameters:
//     hVM
//        VM handle (system VM handle)
//
//     dwRefData
//        reference data from real-mode init.  
//           LOBYTE( LOWORD( dwRefData ) )
//              # 256Kb blocks-1 available on video adapter
//           (dwRefData & ~0xFF)
//              initial [VT_Flags] value
//
//     pCmdTail
//        pointer to WIN386's command tail
//
//     pCRS
//        pointer to client register structure (not used here)                  
//
//  History:   Date       Author      Comment
//              3/ 9/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VDD_Sys_Critical_Init
(
    DWORD           hVM,
    DWORD           dwRefData,
    PSTR            pCmdTail,
    PCRS_32         pCRS
)
{
   int      iError ;
   PVDDCB   pSysVMCB ;
   APIPROC  pAPIProc ;

   UNUSED_PARAM( pCmdTail ) ;
   UNUSED_PARAM( pCRS ) ;

   if (0 == (dwVidPages = (dwRefData & 0xFF)))
   {
      dwVidPages = vmmGetProfileDecimalInt( NULL, szVDD_SVGA_Mem_Ini,
                                            256, &iError ) ;

      // Convert to # of 256Kb blocks

      dwVidPages = (dwVidPages + 255) >> 8 ;
   }

   // Convert to # of 16Kb (4K by 4 plane) pages

   dwVidPages = dwVidPages << 4 ;

   // This might be a bug, but the VDD sources currently force
   // Video_Pages to 16.

   dwVidPages = 16 ;

   TRACEMSGPARAM( "Video_Pages=#AX\r\n", dwVidPages ) ;

   dwVTFlags = dwRefData & ~0xFF ;
   dwVidFlags = 0 ;

   // Allocate part of VM control block for VDD usage

   if (NULL == (dwVidCBOff =
                   vmmAllocateDeviceCBArea( sizeof( VDDCB ), 0 )))
   {
      DEBUGMSG( "VDD ERROR: Could not allocate control block area!\r\n" ) ;

      VDD_FatalMemoryError() ;
      return FALSE ;
   }

   // Allocate PM API procedure thunk and stuff into DDB

   if (NULL == (pAPIProc =
                   vmmwrapThunkAPIProc( VDD_PM_API )))
   {
      DEBUGMSG( "VDD ERROR: could not allocate API procedure thunk!\r\n" ) ;
      VDD_FatalMemoryError() ;
      return FALSE ;
   }

   VDD_DDB.DDB_PM_API_Proc = (DWORD) pAPIProc ;

   pSysVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   // Pass display type to VM CB

   pSysVMCB -> VDD_TFlags = dwVTFlags ;

   // Init state and physical memory data structures

   VDD_State_SysCriticalInit( hVM, pSysVMCB ) ;

   VDD_PHMem_SysCriticalInit( hVM ) ;

   return ( TRUE ) ;

} // end of VDD_Sys_Critical_Init()

//------------------------------------------------------------------------
//  BOOL VDD_Device_Init
//
//  Description:
//     This is a non-system critical initialization procedure.
//     Specify instance data, and do any other initialization
//     necessary.
//
//  Parameters:
//     hVM
//        VM handle (system VM handle)
//
//     pCmdTail
//        pointer to WIN386 command tail
//
//     pCRS
//        pointer to client register structure
//
//  History:   Date       Author      Comment
//              3/ 9/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VDD_Device_Init
(
    DWORD           hVM,
    PSTR            pCmdTail,
    PCRS_32         pCRS
)
{
   PVDDCB  pSysVMCB ;

   UNUSED_PARAM( pCmdTail ) ;

   // If message mode state already initialized then return

   if (dwVidFlags & fVid_MsgInit)
      return ( TRUE ) ;

#ifdef DEBUG
   VDD_DebugInit() ;
#endif

   pSysVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

#ifdef DEBUG
   if (pSysVMCB -> VDD_TFlags & fVT_RsrvC6C7)
      TRACEMSG( "ROM Pages C6, C7 excluded\r\n" ) ;
#endif

#ifdef SysVMin2ndBank

   // Init latch address

   if (dwVTFlags & fVT_SysVMin2ndBank)
      VDD_PHMem_SetSysLatchAddr( 0x0000FFFF, 0, 0, 0 ) ;

#endif

   // Set up interrupt trapping

   if (!VDD_Int_DeviceInit())
      goto DI_FatalError ;

   // Set up virtual video memory state

   if (!VDD_VMMem_DeviceInit( hVM, pSysVMCB, pCRS ))
      goto DI_FatalError ;

   // Initialize port trapping

   if (!VDD_TIO_DeviceInit( pSysVMCB ))
      goto DI_FatalError ;

   // Set up VM controller state

   if (!VDD_State_DeviceInit())
      goto DI_FatalError ;

   // Set up VM windowed update interaction

   VDD_Proc_DeviceInit() ;

   // Specify instanced RAM (BIOS data area)

   if (!vmmAddInstanceItem( &Inst_1, 0 ))
      goto DI_FatalError ;
   if (!vmmAddInstanceItem( &Inst_2, 0 ))
      goto DI_FatalError ;
   if (!vmmAddInstanceItem( &Inst_3, 0 ))
      goto DI_FatalError ;
   if (!vmmAddInstanceItem( &Inst_4, 0 ))
      goto DI_FatalError ;

   return ( TRUE ) ;

DI_FatalError:

   // Fatal_Error <BadDevice>

   return ( FALSE ) ;

} // end of VDD_Device_Init()

//=============================================================================
//                     I C O D E  for VDDFONTS.C
//=============================================================================

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Font_SetupMsgModeInit
//  
//  Description:
//      Setup for INT 10h mode 3 that VDDState is about to do to
//      init message mode.	We either map null in A0 & A1, or
//      alloc memory and map it at A0 & A1.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to system VM
//  
//      PVDDCB pSysVMCB
//         pointer to system VM control block
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Font_SetupMsgModeInit
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB
)
{
   UINT  i ;

   // map null memory for A0-AF to catch font loads

   for (i = 0; i < 16; i++)
      VDD_VMMem_MapNull( hSysVM, pSysVMCB, 0xA0 + i ) ;

   dwVidMsgFontAddr=
      vmmPageAllocate( &dwVidMsgFontHandle, 2, PG_SYS, NULL, NULL,
                       NULL, NULL, NULL, PageFixed + PageZeroInit ) ;
   if (dwVidMsgFontHandle)
      vmmMapIntoV86( dwVidMsgFontHandle, hSysVM, 0xA0, 2, 0, 0 ) ;

   // map null memory for B8-BF

   for (i = 0; i < 8; i++)
      VDD_VMMem_MapNull( hSysVM, pSysVMCB, 0xB8 + i ) ;

   return TRUE ;

} // VDD_Font_SetupMsgModeInit()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Font_SaveMsgFont
//  
//  Description:
//      Calls VDD_State_Set_MemC_Planar to access the font in
//      plane 2.  A VGA font is made up of an array of 32 byte
//      blocks.  One 32 byte block for each character.  This
//      routine determines the actual font height by looking at
//      the cell height register (index 9) in the CRT controller.
//      It then either calls INT 10h function 1130h to find the
//      location of the ROM font, or it compresses the font from
//      video save memory into allocated memory by only copying
//      "cell height" bytes per character instead of 32.
//
//  Parameters:
//      DWORD hSysVM
//         handle to system VM
//  
//      PVDDCB pSysVMCB
//         sys VM's VDD control block
//
//      PCRS_32
//         pointer to client reg struct
//  
//  Return (BOOL):
//      FALSE if unable to allocate memory for font save 
//         or planar mode not initialized (should never occur)  
//
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Font_SaveMsgFont
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB,
    PCRS_32         pCRS
   
)
{
   BYTE   bROMFontHeight ;
   DWORD  dwFontPages ;
   PBYTE  pFontPhysAddr, pFontLinAddr ;
   UINT   iError ;

   vmmBeginTouch1stMeg() ;

   dwVidMsgModeTextRows = (DWORD) *((PBYTE) BIOS_Line_Count) + 1 ;
   dwVidMsgModeTextCols = (DWORD) *((PBYTE) BIOS_Col_Count) ;

   vmmEndTouch1stMeg() ;

   bVidCharHgt = (pSysVMCB -> VDD_Stt.crtc.C_CharHgt & 0x1F) + 1 ;
   dwVidFontLen = ((DWORD) bVidCharHgt) << 8 ;

   fROMFontFlag = vmmGetProfileBoolean( NULL, szVDD_Use_ROM_Font_Ini,
                                        TRUE, &iError ) ;
   if ((!iError) && (fROMFontFlag))
      fROMFontFlag = 1 ;

   if (!dwVidMsgFontHandle)
      fROMFontFlag = 0 ;

   // fROMFontFlag ... 0 = false, -1 = true, 1 = forced true

   if (fROMFontFlag)
   {
      CRS_32  CRS ;

      // use ROM font

      if (bVidCharHgt >= 16)
         bROMFontHeight = 16 ;
      else if (bVidCharHgt < 14)
         bROMFontHeight = 8 ;
      else
         bROMFontHeight = 14 ;

      if ((bROMFontHeight > bVidCharHgt) && !dwVidMsgFontHandle)
      {
         TRACEMSG( "VDD: forced to programmed font...\r\n" ) ;
         goto Pack_Font ;
      }

      vmmSaveClientState( hSysVM, pCRS, &CRS ) ;
      vmmBeginNestExec( hSysVM, pCRS ) ;

      switch (bROMFontHeight)
      {
         case 8:
            ((PCRS_8) pCRS) -> Client_BH = 3 ;

         case 14:
            ((PCRS_8) pCRS) -> Client_BH = 2 ;
            break ;

         case 16:
            ((PCRS_8) pCRS) -> Client_BH = 6 ;
            break ;
      }

      // simulate an int 10h to the VM to get a font pointer 

      ((PCRS_16) pCRS) -> Client_AX = 0x1130 ;
      vmmExecInt( hSysVM, pCRS, 0x10 ) ;
      pFontPhysAddr =
         (PBYTE) (((DWORD) pCRS -> Client_ES << 4) +
                    (DWORD) ((PCRS_16) pCRS) -> Client_BP) ;
      vmmEndNestExec( hSysVM, pCRS ) ;
      vmmRestoreClientState( hSysVM, pCRS, &CRS ) ;

      if (fROMFontFlag == 1)
         goto Use_ROM ;
      else
      {
         vmmGetDeviceV86PagesArray( NULL, aVMPagesBuf, NULL ) ;

         // claimed by a device?

         if (aVMPagesBuf[ 0 ] & (1 << ((DWORD) pFontPhysAddr >> 12)))
            goto No_ROM_Font ;
         else
         {
Use_ROM:
            if (-1 == (pFontLinAddr =
                  (PBYTE) vmmMapPhysToLinear( pFontPhysAddr,
                                              dwVidFontLen, 0 )))
            {
               TRACEMSG( "VDD: cannot address video ROM font memory.\r\n" ) ;

No_ROM_Font:
               if (dwVidMsgFontHandle)
                  goto Pack_Font ;
               else
               {
                  DEBUGMSG( "VDD: save message font failed\r\n" ) ;

                  vmmFatalMemoryError() ;
               }
            }
            else
            {
               bVidCharHgt = bROMFontHeight ;
               dwVidMsgFont = (DWORD) pFontLinAddr ;
               if (dwVidMsgFontHandle)
                  vmmPageFree( dwVidMsgFontHandle, 0 ) ;

               VDD_VMMem_DisableNull( hSysVM, pSysVMCB ) ;
               return TRUE ;
            }
         }
      }
   }
   else
   {
      UINT  i, j ;
      PSTR  pDst, pSrc ;

Pack_Font:
      //
      // Pack the font that was programmed by
      // the ROM during the mode change.
      //

      dwFontPages = (dwVidFontLen + P_SIZE - 1) >> 12 ;

      // force no packing for 2 or more pages

      if (dwFontPages > 1)
         bVidCharHgt = 32 ;

      dwVidMsgFont =
         vmmPageAllocate( &dwVidMsgFontH, 2, PG_SYS, NULL, NULL,
                          NULL, NULL, NULL, PageFixed + PageZeroInit ) ;
      if (!dwVidMsgFontH)
      {
         DEBUGMSG( "VDD: save message font failed\r\n" ) ;

         vmmFatalMemoryError() ;
      }

      pSrc = (PSTR) dwVidMsgFontAddr ;
      pDst = (PSTR) dwVidMsgFont ;

      for (i = 0; i < 256; i++)
      {
         for (j = 0; j < bVidCharHgt; j++)
            *pDst++ = *pSrc++ ;

         // adjust

         pSrc += 32 - bVidCharHgt ;
      }

      // map null memory over the 2 pages of font save
      // memory previously mapped at A0 & A1 so that we
      // can free the pages.  Calling VDD_VMMem_Disable_Null
      // will unmap the null memory later.

      VDD_VMMem_MapNull( hSysVM, pSysVMCB, 0xA0 ) ;
      VDD_VMMem_MapNull( hSysVM, pSysVMCB, 0xA1 ) ;

      // free tmp font mem

      vmmPageFree( dwVidMsgFontHandle, 0 ) ;

      VDD_VMMem_DisableNull( hSysVM, pSysVMCB ) ;
      return TRUE ;
   }

} // VDD_Font_SaveMsgFont()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Font_InitComplete
//  
//  Description:
//      Unlock msg mode font save page and SYS VM's save memory,
//      if system is running with an intelligent (direct hardware)
//      pageswap device.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to system VM
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Font_InitComplete
(
    DWORD           hSysVM
)
{
   PVDDCB pSysVMCB ;

   if (dwVidMsgFontH)
      vmmPageUnLock( dwVidMsgFontH, (dwVidFontLen + P_SIZE - 1) >> 12,
                     0, PageLockedIfDP ) ;

   pSysVMCB = (PVDDCB) (hSysVM + dwVidCBOff) ;
   VDD_VMMem_UnlockSaveMem( pSysVMCB, 4 ) ;

   return ( TRUE ) ;

} // VDD_Font_InitComplete()

//=============================================================================
//                     I C O D E  for VDDINT.C
//=============================================================================

//--------------------------------------------------------------------------
//  
//  BOOL VDD_Int_DeviceInit
//  
//  Description:
//      Set up interrupt vectors and Int 10h stub.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to system VM 
//  
//      PVDDCB pSysVMCB
//         system VM VDD control block
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_Int_DeviceInit
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB
)
{
   EVENTPROC   pCallBack ;
   DWORD       dwCallBackAddr ;
   V86INTPROC  pIntHandler ;

   VDD_Int_MungeInt10h( hSysVM ) ;


   if (NULL == (pIntHandler = vmmwrapThunkV86IntProc( VDD_Int_10h )))
   {
      DEBUGMSG( "Failed to allocate INT 10h callback thunk.\r\n" ) ;
      return FALSE ;
   }

   vmmHookV86IntChain( 0x10, pIntHandler ) ;

   if (NULL == (pIntHandler = vmmwrapThunkV86IntProc( VDD_Int_16h )))
   {
      DEBUGMSG( "Failed to allocate INT 16h callback thunk.\r\n" ) ;
      return FALSE ;
   }

   vmmHookV86IntChain( 0x16, pIntHandler ) ;

   if (NULL == (pIntHandler = vmmwrapThunkV86IntProc( VDD_Int_09h )))
   {
      DEBUGMSG( "Failed to allocate INT 09h callback thunk.\r\n" ) ;
      return FALSE ;
   }

   vmmHookV86IntChain( 0x09, pIntHandler ) ;

   // hook PM int 10h chain

   vmmGetPMIntVector( 0x10, &dwVidI10NextCS, &dwVidI10NextEIP ) ;

   if (NULL == (pCallBack = vmmwrapThunkEventProc( VDD_Int_10h_PM )))
   {
      DEBUGMSG( "Failed to allocate INT 10h PM callback thunk.\r\n" ) ;
      return FALSE ;
   }

   dwCallBackAddr = vmmAllocatePMCallBack( pCallBack, 0 ) ;
   vmmSetPMIntVector( 0x10, HIWORD( dwCallBackAddr ),
                      LOWORD( dwCallBackAddr ) ) ;

   // hook int 2Fh chains

   if (NULL == (pIntHandler = vmmwrapThunkV86IntProc( VDD_Int_2Fh )))
   {
      DEBUGMSG( "Failed to allocate INT 2Fh callback thunk.\r\n" ) ;
      return FALSE ;
   }

   vmmHookV86IntChain( 0x2F, pIntHandler ) ;

   // hook PM int 2Fh chain

   vmmGetPMIntVector( 0x2F, &dwVidI2FNextCS, &dwVidI2FNextEIP ) ;

   if (NULL == (pCallBack = vmmwrapThunkEventProc( VDD_Int_2Fh_PM )))
   {
      DEBUGMSG( "Failed to allocate INT 2Fh PM callback thunk.\r\n" ) ;
      return FALSE ;
   }

   dwCallBackAddr = vmmAllocatePMCallBack( pCallBack, 0 ) ;
   vmmSetPMIntVector( 0x2F, HIWORD( dwCallBackAddr ),
                      LOWORD( dwCallBackAddr ) ) ;

   return TRUE ;

} // VDD_Int_DeviceInit()

//=============================================================================
//                     I C O D E  for VDDPHMEM.C
//=============================================================================

//------------------------------------------------------------------------
//  VOID VDD_PHMem_SysCriticalInit
//
//  Description:
//     Map A0 through BF as physical=linear in SYS VM, allocate
//     the video page table and create the page handle list.
//
//  Parameters:
//     DWORD hVM
//       handle to System VM
//
//  History:   Date       Author      Comment
//              4/10/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_PHMem_SysCriticalInit
(
    DWORD           hVM
)
{
   UNUSED_PARAM( hVM ) ;

   vmmTraceOut( "VDD_PHMem_SysCriticalInit\r\n" ) ;

   // Set up physical addresses for the video memory

   dwVidPhysA0000 = vmmMapPhysToLinear( 0xA0000, 0x20000, 0 ) ;
   if (dwVidPhysA0000 == (DWORD) -1)
   {
      vmmDebugOut( "VDD: Cannot address video memory (A0000h-BFFFFh)\r\n" ) ;
   }

   // Set up other special physical memory pointers

   dwVidLatchAddr = dwVidPhysA0000 + (0xAFFFF - 0xA0000) ;
   bVidLatchBank = 0 ;

   // Allocate Video_Page_Table

   dwVidPageTable =
      vmmHeapAllocate( (dwVidPages << PPD_Index_Factor), HeapZeroInit ) ;

   if (0 == dwVidPageTable)
   {
      vmmDebugOut( "VDD failed to alloc video page table!\r\n" ) ;
   }

   // Create list for page handles

   if (NULL ==
          (VDD_PageHandle_List = vmmListCreate( sizeof( PAGEDATA ), 0 )))
   {
      vmmDebugOut( "VDD failed to create page handle list!\r\n" ) ;

      // Fatal error!!!
      // VMMJmp  Fatal_Memory_Error
   }

} // end of VDD_PHMem_SysCriticalInit()

//=============================================================================
//                     I C O D E  for VDDPROC.C
//=============================================================================

//--------------------------------------------------------------------------
//  
//  VOID VDD_Proc_DeviceInit
//  
//  Description:
//      VDDPROC initialization
//  
//  Parameters:
//      DWORD hSysVM
//         handle to SysVM
//  
//      PVDDCB pSysVMCB
//         pointer to Sys VM control block
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/29/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Proc_DeviceInit
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB
)
{

#ifdef 0

; Get window update timeout value
	push	edi
	mov	eax, 50 		    ; default to 50msecs
	mov	edi, OFFSET32 VDD_Time_Ini
	xor	esi, esi		    ; Use [Win386] section
	VMMcall Get_Profile_Decimal_Int
	mov	[Vid_Win_Update_Time],eax   ; Set Window update time
	add	eax,eax
	mov	[Vid_Mode_Update_Time],eax  ; Twice as long for mode changes
	pop	edi

; Get scroll update counter value
	push	edi
	mov	eax,2
	mov	edi, OFFSET32 VDD_VM_Scroll_Freq_Ini
	xor	esi,esi
	VMMCall	Get_Profile_Decimal_Int
	or	eax,eax			    ; has someone been mischievous?
	jnz	SHORT NoPrankster
	mov	eax,2
NoPrankster:
	mov	[Vid_Scroll_Freq],eax
	pop	edi

; Hook Release Time slice so we can do Window updates when app is idle
	mov	esi,OFFSET32 @VDD_VMIdle
	mov	eax,Release_Time_Slice
	VMMCall Hook_Device_Service
	mov	[Vid_VM_Idle_Chain],esi
	ret

#endif

} // VDD_Proc_DeviceInit()

//=============================================================================
//                     I C O D E  for VDDSTATE.C
//=============================================================================

//------------------------------------------------------------------------
//  VOID VDD_State_SysCriticalInit
//
//  Description:
//     Set up system VM as CRTC and MemC owner, adjust
//     Msg_Pseudo_VM handle.
//
//  Parameters:
//     DWORD hVM
//        handle to system VM
//
//     PVDDCB pSysVMCB
//        Sys VM's control block.
//
//  History:   Date       Author      Comment
//              4/10/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_State_SysCriticalInit
(
    DWORD           hVM,
    PVDDCB          pSysVMCB
)
{
   // System VM has the focus and is running on hardware.

   hVidFocusVM = hVM ;
   hVidCRTCVM = hVM ;
   hVidMemCVM = hVM ;

   // The pseudo VM handle must be adjusted to have the same
   // offset to the start of the VDD info as normal VM CBs.

   hVidMsgPseudoVM = ((DWORD) &VidMsgPseudoCB) - dwVidCBOff ;

} // end of VDD_State_SysCriticalInit()

//=============================================================================
//                     I C O D E  for VDDTIO.C
//=============================================================================

//--------------------------------------------------------------------------
//  
//  BOOL VDD_TIO_DeviceInit
//  
//  Description:
//      Sets up trapping for VDD I/O ports.
//  
//  Parameters:
//      PVDDCB pSysVMCB
//  
//  Return (BOOL):
//      FALSE if error.
//  
//  History:   Date       Author      Comment
//             10/21/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VDD_TIO_DeviceInit
(
    PVDDCB          pSysVMCB
)
{
   UINT       i ;
   VXDIOPROC  pIOThunk ;
      
   // Set initial state for I/O trapping in sys VM

   pSysVMCB -> VDD_PIF |=
      fVidNoTrpTxt | fVidNoTrpLRGrfx | fVidNoTrpHRGrfx ;

   pSysVMCB -> VDD_Flags |= fVDD_IOT ;  // ports trapped in sys VM
   pSysVMCB -> VDD_TrapEnable = -1 ;    // all ports trapped

   // First, allocate callback thunk for I/O proc...

   if (NULL ==
       (pIOThunk = vmmwrapThunkVXDIOProc( VIOT_Disp )))
   {
      DEBUGMSG( "VDD: Unable to allocate I/O callback!\r\n" ) ;
      return ( FALSE ) ;
   }

   Vid_IO_Table.IOT_Ports =
      (sizeof( Vid_IO_Table ) - sizeof( WORD )) / sizeof( VXDIOENTRY ) ;

   for (i = 0; i < Vid_IO_Table.IOT_Ports; i++)
      Vid_IO_Table.IOT_Entries[ i ].VxD_IO_Proc = pIOThunk ;

   return (vmmInstallMultIOHandlers( &Vid_IO_Table )) ;

} // VDD_TIO_DeviceInit()

//=============================================================================
//                     I C O D E  for VDDVMMEM.C
//=============================================================================

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_CheckB0B7Write( VOID )
//
//  Description:
//     Attempts to write to B0000 and return result.
//
//  Parameters:
//     None.
//
//  Return Value:
//     BOOL
//        TRUE if successful, otherwise FALSE
//
//  History:   Date       Author      Comment
//              4/12/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_CheckB0B7Write()
{
   BYTE  bIndx, bSave ;
   WORD  wOrig, wNew, wTest ;
   PWORD pVidPtr ;


   // No DUALDISPLAY specified, check for memory at B0000 with
   // VGA not mapped at B0000.

   // Force mapping at B8000... we're going to write to memory
   // so we don't want anyone to see it.

   bIndx = inp( pGrpIndx ) ;
   outp( pGrpIndx, 6 ) ;
   bSave = inp( pGrpIndx + 1 ) ;
   outp( pGrpIndx + 1, bSave | 0x0C ) ;
   pVidPtr = (PWORD) (dwVidPhysA0000 + 0xB0000 - 0xA0000) ;
   wOrig = *pVidPtr ;
   wTest = wOrig ^ 0x5555 ;
   *pVidPtr = wTest ;
   wNew = *pVidPtr ;
   *pVidPtr = wOrig ;

   // Restore graphics controller

   outp( pGrpIndx + 1, bSave ) ;
   outp( pGrpIndx, bIndx ) ;
   return (wNew == wTest) ;

} // end of VDD_VMMem_CheckB0B7Write()

//------------------------------------------------------------------------
//  VOID VDD_VMMem_ReserveB0B7Pages( PVDDCB pSysVMCB )
//
//  Description:
//     Reserves pages B0-B7 in the V86 global assignment array.
//
//  Parameters:
//     PVDDCB pSysVMCB
//        pointer to system VM control block.
//
//  Return Value:
//     None.
//
//  History:   Date       Author      Comment
//              4/12/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_VMMem_ReserveB0B7Pages
(
    PVDDCB          pSysVMCB
)
{
#ifdef DEBUG
   vmmGetDeviceV86PagesArray( NULL, aVMPagesBuf, NULL ) ;
   if (0 == (aVMPagesBuf[ 0xA0/32 ] & 0xFF0000))
#endif
      if (vmmAssignDeviceV86Pages( 0xB0, 8, NULL, NULL ))
         return ;
   vmmTraceOut( "VDD: Mono pages already reserved\r\n" ) ;
   pSysVMCB -> VDD_TFlags &= ~(fVT_Mono + fVT_ResvB0) ;
   return ;

} // end of VDD_VMMem_ReserveB0B7Pages()

//------------------------------------------------------------------------
//  BOOL VDD_VMMem_DeviceInit( DWORD hVM, PVDDCB pSysVMCB, PCRS_32 pCRS )
//
//  Description:
//     Allocate pagemaps for SYS VM.  Read auto restore INI
//     switches.  Reserve and hook video pages A0-AF, B8-BF and
//     B0-B7, if necessary.  Hook ROM pages C6 and C7 on machines
//     where the video ROM BIOS is not properly identified.
//     And hook in the VDD's page fault handler, VDD_PFault.
//
//  Parameters:
//     DWORD hVM
//        handle to system VM
//
//     PVDDCB pSysVMCB
//        pointer to system VM's control block
//
//     PCRS_32 pCRS
//        pointer to client register structure
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL VDD_VMMem_DeviceInit
(
    DWORD           hVM,
    PVDDCB          pSysVMCB,
    PCRS_32         pCRS
)
{
   int        i, iError ;
   BOOL       fDual ;
   V86PHPROC  pVDD_PFault ;

   if (NULL ==
         (pVDD_Error_CvtFullScreen =
            vmmwrapThunkEventProc( VDD_Error_CvtFullScrn )))
   {
      vmmDebugOut( "VDD ERROR: Could not allocate VDD_Error thunk!\r\n" ) ;
      vmmFatalError( ) ;
      return ( FALSE ) ;
   }

   VDD_VMMem_AllocPageMaps( pSysVMCB, dwVidPages ) ;

#if 0

   // The following switch was removed from 3.1.
   // It wasn't implemented due to lack of testing.

   if (!vmmGetProfileBoolean( NULL, szVDD_SysVM_SaveRes_Ini,
                              FALSE, &iError ))
   {
#endif
      pSysVMCB -> VDD_TFlags |= fVT_NoSaveResSysVM ;
      vmmTraceOut( "auto save/restore of SYS VM disabled\r\n" ) ;
#if 0
   }
#endif

   // Check for no suspend in background INI switch

   if (vmmGetProfileBoolean( NULL, szVDD_NoSusp_Ini, FALSE, &iError ))
      pSysVMCB -> VDD_TFlags |= fVT_NoSuspend ;

   // Check for no message on suspend or video
   // corruption in background INI switch

   if (!vmmGetProfileBoolean( NULL, szVDD_SuspMsg_Ini, TRUE, &iError ))
      pSysVMCB -> VDD_TFlags |= fVT_SuprsMsg ;

   // Check for auto save/restore of other intelligent VMs INI switch

   if (!vmmGetProfileBoolean( NULL, szVDD_VM_SaveRes_Ini, TRUE, &iError ))
   {
      vmmTraceOut( "auto save/restore of other VMs disabled\r\n" ) ;
      pSysVMCB -> VDD_TFlags |= fVT_NoSaveResOthers ;
   }

   // Check for background notify time INI switch

   if (vmmGetProfileBoolean( NULL, szVDD_BkGnd_Notify_Time,
                             TRUE, &iError ))
   {
      vmmTraceOut( "notify bkgnd at pfault time\r\n" ) ;
      pSysVMCB -> VDD_TFlags |= fVT_Notify_at_PF ;
   }

   //
   // Reserve and hook pages A0-AF and B8-BF  AND
   //
   //   IF
   // 	1) Secondary VDD does not exist (it will reserve it)
   // 	    AND
   // 	2) "DUALDISPLAY=YES" OR Secondary display is detected
   // 	    AND
   // 	3) Other VxD (e.g. Upper Memory Blocks) has not reserved pages B0-B7
   //
   //   THEN Reserve and map pages B0-B7 physical for secondary display
   //
   //   ELSE IF
   // 	1) Secondary VDD does not exist (it will reserve it)
   // 	    AND
   // 	2) "DUALDISPLAY=NO" OR Secondary display is not detected
   // 	    AND
   // 	3) NOT ("VGAMONOTEXT=NO")
   // 	    AND
   // 	4) "VGAMONOTEXT=YES" OR Windows started when video in mono mode
   // 	    AND
   // 	5) Other VxD (e.g. Upper Memory Blocks) has not reserved pages B0-B7
   //
   //   THEN Reserve and hook pages B0-B7 for VGA mono support
   //
   //   ELSE VGA Mono not supported, pages B0-B7 not handled
   //
   // If user sets "DUALDISPLAY=NO" with monochrome adapter attached the
   // 	results are undefined.
   //

   // Note that fVT_Mono has been set if user started Windows
   // in mono video mode

   if (VDD2_GetVersion() != -1)
   {
      fDual = vmmGetProfileBoolean( NULL, szVDD_2nd_Ini, FALSE, &iError ) ;

      // If iError is set, there is nothing specified for
      // "DUALDISPLAY", so we need to check for mono display.

      if (iError)
      {
         if (VDD_VMMem_CheckB0B7Write())
         {
            // Determined it's a dual display
            pSysVMCB -> VDD_TFlags &= ~fVT_Mono ;
            pSysVMCB -> VDD_TFlags |= fVT_ResvB0 ;
            VDD_VMMem_ReserveB0B7Pages( pSysVMCB ) ;
         }
         else
         {
            // assume Mono support

            pSysVMCB -> VDD_TFlags |= fVT_Mono ;
            if (vmmGetProfileBoolean( NULL, szVDD_Mono_Text_Ini,
                                      FALSE, &iError ))
               VDD_VMMem_ReserveB0B7Pages( pSysVMCB ) ;
            else
               pSysVMCB -> VDD_TFlags &= ~(fVT_Mono + fVT_ResvB0) ;
         }
      }
      else
      {
         if (fDual)
         {
            pSysVMCB -> VDD_TFlags &= ~fVT_Mono ;
            pSysVMCB -> VDD_TFlags |= fVT_ResvB0 ;
            VDD_VMMem_ReserveB0B7Pages( pSysVMCB ) ;
         }
         else
         {
            // assume Mono support

            pSysVMCB -> VDD_TFlags |= fVT_Mono ;
            if (vmmGetProfileBoolean( NULL, szVDD_Mono_Text_Ini,
                                      FALSE, &iError ))
               VDD_VMMem_ReserveB0B7Pages( pSysVMCB ) ;
            else
               pSysVMCB -> VDD_TFlags &= ~(fVT_Mono + fVT_ResvB0) ;
         }
      }
   }
   else
      pSysVMCB -> VDD_TFlags &= ~(fVT_Mono + fVT_ResvB0) ;

   // Either B0-B7 is reserved or not, and if reserved,
   // fVT_Mono is set, VGA mono supported else B0-B7 should
   // be mapped physical (dual display).

#ifdef DEBUG
   if (pSysVMCB -> VDD_TFlags & fVT_Mono)
      vmmTraceOut( "VGA Mono text support\r\n" ) ;
   if (pSysVMCB -> VDD_TFlags & fVT_ResvB0)
      vmmTraceOut( "Secondary Mono text support\r\n" ) ;
#endif
   vmmGetDeviceV86PagesArray( NULL, aVMPagesBuf, NULL ) ;
   if (aVMPagesBuf[ 0xA0/32 ] & 0xFF00FFFF)
   {
      vmmDebugOut( "VDD ERROR: Video pages already allocated\r\n" ) ;
      vmmFatalError( szVDD_Str_CheckVidPgs ) ;
      return ( FALSE ) ;
   }
   if (!vmmAssignDeviceV86Pages( 0xA0, 16, NULL, NULL ))
   {
      vmmDebugOut( "VDD ERROR: Could not allocate A0-AF video pages\r\n" ) ;
      vmmFatalError( szVDD_Str_CheckVidPgs ) ;
      return ( FALSE ) ;
   }
   if (!vmmAssignDeviceV86Pages( 0xB8, 8, NULL, NULL ))
   {
      vmmDebugOut( "VDD ERROR: Could not allocate B8-BF video pages\r\n" ) ;
      vmmFatalError( szVDD_Str_CheckVidPgs ) ;
      return ( FALSE ) ;
   }

   // Many VGAs have ROM without signature at C6, C7 pages.  Either
   // the real mode INIT set fVT_RsrvC6C7 or there is a SYSTEM.INI entry.

   // MUST! how to handle non-existant values?

   // MUST! fVT_RsrvC6C7 should be cleared if user set this key in
   // the SYSTEM.INI to FALSE.

   if (vmmGetProfileBoolean( NULL, szVDD_Rsrv_C6C7_Ini, FALSE, &iError ))
      pSysVMCB -> VDD_TFlags |= fVT_RsrvC6C7 ;

   if (pSysVMCB -> VDD_TFlags & fVT_RsrvC6C7)
   {
      CRS_32  CRS ;
      DWORD   dwROMSeg ;

      vmmSaveClientState( hVM, pCRS, &CRS ) ;
      vmmBeginNestExec( hVM, pCRS ) ;
      ((PCRS_16) pCRS) -> Client_AX = 0x1130 ;     // get font ptr
      ((PCRS_8) pCRS) -> Client_BH = 0x02 ;
      vmmExecInt( hVM, pCRS, 0x10 ) ;
      dwROMSeg = (DWORD) (pCRS -> Client_ES) ;     // ES = ROM segment
      vmmEndNestExec( hVM, pCRS ) ;
      vmmRestoreClientState( hVM, pCRS, &CRS ) ;

      vmmTraceOutParam( "Video ROM segment=#EAX\r\n", dwROMSeg ) ;

      // Reserve additional space for ROM without ROM ID

      if (((dwROMSeg & 0xFF00) >= 0xC000) && ((dwROMSeg & 0xFFF) == 0))
      {
         // C0000 or E0000...


         if (0 == (aVMPagesBuf[ (dwROMSeg >> 8) / 32 ] & 0xC0))
         {
            dwReservedROMPage = (dwROMSeg >> 8) + 6 ;
            if (!vmmAssignDeviceV86Pages( dwReservedROMPage, 2, NULL, NULL ))
            {
               vmmTraceOut( "VDD: C6-C7 ROM pages already allocated\r\n" ) ;
               pSysVMCB -> VDD_TFlags &= ~fVT_RsrvC6C7 ;
            }

         }
         else
         {
            vmmTraceOut( "VDD: C6-C7 ROM pages already allocated\r\n" ) ;
            pSysVMCB -> VDD_TFlags &= ~fVT_RsrvC6C7 ;
         }
      }
   }

   if (NULL == (pVDD_PFault = vmmwrapThunkV86PHProc( VDD_PFault )))
   {
      vmmDebugOut( "VDD ERROR: Could not allocate VDD_PFault thunk!\r\n" ) ;
      vmmFatalError( ) ;
      return ( FALSE ) ;
   }

   // Hook graphics pages

   for (i = 0; i < 16; i++)
      vmmHookV86Page( 0xA0 + i, pVDD_PFault ) ;

   // Hook text pages

   for (i = 0; i < 8; i++)
      vmmHookV86Page( 0xB8 + i, pVDD_PFault ) ;

   // If mono supported, hook the pages

   if (pSysVMCB -> VDD_TFlags & fVT_Mono)
      for (i = 0; i < 8; i++)
         vmmHookV86Page( 0xB0 + i, pVDD_PFault ) ;

   // Update global flags

   dwVTFlags = pSysVMCB -> VDD_TFlags ;

   // Init SYS VM control block

   VDD_VMMem_CreateVM( hVM, pSysVMCB ) ;

   // Map the ROM in system VM

   VDD_VMMem_VMCriticalInit( hVM, pSysVMCB ) ;

   return ( TRUE ) ;

} // end of VDD_VMMem_DeviceInit()

//--------------------------------------------------------------------------
//  
//  BOOL VDD_VMMem_InitComplete
//  
//  Description:
//      Init dwVidPageAllocFlags & call VDD_Font_InitComplete.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to system VM 
//  
//      PSTR pCmdTail
//         WIN386's command tail
//  
//      PCRS_32 pCRS
//         client regs
//  
//  Return (BOOL):
//      TRUE on success
//  
//  History:   Date       Author      Comment
//             10/26/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL CDECL VDD_VMMem_InitComplete
(
    DWORD           hSysVM,
    PSTR            pCmdTail,
    PCRS_32         pCRS
)
{
   UNUSED_PARAM( pCmdTail ) ;
   UNUSED_PARAM( pCRS ) ;

   dwVidPageAllocFlags = PageZeroInit | PageLockedIfDP ;
   return ( VDD_Font_InitComplete( hSysVM ) ) ;

} // VDD_VMMem_InitComplete()

//---------------------------------------------------------------------------
//  End of File: vddinit.c
//---------------------------------------------------------------------------

