//---------------------------------------------------------------------------
//
//  Module: vddmsg.c
//
//  Purpose:
//     String definitions for VDD.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              4/ 9/93   BryanW      Rewritten in 'C'.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>

#pragma data_seg( "_LDATA", "CODE" ) 

// The following message results from an incorrect installation of
// device drivers.  Another device driver is conflicting with the VDD.

char szVDD_Str_CheckVidPgs[] = "Cannot run Windows because of video device "\
   "conflict.  You need to run the Setup program again. " ;

// The following is a message that results from incorrect initialization
// of the VDD, probably due to insufficient memory, incorrect display
// adapter or bad files.  User should verify that enough system memory
// is available and that SETUP was completed properly.

char szVDD_Str_BadDevice[] = "Video initialization failed.  You need to run "\
   "the Setup program again." ;

// INI entry for window update time.

char szVDD_Time_Ini[] = "WindowUpdateTime" ;

// INI entry to disable message when background application is suspended or
// its display becomes corrupt due to inability to handle video memory access.
// Default is true, the message will be generated.
//
// This allows users to run background applications that they know get
// suspended sometimes but don't want to deal with the error message.

char szVDD_SuspMsg_Ini[] = "VideoBackgroundMsg" ;

// INI entry to allow VM display corruption in background, rather
// than suspending the VM.  When this occurs the "NoMainMem" message
// will be generated unless the SuspMsg INI switch is set false.
// Default is false, the application will be suspended.

char szVDD_NoSusp_Ini[] = "VideoSuspendDisable" ;

// INI entry for initial number of text rows

char szVid_Text_Rows_Ini[] = "ScreenLines" ;
char szVid_Text_Rows_Sect[] = "NonWindowsApp" ; // SYSTEM.INI section name

//
// INI entries that detemine whether the VDDVGA will manage the monochrome
// text (B0000h-B7000h) address area or leave it alone, allowing
// other devices to use the address space.  Note that the
// VGAMONOTEXT entry is new for 3.1.  It provides for the case
// where there is no secondary display or secondary display
// device but the user still does not want VGA mono support.
// Also note that for 3.1 it is not an error if we attempt to
// provide VGA mono support but some other device has grabbed
// the address space already (we just won't support VGA mono).
// Following is a description of the logic followed:
// 
// IF
// 1) Secondary VDD does not exist (it will reserve it)
//    AND
// 2) "DUALDISPLAY=YES" OR Secondary display is detected
//    AND
// 3) Other VxD (e.g. Upper Memory Blocks) has not reserved pages B0-B7
// 
// THEN Reserve and map pages B0-B7 physical for secondary display.
//     The memory is address space is not available for any use other
//     than accessing the physical memory at pages B0-B7.
//
// ELSE IF
// 1) Secondary VDD does not exist (it will reserve it)
//    AND
// 2) "DUALDISPLAY=NO" OR Secondary display is not detected
//    AND
// 3) NOT ("VGAMONOTEXT=NO")
//    AND
// 4) "VGAMONOTEXT=YES" OR Windows started when video in mono mode
//     AND
// 5) Other VxD (e.g. Upper Memory Blocks) has not reserved pages B0-B7
//
// THEN Reserve and hook pages B0-B7 for VGA mono support.
//
// ELSE VGA Mono not supported, pages B0-B7 not handled.  The address space
//    is available for any use.  This may be by a 2nd VDD or by a
//    DOS UMB or by translation buffers from protected mode to virtual
//    mode.
//
// If user sets "DUALDISPLAY=NO" with monochrome adapter attached the results
// are undefined.   In this case we may provide VGA mono support which
// will conflict with the existing secondary monochrome display.
//

char szVDD_2nd_Ini[] = "DualDisplay" ;
char szVDD_Mono_Text_Ini[] = "VGAMonoText" ;

//
// Some VGA adapters have ROM at C6 and C7 but do not put a ROM signature
// in the ROM, so the V86MMGR determines that it is free address space and
// uses it for buffers.  If this is set TRUE, or the VDD detects one of these
// ROMs, we will exclude the memory from C6000h to C7FFFh from other usage
// and the memory will always be addressable.   If this is set FALSE, we will
// not do anything and the address space may be reused by the system or used
// for any other reason.
//

char szVDD_Rsrv_C6C7_Ini[] = "ReserveVideoROM" ;

// INI entry for specifying the amount of video memory on the VGA adapter
// Default is 256, use this switch to specify the amount of memory
// installed on the VGA adapter in kilo bytes (Kb)

char szVDD_SVGA_Mem_Ini[] = "SVGAmemory" ;

//
// INI entry for disabling automatic save/restore of other display API aware
// applications.
// Default is TRUE, so that automatic save/restore is enabled.  Setting this
// switch to FALSE, avoids allocating save memory for applications which
// notify enhanced mode Windows that they know how to restore/repaint their
// screen themselves.   By setting this switch to FALSE, switching to some
// applications may be somewhat slower, because the application will be
// responsible for repainting its screen which usually takes more time than
// if Windows handles the screen restore automatically.
//

char szVDD_VM_SaveRes_Ini[] = "AutoRestoreScreen" ;

//
// INI entry for adjusting the scrolling frequency for dos apps. VDD will attempt
// to scroll once in these many times. This does NOT provide a guarantee that
// an update will be generated once for n lines. Updates might be generated
// more often due to asynch events such as mouse clicks.
//

char szVDD_VM_Scroll_Freq_Ini[] = "ScrollFrequency" ;

//
// INI entry for changing when the automatic save/restore background
// notification is sent.  For displays that don't have special display
// hardware (e.g. blitters), it is possible to hold off sending a
// background notification into a VM until it actually page faults.
// The VDD will then send the notification and map NULL memory so the
// VM can proceed, but doesn't access actual video memory.  For displays
// that do have special display hardware, waiting until a page fault
// may be to late, because it may be trying to access video memory
// directly on the adapter, so there may not even be a page fault!  On
// these displays it is necessary to send the background notification
// as early as possible.
//
// Setting this switch to FALSE forces the early notification, setting it
// to TRUE specifies that the notification can wait until page fault time.
//
// The default setting for VGA is TRUE, and for 8514 it is FALSE.
//

char szVDD_BkGnd_Notify_Time[] = "BkGndNotifyAtPFault" ;

//
// INI entry for disabling the use of the soft font stored in the video ROM for
// the message mode font (the font used in full screen message text and in the
// switcher screen displayed when switching away from a DOS VM.)  Set this
// switch to FALSE to disable.  This switch would be used on machines where a
// different soft font is programmed by a TSR or device driver, etc. or where
// the video ROM is banked so that the font isn't always addressable.  The
// first possibility would be identified by seeing that the font in message
// mode is different from the normally used font.  The second possibility would
// be identified by seeing random dots and weird shapes that appear in rows and
// columns like text, but aren't readable; this condition indicates that the
// font is not programmed correctly.
//

char szVDD_Use_ROM_Font_Ini[] = "UseROMfont" ;

char szVDD_SysVM_SaveRes_Ini[] = "AutoRestoreWindows" ;

//---------------------------------------------------------------------------
//  End of File: vddmsg.c
//---------------------------------------------------------------------------

