        page    60, 132

;******************************************************************************
	title	VDDVGA - real mode initialization
;******************************************************************************
;
;   (C) Copyright Woodruff Software Systems, 1993
;
;   Title:    VDDVGA - real mode initialization
;
;   Module:   VDDRINIT.ASM
;
;   Version:  1.00
;
;   Date:     April 11, 1993
;
;   Author:   Bryan A. Woodruff
;
;******************************************************************************
;
;   Change log:
;
;   DATE      REVISION  DESCRIPTION                                 AUTHOR
;   --------  --------	------------------------------------------  -----------
;    4/11/93  1.00      Pulled from original DDK sources.           BryanW
;
;******************************************************************************
;
;   Functional Description:
;
;      Performs real mode initialization/detection for VDD.
;
;******************************************************************************

        .386p

;==============================================================================
;                     I N C L U D E S   &   E Q U A T E S
;==============================================================================

.xlist
	INCLUDE VMM.INC
	INCLUDE VDDDEF.INC
.list

VxD_REAL_INIT_SEG

;******************************************************************************
;
;   VDD_Real_Init
;
;   DESCRIPTION:
;	This does some detection of video hardware/software and returns
;	an initial value for VT_Flags
;
;   ENTRY:
;	AX = VMM version number
;	BX = flags (see VMM.INC)
;	SI = environment segment
;	EDX = reference data when device specified from INT 2F (not applicable)
;
;   EXIT:
;	AX = Device_Load_Ok
;	BX = 0
;	EDX = Initial value for VT_Flags
;
;   USES:
;	Flags, AX, BX, CX, DX, BP, SI, DI
;
;==============================================================================

BeginProc VDD_Real_init

	mov	ax, 1130h
	mov	bh, 2
	int	10h		    ; es:bp -> ROM font, so use es as ROM segment

;********************************
; VGA real mode adapter detection
;	Note that individual adapter detection routines must either
;	jump to VRI_Exit with EDX = VT_Flags value or preserve
;	the ES and DX register.  ES is the adapter ROM segment (usually C000h)
;	and DX is the status register port address (for resetting the
;	attribute controller index/value toggle).

;*******
; Load status port value in DX
;	This is to prevent duplication in OEM specific detection
;
	mov	dx,pStatColr
	xor	edx,edx
;
; DX = initial flags for VDD
;	Add in fVT_Mono if video currently in mono mode
;
VRI_Exit:
	push	dx
	mov	ax,0F00h
	int	10h			; Get BIOS mode
	pop	dx
	cmp	al,7			; Q: Mono text mode?
	je	SHORT VRI_Mono		;   Y: Assume want Mono VGA support
	cmp	al,0Fh			; Q: Mono graphics mode?
	jne	SHORT VRI_NotMono	;   N: Assume do not want Mono VGA
VRI_Mono:
	or	edx,fVT_Mono	        ; Probably want Mono VGA support

VRI_NotMono:
	xor	bx,bx			; No pages to exclude
	mov	si,bx			; No instance data
	mov	ax, Device_Load_Ok	; Continue with load
	ret

EndProc VDD_Real_init

VxD_REAL_INIT_ENDS

end VDD_Real_Init
