//---------------------------------------------------------------------------
//
//  Module: vddsvc.c
//
//  Purpose:
//     Virtual Display Driver services and grabber API.
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Rewritten in 'C'.
//
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include <shell.h>
#include "vddvga.h"
#include "portio.h"

#pragma data_seg( "_LDATA", "CODE" ) 

DWORD hVidMsgVM = 0 ;               // Msg VM handle

DWORD dwVidMsgModeTextRows = 0 ;
DWORD dwVidMsgModeTextCols = 0 ;

BYTE bVidScrnAttr = 0 ;             // Screen clear attribute
BYTE bVidMsgAttr = 0 ;              // Msg text mode attribute

WORD wPIFSave = 0 ;                 // Pre VM create PIF bits save

DWORD dwVidShadowMemStatusPtr = 0 ;  

//------------------------------------------------------------------------
//  VOID VDD_Get_Version
//
//  Description:
//     Returns version number and device ID.
//
//  Parameters:
//      PREGS pRegs 
//         no parameters in registers
//
//  Return Value:
//     pRegs -> eax = version
//        (AH = major version, AL = minor version)
//
//     pRegs -> esi = ptr to 8 byte ID string
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_Get_Version
(
    PREGS           pRegs
)
{
   static char szVDDType[] = "VIDEOVGA" ;

   pRegs -> eax = VDD_VerNum ;
   pRegs -> esi = (DWORD) szVDDType ;

} // end of VDD_Get_Version()

//------------------------------------------------------------------------
//  VOID VDD_Get_GrbVersion
//
//  Description:
//     Returns version number and device ID.
//
//  Parameters:
//      PREGS pRegs 
//         no parameters in registers
//
//  Return Value:
//     pRegs -> eax = version
//        (AH = major version, AL = minor version)
//
//     pRegs -> esi = ptr to 8 byte ID string
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_Get_GrbVersion
(
    PREGS           pRegs
)
{
   VDD_Get_Version( pRegs ) ;

} // end of VDD_Get_GrbVersion()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Begin_Msg_Mode
//  
//  Description:
//      This routine is called when a Begin_Message_Mode device control
//      is received. It is followed by a call to the VDD_Msg_ClrScrn
//      routine, which initializes (or sets up an event to initialize)
//      the screen.
//
//      When the display is finished, an End_Message_Mode device control
//      is made to restore the focus VM's display (focus may change while
//      in message mode, but the screen will not change until
//      End_Message_Mode is called).
//
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx
//            VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Begin_Msg_Mode
(
    PREGS           pRegs
)
{
   hVidMsgVM = pRegs -> ebx ;
   
   // clear carry and return

   pRegs -> eflags &= ~EFLAGS_CF ;

} // VDD_Begin_Msg_Mode()

//--------------------------------------------------------------------------
//  
//  VOID VDD_End_Msg_Mode
//  
//  Description:
//      This routine is called after a sysmodal message box is complete.
//      It restore's the VM's video state.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx
//            VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_End_Msg_Mode
(
    PREGS           pRegs
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

#ifdef DEBUG
   if (!hVidMsgVM)
      DEBUGMSG( "VDD: End_Msg_Mode not in Msg mode\r\n" ) ;

   if (hVidMsgVM != pRegs -> ebx)
   {
      DEBUGMSG( "VDD: End_Msg_Mode, but not current owner\r\n" ) ;
      pRegs -> eflags &= ~EFLAGS_CF ;
      return ;
   }
#endif

   dwVidFlags |= fVid_SysVMI ;
   if (pRegs -> ebx == hVidMsgVM)
      VDD_Msg_ClrScrn( pRegs ) ;
   dwVidFlags &= ~fVid_MsgA ;
   hVidMsgVM = 0 ;

   hVM = hVidFocusVM ;
   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;
   VDD_Ctl_SetFocus( hVM, pVMCB ) ;

   // clear carry and exit

   pRegs -> eflags &= ~EFLAGS_CF ;

} // VDD_End_Msg_Mode()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Msg_ClrScrn
//  
//  Description:
//      This routine is called to initialize the screen for putting up
//      messages.  It is a service called by the shell.  The screen is
//      cleared immediately.  A Begin_Message_Mode device control call
//      must be issued before this service is utilized.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx = VM handle
//         pRegs -> eax = background attribute
//  
//  Return (VOID):
//      pRegs -> eax = width in columns
//      pRegs -> edx = height in rows
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Msg_ClrScrn
(
    PREGS           pRegs
)
{
   int    i ;
   PWORD  pScreenArea ;


   bVidMsgAttr = bVidScrnAttr =
      (BYTE) ((((pRegs -> eax) & 0x07) << 4) | 0x07) ;

   VDD_State_ScrnOff( pRegs -> ebx ) ;
   VDD_State_SetMessageMode() ;

   pScreenArea = (PWORD) (dwVidPhysA0000 + (0xB8000 - 0xA0000)) ;
   for (i = 0; i < dwVidMsgModeTextRows * dwVidMsgModeTextCols; i++)
      *pScreenArea++ = (WORD) (((WORD) bVidScrnAttr) << 8) + ' ' ;

   VDD_State_ScrnOn( pRegs -> ebx ) ;

   pRegs -> eax = dwVidMsgModeTextCols ;
   pRegs -> edx = dwVidMsgModeTextRows ;

} // VDD_Msg_ClrScrn()

//------------------------------------------------------------------------
//  VOID VDD_Msg_ForColor
//
//  Description:
//     After calling VDD_Begin_Msg_Mode, this service sets up the
//     foreground attribute.
//
//  Parameters:
//     PREGS pRegs
//        pRegs -> ebx = VM handle
//        pRegs -> eax = color (for EGA/VGA driver, a text mode attribute)
//
//  Return Value:
//     Nothing.
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_Msg_ForColor
(
    PREGS           pRegs
)
{
   bVidMsgAttr = ((pRegs -> eax) & 0x0F) | (bVidMsgAttr & 0xF0) ;

} // end of VDD_Msg_ForColor()

//------------------------------------------------------------------------
//  VOID VDD_Msg_BakColor
//
//  Description:
//     After calling VDD_Begin_Msg_Mode, this service sets up the
//     background attribute.
//
//  Parameters:
//     PREGS pRegs
//        pRegs -> ebx = VM handle
//        pRegs -> eax = color (for EGA/VGA driver, a text mode attribute)
//
//  Return Value:
//     Nothing.
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_Msg_BakColor
(
    PREGS           pRegs
)
{
   bVidMsgAttr = (((pRegs -> eax) & 0x07) << 4) | (bVidMsgAttr & 0x0F) ;

} // end of VDD_Msg_BakColor()

//------------------------------------------------------------------------
//  VOID VDD_Msg_TextOut
//
//  Description:
//     After calling VDD_Begin_Msg_Mode and setting up the foreground
//     and background colors, this service puts a string on the screen.
//
//  Parameters:
//     PREGS pRegs
//        pRegs -> esi = address of string
//        pRegs -> ecx = length of string
//        pRegs -> eax = row start
//        pRegs -> edx = column start
//        pRegs -> ebx = VM handle
//
//  Return Value:
//     Nothing.
//
//  History:   Date       Author      Comment
//              4/11/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID CDECL VDD_Msg_TextOut
(
    PREGS           pRegs
)
{
   UINT   i ;
   PBYTE  pStr ;
   PWORD  pScreenArea ;

   if (pRegs -> ecx)
   {
      VDD_State_SetMessageMode( pRegs -> ebx ) ;

      pScreenArea =
         (PWORD) (dwVidPhysA0000 + 0xb8000 - 0xa0000 +
                  pRegs -> eax * 160 + pRegs -> edx * 2) ;
      pStr = (PBYTE) (pRegs -> esi) ;

      for (i = 0; i < pRegs -> ecx; i++)
         *pScreenArea++ = (((WORD) bVidMsgAttr) << 8) | (WORD) (*pStr++) ;
   }

} // end of VDD_Msg_TextOut()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Msg_SetCursPos
//  
//  Description:
//      After calling the VDD_Begin_Msg_Mode this routine sets the
//      cursor position.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> eax
//            row
//
//         pRegs -> edx
//            column
//
//         pRegs -> ebx
//            VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Msg_SetCursPos
(
    PREGS           pRegs
)
{
   WORD   wPort ;
   DWORD  dwPosition ;

   VDD_State_SetMessageMode( pRegs -> ebx ) ;

   dwPosition = pRegs -> eax ;
   if (dwPosition >= dwVidMsgModeTextRows)
      dwPosition = (DWORD) -1 ;
   else
   {
      dwPosition *= dwVidMsgModeTextCols ;
      dwPosition += pRegs -> edx ;
   }

   wPort = VDD_State_GetCRTCIndexPort() ;

   outp( wPort, 0x0F ) ;
   outp( wPort + 1, (BYTE) dwPosition ) ;
   outp( wPort, 0x0E ) ;
   outp( wPort + 1, (BYTE) (dwPosition >> 8) ) ;

} // VDD_Msg_SetCursPos()

//--------------------------------------------------------------------------
//  
//  VOID VDD_PIF_State
//  
//  Description:
//      Informs VDD about PIF bits for newly created VM.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx = VM handle
//         pRegs -> ax = PIF bits
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_PIF_State
(
    PREGS           pRegs
)
{
   PVDDCB  pVMCB ;

   if (vmmTestSysVMHandle( pRegs -> ebx ))
      wPIFSave = (WORD) pRegs -> eax ;
   else
   {
      pVMCB = (PVDDCB) (pRegs -> ebx + dwVidCBOff) ;
      if (pVMCB -> VDD_PIF != (WORD) pRegs -> eax)
      {
         pVMCB -> VDD_PIF = (WORD) pRegs -> eax ;
         VDD_TIO_SetTrap( pRegs -> ebx, pVMCB ) ;
      }
   }

} // VDD_PIF_State()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Set_HCurTrk
//  
//  Description:
//      Sets flag passed to VMDOSAPP indicating that VMDOSAPP should
//      maintain cursor position withing display window for this
//      application.  This is called by the keyboard driver when a
//      keyboard interrupt is simulated into a VM.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx = VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Set_HCurTrk
(
    PREGS           pRegs
)
{
   PVDDCB  pVMCB ;

   pVMCB = (PVDDCB) (pRegs -> ebx + dwVidCBOff) ;
   pVMCB -> VDD_Flags |= fVDD_HCurTrk ;

} // VDD_Set_HCurTrk()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Get_GrabRtn
//  
//  Description:
//      Returns address of video grab routine. The grab routine is
//      called by the SHELL device when the appropriate hot key is
//      pressed by the user. It makes a copy of the visible screen
//      and controller state of the current VM. That copy is then
//      accessible via VDD_Get_GrbState and VDD_Get_GrbMem services.
//  
//  Parameters:
//      PREGS pRegs
//  
//  Return (VOID):
//      pRegs -> esi = address of grab routine.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Get_GrabRtn
(
    PREGS           pRegs
)
{
   pRegs -> esi = (DWORD) VDD_Grab ;

} // VDD_Get_GrabRtn()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grab
//  
//  Description:
//      Attempt to grab a copy of the VM's screen and controller
//      state and schedule an event for VMDOSAPP.  If unsupported
//      mode or memory cannot be allocated, then nothing is done.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx = VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Grab
(
    PREGS           pRegs
)
{
   PVDDCB pVMCB ;

   pVMCB = (PVDDCB) (pRegs -> ebx + dwVidCBOff) ;

   VDD_State_SaveCRTCOwner( pRegs -> ebx, pVMCB ) ;
   if (VDD_Grab_DoGrab( pRegs -> ebx, pVMCB ))
   {
      VDD_State_CopyCtlrState( pRegs -> ebx, pVMCB ) ;
      VDD_VMDA_Grab( pRegs -> ebx, pVMCB ) ;
   }

} // VDD_Grab()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Hide_Cursor
//  
//  Description:
//      If EAX is non zero, sets a hide cursor flag else clears the flag.
//      This is so that, if the mouse is using a hardware cursor, it can
//      turn off that cursor while the VM is windowed(since the VM will
//      no longer own the mouse).
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> eax = 0 if cursor should be displayed,
//                     != 0 if cursor should NOT be displayed
//
//         pRegs -> ebx = VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Hide_Cursor
(
    PREGS           pRegs
)
{
   PVDDCB pVMCB ;

   pVMCB = (PVDDCB) (pRegs -> ebx + dwVidCBOff) ;

   pVMCB -> VDD_Flags &= ~fVDD_Hide ;
   if (pRegs -> eax)
   {
      pVMCB -> VDD_Flags |= fVDD_Hide ;
      *(PWORD)(&pVMCB -> VDD_SttCopy.crtc.C_CAddrH) = 0xFFFF ;
   }

} // VDD_Hide_Cursor()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Query_Access
//  
//  Description:
//     If the VM were to access its video memory now, would the VDD be
//     able to handle it?  Some devices, such as the Mouse device would
//     like to call software in the VM that will access the video.  However,
//     if this will cause a video display error, it would be better to
//     let the side affect occur rather than generate the error (in the
//     case of the mouse, the mouse cursor will be left on when the
//     VM is displayed in a window).  This call should only be done
//     just before returning to the VM in question (i.e. in a VM Event)
//     since subsequent setfocus calls may make the result incorrect.
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx = VM handle
//  
//  Return (VOID):
//      CF = 0 if video access is OK.
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Query_Access
(
    PREGS           pRegs
)
{
   pRegs -> eflags &= ~EFLAGS_CF ;

} // VDD_Query_Access()

//--------------------------------------------------------------------------
// 
// Displaying VM's video memory in a window
// 
//  Parameter passing
//
// There are several services supplied in order to efficiently render
// a VM's video memory into a window. These routines are called by the
// grabber. Since the grabber runs in a virtual machine, parameters
// are passed in the Client registers and in VM memory pointed to by
// the Client registers. Other services available use protect mode
// registers.
// 
//  Updating windowed VMs
//
// The first step is for the SHELL to call Set_VMType with a parameter
// indicating that the VM is to be windowed. When the VM is no longer
// windowed, Set_VMType is called again. When the VMState is not
// windowed, the Get_Mod call will always return no changes and the
// video update message will never be generated.
// 
// Note that when a VM's video state changes, including controller state
// changes such as cursor movement and memory modification, the VDD will
// send VMDOSAPP a display update message. All the changes made to the
// video state will accumulate and be reported by Get_Mod until a
// Clear_Mod call is made. There will only be one display update
// message per Clear_Mod call.  Also note that immediately after
// a mode change, the Update message is delayed to allow the application
// to do a series of mode changes before things settle down.
// 
//--------------------------------------------------------------------------


//--------------------------------------------------------------------------
//  
//  VOID VDD_Set_VMType
//  
//  Description:
//      This service is used to inform the VDD of a VM's type. The
//      parameter explicitly passed is the Windowed flag. Implicitly
//      passed are the VM status flags, Exclusive and Background. This
//      should be called prior to running the VM and each time thereafter
//      that any of the VM paramaters are modified. Note that for a system
//      critical Set_Focus this routine may not be called before the
//      Set_Focus. In that case, the VDD is responsible for doing an
//      implied Set_VMType(not windowed).
//  
//  Parameters:
//      PREGS pRegs
//         pRegs -> ebx
//            VM handle whose state is to change
//
//         pRegs -> eax
//            state flag (= non-zero if changing to a windowed VM)
//
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Set_VMType
(
    PREGS           pRegs
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

   hVM = pRegs -> ebx ;
   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   // assume it's windowed

   pVMCB -> VDD_Flags |= fVDD_Win ;
   pVMCB -> VDD_Mod_Rect = -1 ;

   if (pRegs -> eax)
   {
      // windowed

      if (pVMCB -> VDD_Flags & fVDD_NoSaveRes)
         VDD_VMMem_ChangeSaveMem( hVM, pVMCB ) ;
      pVMCB -> VDD_Flags &= ~fVDD_NoSaveRes ;

      // when windowed from full-screen, force cursor tracking
      // in the window by setting up an update event (which has
      // the cursor change).

      // force screen update

      pVMCB -> VDD_Mod_Flag_Save = fVDD_M_Ctlr + fVDD_M_Curs ;

      // force horz. cursor tracking

      pVMCB -> VDD_Flags |= fVDD_HCurTrk ;

      // update event to track cursor 

      VDD_Update_Window( hVM, pVMCB ) ;
   }
   else
   {
      // full screen

      pVMCB -> VDD_Flags &= ~fVDD_Win ;

      VDD_Grab_DeleteTextCopyBuf( hVM, pVMCB ) ;
      VDD_ClrVMTimeOut( hVM, pVMCB ) ;
      if (pVMCB -> VDD_Flags  & fVDD_CanRestore)
         VDD_Int_SetCanRestore( hVM, pVMCB ) ;
   }

} // VDD_Set_VMType()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Check_Update_Soon
//  
//  Description:
//      This service can be used by VxDs to direct the VDD to check for 
//      updates soon. This service is useful only for windowed dos apps.
//      VxDs should call this service in response to user actions which
//      generally generate a change in the display of apps. eg. the VDD
//      can call this service when it detects a key press/release etc.
//  
//  Parameters:
//      PREGS pRegs
//
//         pRegs -> EBX
//            VM handle
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Check_Update_Soon
(
    PREGS           pRegs
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

   hVM = pRegs -> ebx ;
   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   if ((!pVMCB -> VDD_Recent_User_Action) &&
       (pVMCB -> VDD_Flags & fVDD_Win))
   {
      VDD_ClrVMTimeOut( hVM, pVMCB ) ;
      pVMCB -> VDD_Time_Out_Handle =
         vmmSetVMTimeOut( hVM, 5, pVDD_VMTimeOut, (DWORD) pVMCB ) ;
      pVMCB -> VDD_Recent_User_Action++ ;
      pVMCB -> VDD_Check_Upd_Soon = 4 ;
   }

} // VDD_Check_Update_Soon()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Get_ModTime
//  
//  Description:
//      This routine is used to determine if any video activity has
//      occurred.  The poll device uses it to determine if the VM is
//      idle.
//  
//  Parameters:
//      PREGS pRegs
//
//         pRegs -> ebx
//            VM handle
//  
//  Return (VOID):
//      pRegs -> eax
//         system time at last video modification
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Get_ModTime
(
    PREGS           pRegs
)
{
   DWORD   hVM ;
   PVDDCB  pVMCB ;

   hVM = pRegs -> ebx ;
   pVMCB = (PVDDCB) (hVM + dwVidCBOff) ;

   if ((pVMCB -> VDD_Flags & fVDD_ModeChange) ||
        VDD_State_ChangeQuery( hVM, pVMCB ))
   {
       // middle of mode change... busy now.

       pRegs -> eax = vmmGetLastUpdatedSystemTime() ;
   }
   else
   {
      // update VDD_ModTime if changes

      VDD_VMMem_SaveDirtyState( hVM, pVMCB ) ;
      pRegs -> eax = pVMCB -> VDD_ModTime ;
   }

} // VDD_Get_ModTime()

//--------------------------------------------------------------------------
//  
//  VOID VDD_PM_API
//  
//  Description:
//     This routine dispatches to the various service routines required
//     by the grabber.  It is called only by the grabber in the SYS VM.
//
//  Parameters:
//      DWORD hVM
//         current VM handle = SYS VM handle
//  
//      PCRS_32 pCRS
//         pointer to client stack frame
//            Client_EBX = VM handle (except for function 0)
//            Client_AX = device function
//
//  
//  Return (VOID):
//      dispatched to function handler
//  
//  History:   Date       Author      Comment
//             10/24/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

GRBSVCPROC Grabber_JumpTable[] =
{
   VDD_Grb_Version,        // Get version number and ID string ptr
   VDD_Get_Mem,            // Get selector and map to video memory
   VDD_Get_State,          // Get video state
   VDD_Get_Mod,            // Get state changes
   VDD_Clear_Mod,          // Clear state change accumulation
   VDD_Free_Mem,           // Free memory
   VDD_Get_GrbMem,         // Get selector and map to video grab mem
   VDD_Free_Grab,          // Free grab memory and controller state
   VDD_Get_GrbState,       // Get grab video state
   VDD_Unlock_App          // Unlock the APP
} ;                        

VOID CDECL VDD_PM_API
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   if (vmmTestSysVMHandle( hVM ))
   {
      if (((PCRS_16) pCRS) -> Client_AX <= GRB_Unlock_APP)
      {
         BOOL    fVMValid ;
         DWORD   dwFunc ;
         DWORD   hCallVM ;
         PVDDCB  pVMCB ;

         dwFunc = ((PCRS_16) pCRS) -> Client_AX ;
         if (dwFunc > 0)
         {
            hCallVM = pCRS -> Client_EBX ;
            pVMCB = (PVDDCB) (hCallVM + dwVidCBOff) ;

            fVMValid = vmmValidateVMHandle( hCallVM ) ;
         }

         // call service

         Grabber_JumpTable[ dwFunc ]( hVM, pCRS, hCallVM, pVMCB, fVMValid ) ;
      }
      else
      {
         switch (((PCRS_16) pCRS) -> Client_AX)
         {
            case Private_DspDrvr_1:
               VDD_Init_DspDrvRing0() ;
               break ;

            case DspDrvr_Version:
               VDD_SVC_DspVersion( pCRS ) ;
               break ;

            case DspDrvr_Addresses:
               VDD_SVC_SetAddresses( hVM, pCRS ) ;
         }
      }
   }
   else
      DEBUGMSG( "VDD: API call made by non system VM\r\n" ) ;


} // VDD_PM_API()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Grb_Version
//  
//  Description:
//      Returns the grabber version information.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         Client_ES:Client_EDI
//            ptr to string buffer for grabber name
//  
//      DWORD hVM
//         not used
//  
//      PVDDCB pVMCB
//         not used
//  
//      BOOL fVMValid
//         not used
//  
//  Return (VOID):
//      pCRS -> Client_EAX = version
//      pCRS -> Client_ES:Client_EDI = string name
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Grb_Version
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   REGS   regs ;
   PDWORD pDst, pSrc ;

   UNUSED_PARAM( hVM ) ;
   UNUSED_PARAM( pVMCB ) ;
   UNUSED_PARAM( fVMValid ) ;

   VDD_Get_GrbVersion( &regs ) ;
   pCRS -> Client_EAX = regs.eax ;

   if (-1 !=
         (pDst = (PDWORD) vmmSelectorMapFlat( hSysVM, pCRS -> Client_ES, 0 )))
   {
      pDst = (PDWORD) ((DWORD) pDst + pCRS -> Client_EDI) ;
      pSrc = (PDWORD) regs.esi ;
      *pDst++ = *pSrc++ ;
      *pDst++ = *pSrc++ ;
   }

} // VDD_Grb_Version()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Get_State, VDD_Get_GrbState
//  
//  Description:
//      These routines are called by the grabber. Since the grabber is a
//      virtual machine piece of code, all parameters exist on the stack
//      in the Client registers. These routines return a video state
//      structure (see VDD.INC). The Get_GrbState version of the
//      routine returns the state of the video for the VM specified at the
//      point when the GrabRtn was called. The Get_State version of the
//      routine returns the current state of the VM. The Get_State routine
//      should be called after the Get_Mem call, before the Free_Mem call.
//
//      Note that a state structure size of zero, returned in CX,
//      indicates an error.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         Client_ES:Client_EDI
//            SEL:OFF of buffer to store state
//
//         Client_CX
//            size of buffer to hold state structure (DEBUG ONLY)
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//      pCRS -> Client_SI = size of state structure returned (DEBUG ONLY!)
//
//      State structure saved in buffer indicated
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Get_State
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   PBYTE  pDst ;

   if (!fVMValid)
   {
      // error

      pCRS -> Client_ECX = 0 ;
      return ;
   }

#ifdef DEBUG
   if (((PCRS_16) pCRS) -> Client_CX < sizeof( VDAEGASTATE ))
   {
      DEBUGMSG( "VDD: Get_State buffer too small.\r\n" ) ;

      // error 
      pCRS -> Client_ECX = 0 ;
      return ;
   }
   else
      ((PCRS_16) pCRS) -> Client_CX = sizeof( VDAEGASTATE ) ;
#endif

   if (-1 ==
         (pDst = (PSTR) vmmSelectorMapFlat( hSysVM, pCRS -> Client_ES, 0 )))
   {
      DEBUGMSG( "VDD: SelectorMapFlat failed for Get_State.\r\n" ) ;

      // error
      pCRS -> Client_ECX = 0 ;
      return ;
   }

   pDst += pCRS -> Client_EDI ;

   VDD_State_GetState( hVM, pVMCB, pDst, GETSTATEF_MAINSTATE ) ;

} // VDD_Get_State()

VOID VDD_Get_GrbState
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   PBYTE  pDst ;

   if (!fVMValid)
   {
      // error

      pCRS -> Client_ECX = 0 ;
      return ;
   }

#ifdef DEBUG
   if (((PCRS_16) pCRS) -> Client_CX < sizeof( VDAEGASTATE ))
   {
      DEBUGMSG( "VDD: Get_State buffer too small.\r\n" ) ;

      // error 
      pCRS -> Client_ECX = 0 ;
      return ;
   }
   else
      ((PCRS_16) pCRS) -> Client_CX = sizeof( VDAEGASTATE ) ;
#endif

   if (-1 ==
         (pDst = (PSTR) vmmSelectorMapFlat( hSysVM, pCRS -> Client_ES, 0 )))
   {
      DEBUGMSG( "VDD: SelectorMapFlat failed for Get_State.\r\n" ) ;

      // error
      pCRS -> Client_ECX = 0 ;
      return ;
   }

   pDst += pCRS -> Client_EDI ;

   VDD_State_GetState( hVM, pVMCB, pDst, GETSTATEF_VMCOPYSTATE ) ;

} // VDD_Get_GrbState()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Get_Mod
//  
//  Description:
//      This routine is called to get changes in a VM's video state.
//      The changes are passed to the grabber in a buffer that includes
//      a flag indicating what kind of changes occured and what type of
//      a memory change list follows. The flag is followed by a count of
//      memory change list entries and the change list itself. Note that
//      a change structure size of zero, returned in CX, indicates an
//      error.
//
//      This routine will return cumulative changes until Clear_Mod is
//      called.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         Client_ES:Client_EDI
//            SEL:OFF of buffer to store chgs structure
//
//         Client_CX = size of buffer to hold changes structure (DEBUG ONLY!)
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//      Client_CX
//         size of change structure returned (DEBUG ONLY!)
//
//      Client_AX
//         location (bottom,right)
//     
//      Client_BX 
//         location (top,left)
//
//      Client_DX
//         A_Mode
//
//      Changes structure saved in buffer indicated
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Get_Mod
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   PBYTE  pDst ;

   if (!fVMValid)
   {
      // error

      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      goto VGM_Exit_NoMod ;
   }

#ifdef DEBUG
   if (((PCRS_16) pCRS) -> Client_CX < sizeof( VDDMODSTATE ))
   {
      DEBUGMSG( "VDD: Get_Mod buffer too small.\r\n" ) ;

      // error 
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      goto VGM_Exit_NoMod ;
   }
#endif

   if (-1 ==
         (pDst = (PSTR) vmmSelectorMapFlat( hSysVM, pCRS -> Client_ES, 0 )))
   {
      DEBUGMSG( "VDD: SelectorMapFlat failed for Get_Mod.\r\n" ) ;

      // error
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      goto VGM_Exit_NoMod ;
   }

   pDst += pCRS -> Client_EDI ;

   if (!VDD_VMMem_ChkSaveMem( hVM, pVMCB ))
   {
      VDD_Error_NoMainMem() ;
      ((PVDDMODSTATE) pDst) -> VDD_Mod_Flag = fVDD_M_Err ;
      ((PVDDMODSTATE) pDst) -> VDD_Mod_Count = 0 ;

      goto VGM_Exit ;
   }

   // Report no changes if not initialized or during a mode change

   if ((0 == (pVMCB -> VDD_Flags & fVDD_MInit)) ||
       (pVMCB -> VDD_Flags & fVDD_ModeChange))
   {
      ((PVDDMODSTATE) pDst) -> VDD_Mod_Flag = 0 ;
      ((PVDDMODSTATE) pDst) -> VDD_Mod_Count = 0 ;

      goto VGM_Exit ;
   }

   VDD_State_GetChgs( hVM, pVMCB, pDst ) ;
   ((PCRS_16) pCRS) -> Client_CX = VDD_Grab_GetChgs( hVM, pVMCB, pDst ) ;

VGM_Exit: 
   ((PCRS_16) pCRS) -> Client_CX +=
      FIELDOFFSET( VDDMODSTATE, VDD_Mod_List ) ;

VGM_Exit_NoMod:

   // return (a, b)
   ((PCRS_16) pCRS) -> Client_BX = (WORD) pVMCB -> VDD_Mod_Rect ;

   // return (c, d)
   ((PCRS_16) pCRS) -> Client_AX = (WORD) (pVMCB -> VDD_Mod_Rect >> 16) ;

   // zero out rect
   pVMCB -> VDD_Mod_Rect = 0xFFFF0000 ;

   ((PCRS_16) pCRS) -> Client_DX = (WORD) pVMCB -> VDD_Stt.A_Mode ;

} // VDD_Get_Mod()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Clear_Mod
//  
//  Description:
//      Called to clear a change state of a VM.
//  
//  Parameters:
//      DWORD hSysVM
//         not used
//  
//      PCRS_32 pCRS
//         not used
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Clear_Mod
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   UNUSED_PARAM( hSysVM ) ;
   UNUSED_PARAM( pCRS ) ;

   if (fVMValid)
   {
      // clear ctlr changes

      VDD_State_ClearModState( hVM, pVMCB ) ;

      // clear memory changes

      VDD_Grab_ClearModState( hVM, pVMCB ) ;
   }

} // VDD_Clear_Mod()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Get_Mem
//  
//  Description:
//      Return flat address and allocation bit map for VM's
//      video memory.
//      
//      The main difference between these two routines is that
//      Get_Mem returns the main video save memory and Get_GrbMem
//      a copy of the memory that was made by VDD_Grab.
//      An additional important difference is that Get_Mem boosts
//      the scheduling priority of the system VM, as mentioned below.
//      
//      In order to assure that the memory does not change from the
//      point that Get_Mem is called until Free_Mem is called, we
//      will boost the system VM's priority by a Low_Priority_Device_Boost.
//      This will inhibit normal scheduling but will not hamper
//      scheduling of VM's to handle events such as interrupts.
//      
//      THE ALLOCATION OF MEMORY TO THE BIT MAP IS DYNAMIC. The
//      grabber must make this call each time that it wants to access
//      the memory.
//      
//      Get_Mem call MUST BE FOLLOWED by a Free_Mem call AS SOON AS
//      POSSIBLE. No other VM's will be scheduled until the Free_Mem
//      call is made.
//      
//      If the client buffer is not big enough or the selector is bad
//      or if GrbMem is called when no grab is active, an error return
//      will be done, that is, Client SI = 0.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         Client_ES:Client_EDI
//            SEL:OFF of buffer to store VDA_Mem_State
//
//         Client_CX
//            size of buffer to hold VDA_Mem_State (DEBUG ONLY!)
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//     Mem structure modified for current state of memory
//
//     Client_CX
//        size of structure (=0 if error)
//
//     Client_DX
//        Attribute mode controller register value
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Get_Mem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   if (!fVMValid)
   {
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      return ;
   }

   pVMCB -> VDD_Get_Mem_Count++ ;

   if (pVMCB -> VDD_Stt.V_BIOSMode >= 0x0D)
      pVMCB -> VDD_Upd_Start = vmmGetVMExecTime( hSysVM ) ;

   if (VGMm_SetStruc( hSysVM, pCRS, hVM, pVMCB, &pVMCB -> VDD_MStt ))
      VDD_Suspend_NormSch( hSysVM ) ;

} // VDD_Get_Mem()

VOID VDD_Get_GrbMem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   PMEMSTATE pMS ;

   if (!fVMValid)
   {
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      return ;
   }

   pMS = &pVMCB -> VDD_MSttCopy ;
   if (0 == pMS -> MSS_planes[ 0 ].MSS_plane_addr)
      pMS = &pVMCB -> VDD_MStt ;

   VGMm_SetStruc( hSysVM, pCRS, hVM, pVMCB, pMS ) ;
      
} // VDD_Get_GrbMem()

//--------------------------------------------------------------------------
//  
//  BOOL VGMm_Set_Struc
//  
//  Description:
//      Copy memory state structure to client buffer, including
//      addresses for each of the planes and size of each of the
//      four planes (zero size for planes not allocated).
//  
//  Parameters:
//      DWORD hSysVM
//         handle to System VM
//  
//      PCRS_32 pCRS
//         Client_CX
//            size of buffer to hold mem structure
//
//         Client_ES:Client_EDI
//            SEL:OFF of buffer to store mem structure
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      PMEMSTATE pMS
//         pointer to MEMSTATE structure
//  
//  Return (BOOL):
//     FALSE if error, otherwise TRUE and Mem structure copied to
//     client buffer
//
//     Client_ECX
//        size of structure copied (=0 if error)
//
//     Client_DX
//        Attribute controller register 10h
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

BOOL VGMm_Set_Struc
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS
)
{
   PVDAMEMSTATE  pDst ;

   ((PCRS_16) pCRS) -> Client_DX = pVMCB -> VDD_Stt.A_Mode ;

#ifdef DEBUG
   if (((PCRS_16) pCRS) -> Client_CX < sizeof( VDAMEMSTATE ))
   {
      DEBUGMSG( "VDD: Set_Struc buffer too small.\r\n" ) ;

      // error
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      return FALSE ;

   }
#endif

   ((PCRS_16) pCRS) -> Client_CX = sizeof( VDAMEMSTATE ) ;
   if (-1 ==
         (pDst = (PVDAMEMSTATE) vmmSelectorMapFlat( hSysVM,
                                                    pCRS -> Client_ES, 0 )))
   {
      DEBUGMSG( "VDD: SelectorMapFlat failed for Set_Struc.\r\n" ) ;

      // error
      ((PCRS_16) pCRS) -> Client_CX = 0 ;
      return FALSE ;
   }
   pDst = (PVDAMEMSTATE) ((DWORD) pDst+ pCRS -> Client_EDI) ;

   pDst -> VDA_Mem_DACOff = (DWORD) &pVMCB -> VDD_DAC ;
   pDst -> VDA_Mem_Addr_P0 =
      pMS -> MSS_planes[ 0 ].MSS_plane_addr ;
   pDst -> VDA_Mem_Size_P0 =
      pMS -> MSS_planes[ 0 ].MSS_plane_size << 12 ;

   if (pVMCB -> VDD_Stt.G_Misc & fGrp6Char)
   {
      pDst -> VDA_Mem_Addr_P1 =
         pMS -> MSS_planes[ 1 ].MSS_plane_addr ;
      pDst -> VDA_Mem_Size_P1 =
         pMS -> MSS_planes[ 1 ].MSS_plane_size << 12 ;

      pDst -> VDA_Mem_Addr_P2 =
         pMS -> MSS_planes[ 2 ].MSS_plane_addr ;
      pDst -> VDA_Mem_Size_P2 =
         pMS -> MSS_planes[ 2 ].MSS_plane_size << 12 ;

      pDst -> VDA_Mem_Addr_P3 =
         pMS -> MSS_planes[ 3 ].MSS_plane_addr ;
      pDst -> VDA_Mem_Size_P3 =
         pMS -> MSS_planes[ 3 ].MSS_plane_size << 12 ;
   }
   else
   {
      VDD_Grab_GetTextCopyBuf( hVM, pVMCB, &pDst -> VDA_Mem_Addr_P1,
                               &pDst -> VDA_Mem_Size_P1 ) ;
      pDst -> VDA_Mem_Size_P1 <<= 12 ;

      // pass font to grabber

      pDst -> VDA_Mem_Addr_P2 =
         pMS -> MSS_planes[ 2 ].MSS_plane_addr ;
      pDst -> VDA_Mem_Size_P2 =
         pMS -> MSS_planes[ 2 ].MSS_plane_size << 12 ;

      pDst -> VDA_Mem_Size_P3 = 0 ;
   }

   pDst -> VDA_Mem_DPagOff = VDD_State_GetScreenStart( hVM, pVMCB ) ;

} // VGMm_Set_Struc()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Suspend_NormSch
//  
//  Description:
//      Boosts the priority of the Sys VM.
//  
//  Parameters:
//      DWORD hSysVM
//         System VM handle
//  
//  Return (VOID):
//      Priority is boosted.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Suspend_NormSch
(
    DWORD           hSysVM
)
{
   PVDDCB pSysVMCB ;

   pSysVMCB = (PVDDCB) (hSysVM + dwVidCBOff) ;
   if (0 == (pSysVMCB -> VDD_Flags & fVDD_Boost))
   {
      vmmAdjustExecPriority( hSysVM, VDD_Pri_Device_Boost ) ;
      pSysVMCB -> VDD_Flags |= fVDD_Boost ;
   }

} // VDD_Suspend_NormSch()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Free_Grab
//  
//  Description:
//      This routine is used to release the copy of the video memory
//      that was allocated when a screen grab was done.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM (not used)
//  
//      PCRS_32 pCRS
//         pointer to client regs (not used)
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Free_Grab
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   UNUSED_PARAM( hSysVM ) ;
   UNUSED_PARAM( pCRS ) ;

   if (fVMValid)
   {
      if (0 == (pVMCB -> VDD_Flags & fVDD_Win))
         VDD_Grab_DCopy( hVM, pVMCB ) ;
   }

} // VDD_Free_Grab()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Free_Mem
//  
//  Description:
//      This routine is used to release the scheduling freeze that
//      occurred when the Get_Mem call was made. See Get_Mem
//      documentation for details.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Free_Mem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   if (fVMValid)
   {
      pVMCB -> VDD_Get_Mem_Count-- ;

      VDD_Unlock_App( hSysVM, pCRS, hVM, pVMCB, fVMValid ) ;
   }

} // VDD_Free_Mem()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Unlock_App
//  
//  Description:
//      This routine is used to Unlock the Windowed OLDAPP from the
//      grabber.  Called from VDD_Free_Mem also.
//
//      This does not take into account updating TWO VMs at once or
//      that other windows programs are running, but is the best we
//      can do.
//  
//  Parameters:
//      DWORD hSysVM
//         handle to Sys VM
//  
//      PCRS_32 pCRS
//         pointer to client regs
//  
//      DWORD hVM
//         VM grabber is calling on behalf
//  
//      PVDDCB pVMCB
//         hVM's control block
//  
//      BOOL fVMValid
//         VM handle is valid?
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Unlock_App
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
)
{
   PVDDCB  pSysVMCB ;
   DWORD   dwCurTime ;

   UNUSED_PARAM( pCRS ) ;

   if (fVMValid && (0 == pVMCB -> VDD_Get_Mem_Count))
   {
      pSysVMCB = (PVDDCB) (hSysVM + dwVidCBOff) ;
      if (pSysVMCB -> VDD_Flags & fVDD_Boost)
      {
         pSysVMCB -> VDD_Flags &= ~fVDD_Boost ;

         // unboost Sys VM

         vmmAdjustExecPriority( hSysVM, -VDD_Pri_Device_Boost ) ;
      }

      dwCurTime = vmmGetVMExecTime( hSysVM ) ;
      dwCurTime -= pVMCB -> VDD_Upd_Start ;
      vmmAdjustExecutionTime( hSysVM, dwCurTime ) ;
      vmmAdjustExecutionTime( hVM, -dwCurTime ) ;
   }

} // VDD_Unlock_App()

//--------------------------------------------------------------------------
//  
//  VOID VDD_SVC_DspVersion
//  
//  Description:
//      Service for the display driver to determine level of
//      support provided by VDD.
//  
//  Parameters:
//      PCRS_32 pCRS
//         pointer to client regs
//  
//  Return (VOID):
//      pCRS -> Client_AX
//         version
//
//      pCRS -> Client_DX
//         flags, currently 0
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_SVC_DspVersion
(
    PCRS_32         pCRS
)
{
   ((PCRS_16) pCRS) -> Client_AX = VDD_VerNum ;
   ((PCRS_16) pCRS) -> Client_DX = 0 ;

} // VDD_SVC_DspVersion()

//--------------------------------------------------------------------------
//  
//  VOID VDD_SVC_SetAddresses
//  
//  Description:
//      This service is called by the display driver BEFORE it
//      does the INT 2Fh to indicate that it knows how to restore
//      its screen.  This service is used to tell us where we
//      can interface with the display driver.
//      
//      One of the addresses passed is the location of a byte of
//      video memory that we can safely use to save/restore latches.
//      
//      The 2nd address passed is the location of the flag byte
//      used by the display driver to determine the availability
//      and validity of the "save screen bits" area.  The display
//      driver copies portions of the visible screen to non-visible
//      memory to save original contents when dialog boxes or menus
//      area displayed.  Since this is non-visible memory, it is
//      subject to demand paging and be stolen for use by a
//      different VM.  If a page is stolen from the sys VM, then
//      any data that was in the page is lost, because we don't
//      maintain a copy, so we need to indicate to the display
//      driver that the "save screen bits" area is now invalid.
//  
//  Parameters:
//      DWORD hVM
//         handle to VM
//  
//      PCRS_32 pCRS
//         Client_AX
//            function #
//
//         Client_BX
//            offset of address in display segment
//
//         Client_DX
//            is reserved and must be 0
//
//         ClientDS:Client_SI
//            pointer to shadow_mem_status
//  
//  Return (VOID):
//     Client_AX
//        is returned with a copy of Client_BX to indicate that
//        the service is implemented
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_SVC_SetAddresses
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   PVDDCB  pVMCB ;
   DWORD   dwVisPages ;

   pVMCB = (PVDDCB)(hVM + dwVidCBOff) ;

   ((PCRS_16) pCRS) -> Client_AX = ((PCRS_16) pCRS) -> Client_BX ;

   // get number of visible pages - this assumes
   // latch page is the last visible page
                  
   dwVisPages = ((PCRS_16) pCRS) -> Client_BX ;
   dwVisPages >>= 12 ;
   dwVisPages++ ;

   pVMCB -> VDD_Flags |= fVDD_DspDrvrAware ;

   VDD_PHMem_SetSysLatchAddr( ((PCRS_16) pCRS) -> Client_BX, 0, 0,
                              dwVisPages ) ;

   dwVidShadowMemStatusPtr =
      (DWORD) (((PCRS_16) pCRS) -> Client_DS) << 16 |
         ((PCRS_16) pCRS) -> Client_SI ;

} // VDD_SVC_SetAddresses()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Init_DspDrvRing0
//  
//  Description:
//      Initializes display driver ring 0 callback hack.
//  
//  Parameters:
//      DWORD hVM
//         VM handle
//  
//      PCRS_32 pCRS
//         Client_CX
//            ring 3 CS
//
//  Return (VOID):
//      Client_DX:AX
//         call address to switch to ring 0
//
//      Client_CX:BX
//         call address to switch back to ring 3
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Init_DspDrvRing0
(
    DWORD           hVM,
    PCRS_32         pCRS
)
{
   WORD          wRing3CS ;
   DWORD         dw1, dw2, dwCallBack, dwRing0CS ;
   static DWORD  VDD_Return_To_Ring3_CS = 0 ;

   if (!VDD_Return_To_Ring3_CS)
   {
      vmmBuildDescriptorDWORDs( &dw1, &dw2, (PSTR) VDD_Jump_To_Ring3,
                                0x1000, Code_Type, D_DEF32, BDDExplicitDPL ) ;
      if (!VDD_Return_To_Ring3_CS = vmmAllocateGDTSelector( dw1, dw2, 0 ))
      {
         DEBUGMSG( "VDD: failed to allocated GDT selector for return_to_ring3\r\n" ) ;

         vmmFatalError( szDspDrvInitFailed ) ;
      }
   }

   ((PCRS_16) pCRS) -> Client_BX = 0 ;
   wRing3CS = ((PCRS_16) pCRS) -> Client_CX ;
   ((PCRS_16) pCRS) -> Client_CX = (WORD) VDD_Return_To_Ring3_CS ;

   vmmGetDescriptor( &dw1, &dw2, wRing3Cs, hVM, 0 ) ;

   // make DPL=0

   dw1 &= ~D_DPL3 ;
   if (!dwRing0CS = vmmAllocateGDTSelector( dw1, dw2, 0 ))
      vmmFatalError( szDspDrvInitFailed ) ;

   if (!dwCallBack = vmmAllocatePMCallBack( VDD_Jump_To_Ring0, dwRing0CS ))
      vmmFatalError( szDspDrvInitFailed ) ;

   ((PCRS_16) pCRS) -> Client_AX = (WORD) dwCallBack ;
   ((PCRS_16) pCRS) -> Client_DX = (WORD) (dwCallBack >> 16) ;

} // VDD_Init_DspDrvRing0()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Error_CvtFullScrn
//  
//  Description:
//      Force a full screen message to inform the user that their
//      application is being forced to full screen.
//  
//  Parameters:
//      DWORD hVM
//         VM handle
//  
//      DWORD dwRefData
//         not used
//  
//      PCRS_32 pCRS
//         not used
//  
//  Return (VOID):
//      Nothing.
//
//  Comments:
//      This routine may be called by a scheduled event from
//      VDD_PFault, the extra parameters are to maintain compatibility
//      as an event handler.  This routine is thunked for this case.
//  
//  History:   Date       Author      Comment
//             10/23/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID CDECL VDD_Error_CvtFullScrn
(
    DWORD           hVM,
    DWORD           dwRefData,
    PCRS_32         pCRS
)
{
   UNUSED_PARAM( dwRefData ) ;
   UNUSED_PARAM( pCRS ) ;

   shellSYSMODALMessage( hVM, szVMB_Str_CvtFullScrn, NULL,
                         MB_OK + MB_ICONEXCLAMATION + MB_ASAP +
                         MB_SYSTEMMODAL + MB_NOWINDOW ) ;

} // VDD_Error_CvtFullScrn()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Error_NoPagesAvail
//  
//  Parameters:
//      DWORD hVM
//         VM for whom message is generated
//  
//      PVDDCB pVMCB
//         VN's VDD CB ptr
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Error_NoPagesAvail
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   VDD_ErrorMessage( hVM, pVMCB, szVMB_Str_NoPagesAvail, bVDE_NoPAMsg ) ;

} // VDD_Error_NoPagesAvail()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Error_NoFGrnd
//  
//  Parameters:
//      DWORD hVM
//         VM for whom message is generated
//  
//      PVDDCB pVMCB
//         VN's VDD CB ptr
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Error_NoFGrnd
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   VDD_ErrorMessage( hVM, pVMCB, szVMB_Str_NoFGrnd, bVDE_NoFGMsg ) ;

} // VDD_Error_NoFGrnd()

//--------------------------------------------------------------------------
//  
//  VOID VDD_Error_NoMainMem
//  
//  Parameters:
//      DWORD hVM
//         VM for whom message is generated
//  
//      PVDDCB pVMCB
//         VN's VDD CB ptr
//  
//  Return (VOID):
//      Nothing.
//  
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//  
//--------------------------------------------------------------------------

VOID VDD_Error_NoMainMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
)
{
   VDD_ErrorMessage( hVM, pVMCB, szVMB_Str_NoMainMem, bVDE_NoMMsg ) ;

} // VDD_Error_NoMainMem()

//------------------------------------------------------------------------
//  VOID VDD_ErrorMessage
//
//  Description:
//     VDD_Error_Message is the same as VDD_Message except that it
//     forces EAX to MB_OK+MB_ICONEXCLAMATION+MB_ASAP (the standard
//     flags for error dialogs)
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     PSTR pMsg
//        pointer to message string
//
//     UINT uEFlagBit
//        VDD_EFlags bit to set and check to avoid duplication of
//        error messages
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_ErrorMessage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PSTR            pMsg,
    UINT            uEFlagBit
)
{
   VDD_Message( hVM, pVMCB, MB_OK | MB_ICONEXCLAMATION | MB_ASAP,
                pMsg, uEFlagBit ) ;

} // end of VDD_ErrorMessage()

//------------------------------------------------------------------------
//  VOID VDD_Message( DWORD hVM, PVDDCB pVMCB, UINT uFlags,
//                    PSTR pMsg, UINT uEFlagBit )
//
//  Description:
//     Call SHELL_Message to display a dialog box message, and if
//     SHELL fails the call, then call SHELL_SYSMODAL_Message to
//     force a fullscreen error dialog.
//
//  Parameters:
//     DWORD hVM
//        handle to VM
//
//     PVDDCB pVMCB
//        pointer to VM control block
//
//     UINT uFlags
//        message box flags
//
//     PSTR pMsg
//        pointer to message string
//
//     UINT uEFlagBit
//        VDD_EFlags bit to set and check to avoid duplication of
//        error messages
//
//  History:   Date       Author      Comment
//              4/25/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

VOID VDD_Message
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uFlags,
    PSTR            pMsg,
    UINT            uEFlagBit
)
{
   if (0 == pVMCB -> VDD_EFlags & (1 << uEFlagBit))
   {
      pVMCB -> VDD_EFlags |= (1 << uEFlagBit) ;

      if (!shellMessage( hVM, pMsg, NULL, uFlags, NULL, NULL ))
         shellSYSMODALMessage( hVM, pMsg, NULL, uFlags + MB_SYSTEMMODAL ) ;
   }

} // end of VDD_Message()

//---------------------------------------------------------------------------
//  End of File: vddsvc.c
//---------------------------------------------------------------------------

