//---------------------------------------------------------------------------
//
//  Module: vddvga.h
//
//  Purpose:
//     Header file for vddvga.c
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              3/ 9/93   BryanW      Wrote it.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include "vmdavga.h"

//----------------------------------------------------------------------------
//
//                          Version equates
//
//----------------------------------------------------------------------------

#define VDD_Major_Ver        0x02
#define VDD_Minor_Ver        0x01

#define VDD_VerNum           0x30A

//----------------------------------------------------------------------------
//
//                              Macros
//
//----------------------------------------------------------------------------

#define UNUSED_PARAM( x ) x = x ;

#ifdef DEBUG
   #define TRACEMSG( sz )       (vmmTraceOut( sz ))
   #define TRACEPMSG( sz, x )   (vmmTraceOutParam( sz, x ))
   #define DEBUGMSG( sz )       (vmmDebugOut( sz ))
#else
   #define TRACEMSG( sz )       
   #define TRACEPMSG( sz, x )   
   #define DEBUGMSG( sz )       
#endif

#define FIELDOFFSET(type, field)    ((int)(&((type *)1)->field)-1)

//----------------------------------------------------------------------------
//
//               Specific video adapter conditionals
//
//----------------------------------------------------------------------------

// Default VGA

#define RESERVEROMC6C7  -1      // Turn on detection and reserve of ROM

//----------------------------------------------------------------------------
//
//                    PIF_State service definitions
//
//----------------------------------------------------------------------------

// *** NOTE: These must be updated if VDD.INC's definitions change!!!! ***

#define fVidTxtEmulate   0x0001      // Do INT 10h TTY and cursor emulation
#define fVidNoTrpTxt     0x0002      // Do not trap text mode apps
#define fVidNoTrpLRGrfx  0x0004      // Do not trap lo res graphics mode apps
#define fVidNoTrpHRGrfx  0x0008      // Do not trap hi res graphics mode apps
#define fVidTextMd       0x0010      // Allocate text mode mem
#define fVidLowRsGrfxMd  0x0020      // Allocate lo res graphics mode mem
#define fVidHghRsGrfxMd  0x0040      // Allocate hi res graphics mode mem
#define fVidRetainAllo   0x0080      // Never deallocate once allocated

//----------------------------------------------------------------------------
//
//                       Structure definitions
//
//----------------------------------------------------------------------------

typedef struct tagMEMCMOD
{
   WORD  MMS_data_port ;        // port value
   BYTE  MMS_index ;            // index, if necessary (graphics controller,
                                //                      sequencer)
   BYTE  MMS_ANDmask ;          // mask ANDed with new value
   BYTE  MMS_ORmask ;           // mask ORed with AND result
   BYTE  MMS_flags ;
   BYTE  MMS_resvd[ 2 ] ;       // 8 byte aligned

} MEMCMOD ;

#define MaxMemCMods  16

#define fMMS_OEM_extension     0x1
#define bMMS_OEM_extension     0
#define fMMS_no_index          0x2
#define bMMS_no_index          1

typedef struct tagMSSPLANEINFO
{
   DWORD  MSS_plane_handle ;     // handle of memory alloc'ed or 0
   DWORD  MSS_plane_addr ;       // linear address of memory alloc'ed
   WORD   MSS_plane_size ;       // size needed for plane
   WORD   MSS_plane_alloced ;    // size alloc'ed for plane
   
} MSSPLANEINFO, *PMSSPLANEINFO ;

typedef struct tagMEMSTATE
{
   MSSPLANEINFO   MSS_planes[ 4 ] ;
   BYTE           MSS_fontmap[ 8 ] ;

} MEMSTATE, *PMEMSTATE ;

#define MSS_plane0  MSS_planes[ 0 ] ;
#define MSS_plane1  MSS_planes[ 1 ] ;
#define MSS_plane2  MSS_planes[ 2 ] ;
#define MSS_plane3  MSS_planes[ 3 ] ;

//----------------------------------------------------------------------------
//                CRTC registers structure definition
//----------------------------------------------------------------------------

#define CL_Resvd  (7+32)         // Fill out to 64 bytes (power of 2)

typedef struct tagCRTC
{
   BYTE  C_HTotal ;              // Horizontal total
   BYTE  C_HDisp ;               // Horizontal displacement
   BYTE  C_HBlnkStrt ;           // Horizontal blank start
   BYTE  C_HBlnkEnd ;            // Horizontal blank end
   BYTE  C_HRTCStrt ;            // Horizontal retrace start
   BYTE  C_HRTCEnd ;             // Horizontal retrace end
   BYTE  C_VTotal ;              // Vertical total
   BYTE  C_Ovflw ;               // 8th bit for following
   BYTE  C_RowScn ;              // First row after VRTC
   BYTE  C_CharHgt ;             // Character Cell Height
   BYTE  C_CStart ;              // Cursor start scan line
   BYTE  C_CEnd ;                // Cursor end scan line
   BYTE  C_AddrH ;               // VRAM address high byte
   BYTE  C_AddrL ;               // VRAM address low byte
   BYTE  C_CAddrH ;              // Cursor address high byte
   BYTE  C_CAddrL ;              // Cursor address low byte

   // Note... the following two register have
   // read values as noted

   BYTE  C_VRTCStrt ;            // Vertical retrace start (w/o)
                                 //    or light pen address high byte (r/o)
   BYTE  C_VRTCEnd ;             // Vertical retrace end (w/o)
                                 //    or light pen address low byte (r/o)

   BYTE  C_VDspEnd ;             // Vertical display end line
   BYTE  C_LnOff ;               // Offset to next line
   BYTE  C_UndrLn ;              // Underline
   BYTE  C_VBlnkStrt ;           // Vertical blanking start
   BYTE  C_VBlnkEnd ;            // Vertical blanking end
   BYTE  C_Mode ;                // Mode
   BYTE  C_LnCmp ;               // Line compare for MA clearing

   // Fill out to power of 2

   BYTE  C_Extension[ CL_Resvd - 1 ] ;
   BYTE  C_Last ;

} CRTC, *PCRTC ;

#define C_MaxScan    (CRTC.C_CharHgt)

//
// Extension to CRTC state for various versions defined as offsets
// from the last entry.  Set C_Last to the offset of the last
// defined entry.
//

#define SL_Resvd           11           // OEM extensions to Sequencer
#define GL_Resvd           23           // OEM extensions to Graphics cntrllr
#define Extended_Reg_Size  0x80         // OEM other extended regs

typedef struct tagREGSTATE
{
   // CRT controller state must be first so we can access
   // using CRTC structure definitions...

   CRTC    crtc ;

   // Index registers

   BYTE    A_Indx ;                        // Attribute controller index reg
   BYTE    S_Indx ;                        // Sequencer index reg
   BYTE    C_Indx ;                        // CRT controller index reg
   BYTE    G_Indx ;                        // Graphics controller index reg

   // Attribute controller

   BYTE    A_Pal[ 16 ] ;                   // Color palette
   BYTE    A_Mode ;                        // Mode
   BYTE    A_OvScn ;                       // Over scan color
   BYTE    A_PlEna ;                       // Plane enable to attr reg
   BYTE    A_HPan ;                        // Horizontal panning
   BYTE    A_Color ;                       // VGA color register
   BYTE    A_Extension[ 10 ] ;             // up to 32 (power of 2)
   BYTE    A_Last ;        

   // Sequencer

   BYTE    S_Rst ;                         // Reset
   BYTE    S_ClMode ;                      // Clocking mode
   BYTE    S_Mask ;                        // Map mask
   BYTE    S_ChMap ;                       // Char map select
   BYTE    S_MMode ;                       // Memory mode
   BYTE    S_Extension[ SL_Resvd - 1 ] ;   // Up to 16 (power of 2)
   BYTE    S_Last ;

   // Graphics controller

   BYTE    G_SetRst ;                      // Set/Reset on write
   BYTE    G_SREna ;                       // Set/Reset enable
   BYTE    G_CComp ;                       // Color compare
   BYTE    G_Func ;                        // Data rotate and function
   BYTE    G_MSlct ;                       // Read Map select
   BYTE    G_Mode ;                        // Mode
   BYTE    G_Misc ;                        // Miscellaneous
   BYTE    G_DCare ;                       // Color don't care on compare reads
   BYTE    G_Mask ;                        // Mask for data/latches on write
   BYTE    G_Extension[ GL_Resvd - 1 ] ;   // up to 32 (power of 2)
   BYTE    G_Last ;

   // Miscellaneous

   BYTE    V_Misc ;                        // Miscellaneous output

   // Miscellaneous registers

   BYTE    G_Pos1 ;                        // Graphics 1 position (must be 0)
   BYTE    G_Pos2 ;                        // Graphics 1 position (must be 1)
   BYTE    V_Feat ;                        // Feature connector output
   BYTE    V_VGAEna ;                      // VGA - subsystem enable

   // Status (input) registers

   BYTE    V_Stat0 ;                       // Status 0
   BYTE    V_Stat1 ;                       // Status 1

   BYTE    V_BIOSMode ;                    // Video BIOS mode
   BYTE    V_Extend[ Extended_Reg_Size ] ; // Save area for extended regs.

} REGSTATE, *PREGSTATE ;

//----------------------------------------------------------------------------
//              Control Block (CB) structure definition
//----------------------------------------------------------------------------

#define DAC_State_Size        (256*4)
#define DAC_Msg_State_Size    (16*4)

#define ModListMax            64

#define VRAM_Mod_List_Size    (32/8)      // 32 pages/8 bits per byte

#if (!(VRAM_Mod_List_Size <= ModListMax))
#error assertion failed: VRAM_Mod_List_Size <= ModListMax
#endif

#define DAC_State_Size        (256*4)
#define DAC_Msg_State_Size    (16*4)

typedef struct tagVDDCB
{
   // Flags, state of VM's controller, state of VM's video memory

   DWORD      VDD_Flags ;             // VDD flags
   DWORD      VDD_TFlags ;            // VDD type flags
   DWORD      VDD_EFlags ;            // VDD event flags

   REGSTATE   VDD_Stt ;               // Controller state structure

   // DAC save state... this must follow VDD_Stt!!!

   BYTE       VDD_DAC[ DAC_State_Size ] ;

   DWORD      VDD_PEL_Addr ;          // Address withing dsVDD_DAC array
   BYTE       VDD_PEL_Idx ;           // Byte index within PEL DWORD entry
   BYTE       VDD_PEL_Mode ;          // 0 - read mode, 3 - write mode

   BYTE       VDD_fontpages ;         // a set bit indicates the loaded font
                                      // in this order 73625140

   BYTE       VDD_MemC_ModCnt ;
   MEMCMOD    VDD_MemC_Modsp[ MaxMemCMods ] ;

   DWORD      VDD_TrapEnable ;         // one bit for each port trapped

   BYTE       VDD_Latches[ 4 ] ;       // video memory latches

   REGSTATE   VDD_SttCopy ;            // controller state copy
   MEMSTATE   VDD_MStt ;               // memory state structure
   MEMSTATE   VDD_MSttCopy ;           // memory state copy

   DWORD      VDD_VideoPages ;         // # entries in following tables
   PDWORD     VDD_PageMap ;            // pointer to page map array
   PBYTE      VDD_PageFlags ;          // pointer to page flags array

   DWORD      VDD_VRAM_Mods[ VRAM_Mod_List_Size / 4 ] ;
   DWORD      VDD_Grab_Mods[ VRAM_Mod_List_Size / 4 ] ;

   DWORD      VDD_Mod_Rect ;           // rectangle on change <(a,b),<c,d)>

   WORD       VDD_Recent_User_Action ; // recently event occured
   WORD       VDD_Check_Upd_Soon ;     // check update
   WORD       VDD_Scroll_Cntr ;        // generate 1 update for so many scrolls
   WORD       VDD_Last_Int10 ;         // what was the last int 10?

   DWORD      VDD_NotifyEvent ;        // event handle for VDD_Int_FGrnd_Notify
                                       //    or VDD_Int_BGrnd_Notify events
   DWORD      VDD_nullpages ;          // bitmap of mapped null pages
   DWORD      VDD_SavedStt ;           // address of saved VDD_STT for VM's
                                       //    with CanRestore set in VDD_Flags
   BYTE       VDD_VertCnt ;            // Hold VRTC Counter
   BYTE       VDD_HorizCnt ;           // Toggle HRTC Counter
   WORD       VDD_PIF ;                // VDD PIF bits
   DWORD      VDD_Time_Out_Handle ;    // Display update timeout handle
   DWORD      VDD_ModTime ;            // System time at last video mod
   DWORD      VDD_Upd_Start ;          // Start time of window update
   DWORD      VDD_Mod_Flag_Save ;      // cumulative flags passed to grabber

   MSSPLANEINFO  VDD_Copy_Buf ;        // buffer used by grabber for
                                       //    determining memory changes

   DWORD      VDD_Get_Mem_Count ;      // Match calls VDD_Get_Mem/VDD_Free_Mem
                                       //    VDD_UnLock_APP unfreezes the App
                                       //    only if this count is Zero.
   BYTE       VDD_Update_Hit_Cnt ;
   BYTE       VDD_savemem_maps ;       // # of save mem pages mapped in VM
   BYTE       VDD_first_page ;   
   BYTE       VDD_visible_pages ; 

   DWORD      VDD_saved_visible ;

   BYTE       VDD_Postponed_TimeOuts ; // # of time-outs that have postponed
   BYTE       VDD_Owned_Pages ;        // count of physical pages alloc'ed
   BYTE       VDD_Padding[ 2 ] ;       // for padding only, reclaim as needed

} VDDCB, *PVDDCB ;

//----------------------------------------------------------------------------
//                 VDD flag bit definitions for VDD_Flags
//----------------------------------------------------------------------------

#define fVDD_Win  (1 << 0)             // VM is displayed in window
#define bVDD_Win  0               
#define fVDD_ModeSet (1 << 1)          // Video in INT 10 mode set
#define bVDD_ModeSet 1 
#define fVDD_ModeChange (1 << 2)       // Mode change in progress for
                                       //    windowed VM
#define bVDD_ModeChange 2
#define fVDD_MdSvRAM (1 << 3)          // RAM not destroyed during mode chg
#define bVDD_MdSvRAM 3
#define fVDD_PageChg (1 << 4)          // Video in INT 10 page change
#define bVDD_PageChg 4
#define fVDD_Font (1 << 5)             // Memory mapped for font load INT 10
#define bVDD_Font 5
#define fVDD_DirFont (1 << 6)          // Mem mapped for font load
                                       //    (directly programmed)
#define bVDD_DirFont 6
#define fVDD_WaitAtt (1 << 7)          // VDD suspended VM until attached
#define bVDD_WaitAtt 7

#define fVDD_DspDrvrAware (1 << 8)     // Set if display driver told us where
                                       //    its video memory resident data
                                       //    is kept
#define bVDD_DspDrvrAware 8

#define fVDD_HCurTrk (1 << 9)          // Track cursor horizontally
#define bVDD_HCurTrk   9
#define fVDD_Grab (1 << 10)            // Grab event pending
#define bVDD_Grab 10
#define fVDD_Upd (1 << 11)             // Update event pending
#define bVDD_Upd 11
#define fVDD_IdleUpdate (1 << 12)      // VM went idle during window update
#define bVDD_IdleUpdate 12
#define fVDD_MemCmods (1 << 13)        // Set if MemC mods are required for VM
#define bVDD_MemCmods 13
#define fVDD_MappedA0 (1 << 14)        // Mem mapped at A0 in CGA/text mode
#define bVDD_MappedA0 14
#define fVDD_AxRd (1 << 15)            // Attribute index read(VGA)
#define bVDD_AxRd 15
#define fVDD_IOT (1 << 16)             // I/O ports presently trapped
#define bVDD_IOT 16
#define fVDD_CanRestore (1 << 17)      // VM knows how to restore its screen
#define bVDD_CanRestore 17
#define fVDD_Boost (1 << 18)           // SYS VM boosted for memory read
#define bVDD_Boost 18
#define fVDD_Hide (1 << 19)            // Hide cursor in window
#define bVDD_Hide 19
#define fVDD_PVGA_Ena (1 << 20)        // Paradise VGA regs are enabled
#define bVDD_PVGA_Ena 20
#define fVDD_ATIEna (1 << 21)          // ATI register is enabled
#define bVDD_ATIEna 21
#define fVDD_NoSaveRes (1 << 22)       // set if VM must restore screen
#define bVDD_NoSaveRes 22
#define fVDD_IOTmono (1 << 23)         // mono CRTC I/O ports presently trapped
#define bVDD_IOTmono 23
#define fVDD_InInt2F (1 << 24)         // set while exec'ing foreground INT 2Fh
#define bVDD_InInt2F 24
#define fVDD_NeedRestore (1 << 25)     // modified VM state but not hardware
#define bVDD_NeedRestore 25
#define fVDD_PageFault (1 << 26)       // page fault occurred (timeout clears)
#define bVDD_PageFault 26
#define fVDD_InPFault (1 << 27)        // currently processing a page fault
#define bVDD_InPFault 27
#define fVDD_DacChg (1 << 28)          // DACs have changed since clear_mod
#define bVDD_DacChg 28
#define fVDD_BoostUpdate (1 << 29)     // Boost complete update
#define bVDD_BoostUpdate 29
#define fVDD_MInit (1 << 31)           // Display mem initialized
#define bVDD_MInit 31

//----------------------------------------------------------------------------
//          VDD_EFlags defnitions error and error msg status
//----------------------------------------------------------------------------

// Note that Msg bits are error bits shifted left 8

#define fVDE_NoMain (1 << 8)           // Out of main, display not accurate
#define bVDE_NoMain 8
#define fVDE_NoCopy (1 << 9)           // Out of copy, window not accurate
#define bVDE_NoCopy 9
#define fVDE_NoFGrnd (1 << 10)         // Cannot run bkgrnd with fgrnd VM
#define bVDE_NoFGrnd 10
#define fVDE_NoWind (1 << 11)          // Cannot window this VM
#define bVDE_NoWind 11
#define fVDE_Excl (1 << 12)            // Background high Res excluded
#define bVDE_Excl 12
#define fVDE_NoSupMode (1 << 13)       // Mode not supported
#define bVDE_NoSupMode 13
#define fVDE_NoPgAvail (1 << 14)       // Unable to satisfy video mem request
#define bVDE_NoPgAvail 14
#define fVDE_NoMono (1 << 15)          // Mono when mono pages unsupported
#define bVDE_NoMono 15
#define fVDE_NoMMsg (1 << 16)          // User told out of main
#define bVDE_NoMMsg 16
#define fVDE_NoCMsg (1 << 17)          // User told out of copy
#define bVDE_NoCMsg 17
#define fVDE_NoFGMsg (1 << 18)         // User told cannot run with fgrnd
#define bVDE_NoFGMsg 18
#define fVDE_NoWMsg (1 << 19)          // User told cannot window
#define bVDE_NoWMsg 19
#define fVDE_ExclMsg (1 << 20)         // User told VM exclusive
#define bVDE_ExclMsg 20
#define fVDE_NoSMsg (1 << 21)          // User told mode not supported
#define bVDE_NoSMsg 21
#define fVDE_NoPAMsg (1 << 22)         // Unable to satisfy video mem request
#define bVDE_NoPAMsg 22
#define fVDE_NoMonoMsg (1 << 23)       // Mono when mono pages unsupported
#define bVDE_NoMonoMsg 23

//
// The following equates make sure that the I/O trap routines do not
// write beyond the end of the save area reserved for each of the
// controllers.
//

// Since 'C' doesn't allow us to get offsets of structures here, we
// can't "assert" anything, so we just define the masks. (See VDDDEF.INC
// from the VDD sources for more information.)

#define mx_Attr     (0x1F)             // mask for Attribute index
#define mx_Grp      (0x1F)             // mask for graphics controller index
#define mx_Seq      (0x0F)             // mask for Sequencer index
#define mx_CRTC     (0x3F)             // mask for CRTC index

//----------------------------------------------------------------------------
//                      VDD global flag definitions
//----------------------------------------------------------------------------

// There is a global copy kept in dwVTFlags and local copy in each CB,
// for efficiency.  The initial value of manyu of these flags are passed
// from real mode initialization code that detects Video adapter types.

// Flags valued from 0-7 cannot be passed from real-mode init.  They
// are initialized to zero.  The size of the video mem is passed in place.

#define fVT_NoSaveResSysVM (1 << 0)    // Save/restore SYS VM enabled
#define bVT_NoSaveResSysVM 0
#define fVT_NoSaveResOthers (1 << 1)   // Save/restore other VMs enabled
#define bVT_NoSaveResOthers 1
#define fVT_Mono (1 << 2)              // VDD supports Mono VGA memory (B0-B7)
#define bVT_Mono 2
#define fVT_MonoPorts (1 << 3)         // VDD supports Mono VGA ports
#define bVT_MonoPorts 3
#define fVT_Notify_at_PF (1 << 4)      // bkgnd INT 2F notification done
                                       //    at pfault time
#define bVT_Notify_at_PF 4
#define fVT_ResvB0 (1 << 5)            // Reserve memory from B0 to B7
#define bVT_ResvB0 5
#define fVT_SuprsMsg (1 << 6)          // Do not message NoFGrnd and
                                       //    NoAvailPages
#define bVT_SuprsMsg 6
#define fVT_NoSuspend (1 << 7)         // Do not suspend application, 
                                       //    map null memory
#define bVT_NoSuspend 7
#define fVT_SysVMnot1stBank (1 << 8)   // Set if Sys VM not running in
                                       //    1st bank of video memory
#define bVT_SysVMnot1stBank 8

#define fVT_SysVMin2ndBank (1 << 9)    // Set if Sys VM is running in 2nd
                                       //    bank of video memory 
                                       //    (VT_SysVMnot1stBank must be
                                       //     set also)
#define bVT_SysVMin2ndBank 9
         
#define fVT_TVGA (1 << 16)             // Trident TVGA
#define bVT_TVGA 16
#define fVT_TVGA9100 (1 << 17)         // Trident TVGA 9100
                                       //    (needs special care)
#define bVT_TVGA9100 17
#define fVT_ATIVGA (1 << 18)           // ATI VGA Wonder
#define bVT_ATIVGA 18
#define fVT_ATIVGA3 (1 << 19)          // ATI VGA Wonder V3 and below
#define bVT_ATIVGA3 19
#define fVT_NEWCLVGA (1 << 20)         // Cirrus Logic VGA 6410
#define bVT_NEWCLVGA 20
#define fVT_CLVGA (1 << 21)            // Cirrus Logic VGA, 0ECh enable
                                       //    (more regs)
#define bVT_CLVGA 21
#define fVT_CLVGA_RevC (1 << 22)       // Cirrus Logic 610/620 chip set
                                       //    Rev C (this bit will only be
                                       //    set, if VT_CLVGA is)

#define bVT_CLVGA_RevC 22
         
#define fVT_V7VGA (1 << 23)            // Video 7 VGA (more regs)
#define bVT_V7VGA 23
#define fVT_8514 (1 << 24)             // IBM 8514
#define bVT_8514 24
#define fVT_XGAbit (1 << 25)           // Set for machines where IBM has
                                       //    redefined the meaning of bit 6 of
                                       //    CRT.17h.  Normally this bit
                                       //    selects byte/word addressing for
                                       //    the CRT controller and belongs to
                                       //    the CRTC owner.  On IBM's 
                                       //    implementation of XGA this bit
                                       //    selects byte/word addressing of
                                       //    the graphics controller and thus
                                       //    belongs to the MemC owner!
#define bVT_XGAbit 25
#define fVT_CTVGA (1 << 26)            // CHIPS VGA(82c441) controller
#define bVT_CTVGA 26
#define fVT_PVGA (1 << 27)             // Paradise VGA (more regs)
#define bVT_PVGA 27
#define fVT_PVGA1F (1 << 28)           // Paradise VGA version 1F (more regs)
#define bVT_PVGA1F 28
#define fVT_TLVGA (1 << 29)            // Tseng Labs VGA (more regs!)
#define bVT_TLVGA 29
#define fVT_TL_ET4000 (1 << 30)        // Tseng Labs ET4000 VGA
#define bVT_TL_ET4000 30
#define fVT_RsrvC6C7 (1 << 31)         // Reserve C6 & C7 as ROM
                                       //    without a signature
#define bVT_RsrvC6C7 31

#define fVid_MsgInit (1 << 0)          // Message mode INIT complete
#define bVid_MsgInit 0
#define fVid_DOff (1 << 1)             // Display is off(during save/restore)
#define bVid_DOff 1
#define fVid_SysVMI (1 << 2)           // Sys VM initialization complete
#define bVid_SysVMI 2
#define fVid_Crit (1 << 3)             // In video critical section
#define bVid_Crit 3
#define fVid_VMsSuspended (1 << 4)     // VM's have been suspended by VDD
#define bVid_VMsSuspended 4
#define fVid_MsgA (1 << 6)             // System message mode active
#define bVid_MsgA 6
#define fVid_SFPending (1 << 7)        // Set Focus is pending
#define bVid_SFPending 7
#define fVid_PlanarInit (1 << 8)       // Planar mode is initialized
#define bVid_PlanarInit 8
#define fVid_DspDrvr_Init (1 << 9)     // set on first INT 10h mode
                                       //    call from pmode
#define bVid_DspDrvr_Init 9
#define fVid_Dont_Set_Mode (1 << 10)   // set after a forced mode change,
                                       //    if done before Vid_DspDrvr_Init
                                       //    is set

#define bVid_Dont_Set_Mode 10
         
#define fVid_SR_Section (1 << 12)      // Special save restore "critical"
                                       //    section
#define bVid_SR_Section 12
#define fVid_SR_Latch_res (1 << 13)    // Restore latches at end of SR section
#define bVid_SR_Latch_res 13
#define fVid_SR_regs_saved (1 << 14)   // regs saved for CRTC owner
                                       //    while in SR section
#define bVid_SR_regs_saved 14

//----------------------------------------------------------------------------
//                  Video display register definitions
//----------------------------------------------------------------------------

// ...Colr are used with CGA and EGA color modes
// ...Mono are used with EGA/VGA monochrome modes

// NAMING CONVENTIONS
//    pXXXXX is a I/O port definition
//    fXXXXX is a bit mask
//    bXXXXX is a bit number
//    mXXXXX is a bit mask (multiple fXXXX)
//    vXXXXX is a value to output/input to/from a port
//    xXXXXX is a specific index

#define pCRTCIndxColr      0x3D4      // CRTC index register
#define pCRTCIndxMono      0x3B4

#define pCRTCDataColr      0x3D5      // CRTC data register
#define pCRTCDataMono      0x3B5

#define pStatColr          0x3DA      // Status register
#define pStatMono          0x3BA
#define fStatEna           0x0001     // Video RAM access OK for processor
#define fStatLPTr          0x0002     // Light pen triggered
#define fStatLPSw          0x0004     // Light pen switch is "on"
#define fStatVRTC          0x0008     // Vertical retrace in progress

// Initial value for pseudo status

#define mStat_IVal         (fStatEna+fStatVRTC)

#define pFeatColr          pStatColr  // Feature control
#define pFeatMono          pStatMono
#define pFeatVGAIn         0x3CA      // VGA feature control read
#define fFeatFC0           0x0001     //   FC0
#define fFeatFC1           0x0002     //   FC1

#define pLPen1Colr         0x3DB      // Light pen latch clear
#define pLPen1Mono         0x3BB
#define pLPen2Colr         0x3DC      // Light pen latch set
#define pLPen2Mono         0x3BC

#define pMisc              0x3C2      // Miscellaneous output
#ifndef CTVGA
#define pMiscIn            0x3CC      // VGA read Misc. output
#endif

#define fMiscPNum          0x0001     //   3Dx port numbers(vs. 3Bx)
#define fMiscREna          0x0002     //   Ram enable
#define fMiscDot           0x000C     //   Dot clock select
#define fMiscOSrc          0x0010     //   Output source
#define fMiscPage          0x0020     //   Page bit for odd/even mode
#define fMiscHPol          0x0040     //   Horizontal retrace polarity
#define fMiscVPol          0x0080     //   Vertical retrace polarity

// Mask for CRTC bits

#define mMiscCRTC          (fMiscREna+fMiscPage)

// Mask for MemC bits

#define mMiscMemC          0xFFh - mMiscCRTC

#define pAttr              0x03c0     // Attribute (palette) address/data
#define fVAI_Indx          0x0080     //    sign bit of attribute index
#define bVAI_Indx          7          //    flag indicating port is index
#define fVAI_ScOn          0x0020     //    Indicates screen is on
#define bVAI_ScOn          5         

#define pSeqIndx      0x3C4           // Sequencer address
#define pSeqData      0x3C5           // Sequencer data
#define fSeq0ARst     0x0001          //   0:Async reset
#define fSeq0SRst     0x0002          //   0:Sync reset
#define fSeq1DPCh     0x0001          //   1:Dots per character
#define fSeq1BWid     0x0002          //   1:Band Width
#define fSeq1ShLd     0x0004          //   1:Shift load(0=every char, 1=skip)
#define fSeq1DClk     0x0008          //   1:Dot clock(=1 if halved)
#define mSeq2WMsk     0x000F          //   2:Write mask
#define mSeq3ChrB     0x0003          //   3:Char map B select(attr bit 3 = 0)
#define mSeq3ChrA     0x000C          //   3:Char map A select(attr bit 3 = 1)
#define fSeq4Alph     0x0001          //   4:Alpha mode(char gen enabled)
#define fSeq4ExtM     0x0002          //   4:Extended memory installed
#define fSeq4SqAd     0x0004          //   4:Seq'l mem addr'ing(vs. odd/even)
#define fSeq4Chain4   0x0008          //   4:Chain 4 (double odd/even) for
                                      //     256 color (mode 13)
#define pGrp1Pos           0x3CC      // Graphics posn 1(=0)
#define pGrp2Pos           0x3CA      // Graphics posn 2(=1)
#define pGrpIndx           0x3CE      // Graphics controller address
#define pGrpData           0x3CF      // Graphics controller data

#define fGrp0StV0          0x0001     // 0:Value for plane 0 if enabled
#define fGrp0StV1          0x0002     // 0:Value for plane 1 if enabled
#define fGrp0StV2          0x0004     // 0:Value for plane 2 if enabled
#define fGrp0StV3          0x0008     // 0:Value for plane 3 if enabled
#define fGrp1Ena0          0x0001     // 1:Enable for plane 0
#define fGrp1Ena1          0x0002     // 1:Enable for plane 1
#define fGrp1Ena2          0x0004     // 1:Enable for plane 2
#define fGrp1Ena3          0x0008     // 1:Enable for plane 3
#define mGrp2Colr          0x000F     // 2:Color compare value
#define mGrp3RCnt          0x0007     // 3:Rotate left count for mode 0
#define mGrp3Func          0x0018     // 3:Function for modes 0 and 2
#define mGrp4RMsk          0x0007     // 4:Read map select
#define mGrp5WMod          0x0003     // 5:Write mode
#define fGrp5Test          0x0004     // 5:Test condition
#define fGrp5RMod          0x0008     // 5:Read mode(0=plane,1=compare)
#define fGrp5SqAd          0x0010     // 5:Seq'l mem addr'ing(vs. odd/even)
#define fGrp5CGA           0x0020     // 5:CGA 4 color mode addressing
#define fGrp6Char          0x0001     // 6:Char or graphics
#define fGrp6Chain         0x0002     // 6:Chain odd maps after even
#define mGrp6Addr          0x000C     // 6:Processor view of VRAM start addr
                                      //   00=A000 for 128kb
                                      //   01=A000 for 64kb
                                      //   10=B000 for 32kb
                                      //   11=B800 for 32kb
#define bGrp6Addr          0x0002     // First of 2 bits for start addr
#define mGrp7Colr          0x000F     // 7:Color don't care for read compare
                                      // 8:Mask: use latch value vs. data

// VGA DAC registers...

#define pDACWindx          0x3C8         // DAC Write index (R/W)
#define pDACRindx          0x3C7         // DAC Read index (RO)
#define pDACState          (pDACRindx)   // DAC state
#define pDACData           0x3C9         // DAC data (3 successive accesses)
#define pDACMask           0x3CA         // DAC mask

// VGA read access to write only registers

#define pMiscRead          0x3CC      // Port to read Misc output register
#define pFeatRead          0x3CA      // Port to read Feature output register

//----------------------------------------------------------------------------
//
//                        Miscellaneous equates
//
//----------------------------------------------------------------------------

// BIOS instance definitions

#define EGA_I_Addr1  0x449
#define EGA_I_Len1   0x1E
#define EGA_I_Addr2  0x484
#define EGA_I_Len2   0x07
#define EGA_I_Addr3  0x4A8
#define EGA_I_Len3   0x4
#define EGA_I_Addr4  0x410
#define EGA_I_Len4   0x2

#define BIOS_Col_Count  0x44A
#define BIOS_Line_Count 0x484

// timeout period to check for video updates when VM is windowed

#define VDD_Def_Update_Time  50

// boost for system VM when grabber is rendering VM into window

#define VDD_Pri_Device_Boost (Low_Pri_Device_Boost)

// flag bits in shadow_mem_status

#define SHADOW_EXITS    0x0001   // Shadow memory exists
#define SHADOW_IN_USE   0x0002   // Shadow contains valid image
#define SHADOW_TRASHED  0x0004   // Shadow contents have been destroyed

//----------------------------------------------------------------------------
//
//                        External references
//
//----------------------------------------------------------------------------

// svcthunk.asm:

extern SERVICEPROC  VDD_Service_Table[] ;

// vddctl.c:

extern DDB    VDD_DDB ;

extern DWORD  dwVidCBOff, dwVidFlags, dwVTFlags ;

extern DWORD  hVidFocusVM, hVidCRTCVM, hVidMemCVM,
              hVidLatchesVM, hVidStartSRMemCVM ;

extern DWORD  hVidMsgPseudoVM ;           
extern VDDCB  VidMsgPseudoCB ;

extern DWORD  dwVidScrollFreq ;

// vddfonts.c:

extern BYTE   bVidCharHgt ;
extern DWORD  dwVidMsgFontH, dwVidMsgFont ;

// vddinit.c:

extern DWORD  dwVidFontLen ;

// vddint.c:

extern DWORD  dwVidI10NextCS, dwVidI10NextEIP ;
extern DWORD  dwVidI2FNextCS, dwVidI2FNextEIP ;

// vddinta.asm:

extern DWORD  Vid_I10_OldVector ;

// vddproc.c:

extern DWORD  dwVidVMIdleChain ;

// vddphmem.c:

extern BYTE   bVidLatchBank, bVidSysVMLatchBank;
extern DWORD  dwVidLatchAddr, dwVidSysVMLatchAddr, 
              dwVidPages, dwVidPageTable, dwVidPhysA0000 ;

extern DWORD  VDD_PageHandle_List ;

// vddstate.c:

extern BYTE   bVidInitialTextRows ;
extern BYTE   bVidMsgPages ;
extern DWORD  hVidCRTCVM ;

// vddsvc.c

extern DWORD  dwVidShadowMemStatusPtr ;
extern DWORD  dwVidMsgModeTextRows, dwVidMsgModeTextCols ;
extern WORD   wPIFSave ;

// vddvmmem.c:

extern DWORD      dwReservedROMPage ;
extern DWORD      dwVidPageAllocFlags ;
extern DWORD      dwVidUnaccessedPage ;

extern EVENTPROC  pVDD_Error_CvtFullScreen ;

//
// Message references (see vddmsg.c)
//

extern char szVDD_SVGA_Mem_Ini[] ;
#ifdef SysVMin2ndBank
extern char szVDD_2ndBank_Ini[] ;
#endif

extern char szVDD_Str_BadDevice[] ;
extern char szVDD_SVGA_Mem_Ini[] ;
extern char szVDD_Time_Ini[] ;
extern char szVDD_SuspMsg_Ini[] ;
extern char szVDD_NoSusp_Ini[] ;
extern char szVid_Text_Rows_Ini[] ;
extern char szVid_Text_Rows_Sect[] ;
extern char szVDD_2nd_Ini[] ;
extern char szVDD_Mono_Text_Ini[] ;
extern char szVDD_Rsrv_C6C7_Ini[] ;
extern char szVDD_SVGA_Mem_Ini[] ;
extern char szVDD_VM_SaveRes_Ini[] ;
extern char szVDD_VM_Scroll_Freq_Ini[] ;
extern char szVDD_BkGnd_Notify_Time[] ;
extern char szVDD_Use_ROM_Font_Ini[] ;

extern char szVDD_SysVM_SaveRes_Ini[] ;

//----------------------------------------------------------------------------
//
//                        Function prototypes
//
//----------------------------------------------------------------------------

//---
// vddctl.c:
//---

BOOL CDECL VDD_Create_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_Sys_VM_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_VM_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_Set_Device_Focus
(
    DWORD           hVM,
    DWORD           dwVID,
    DWORD           dwFlags,
    DWORD           hAssocVM
) ;

BOOL VDD_Ctl_Set_Focus
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Ctl_End_Dsp_Crit() ;

BOOL CDECL VDD_Destroy_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_Suspend_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_Resume_VM
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_System_Exit
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

VOID CDECL VDD_Sys_Critical_Exit
(
    DWORD           hVM
) ;

BOOL CDECL VDD_VM_Critical_Init
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

//---
// vdddebug.c
//---


//---
// vddfonts.c
//---

VOID VDD_Font_SaveFont
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPage
) ;

VOID VDD_Font_RestoreFont
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPage
) ;

VOID VDD_Font_ExpandMsgFont
(
    PSTR            pDst
) ;

BOOL CDECL VDD_Ctl_Begin_Msg_Mode
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_Ctl_End_Msg_Mode
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

//---
// vddgrab.c
//---

BOOL VDD_Grab_DoGrab
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;


BOOL VDD_Grab_ACopy
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Grab_DCopy
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Grab_GetTextCopyBuf
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PDWORD          pAddr,
    PDWORD          pSize
) ;

VOID VDD_Grab_DeleteTextCopyBuf
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

BOOL VDD_Grab_ChangeQuery
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Grab_ClearModState
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;


VOID VDD_Grab_SaveDirtyPages
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

DWORD VDD_Grab_GetChgs
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PVDDMODSTATE    pModState
) ;

//---
// vddinit.c
//---

BOOL CDECL VDD_Sys_Critical_Init
(
    DWORD           hVM,
    DWORD           dwRefData,
    PSTR            pCmdTail,
    PCRS_32         pCRS
) ;


BOOL CDECL VDD_Device_Init
(
    DWORD           hVM,
    PSTR            pCmdTail,
    PCRS_32         pCRS
) ;

BOOL VDD_Font_SetupMsgModeInit
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB
) ;

BOOL VDD_Font_SaveMsgFont
(
    DWORD           hSysVM,
    PVDDCB          pSysVMCB,
    PCRS_32         pCRS
) ;

BOOL VDD_Font_InitComplete
(
    DWORD           hSysVM
) ;

VOID VDD_PHMem_SysCriticalInit
(
    DWORD           hVM
) ;

VOID VDD_State_SysCriticalInit
(
    DWORD           hVM,
    PVDDCB          pSysVMCB
) ;

BOOL VDD_TIO_DeviceInit
(
    PVDDCB          pSysVMCB
) ;

BOOL VDD_VMMem_CheckB0B7Write() ;

VOID VDD_VMMem_ReserveB0B7Pages
(
    PVDDCB          pSysVMCB
) ;

BOOL VDD_VMMem_DeviceInit
(
    DWORD           hVM,
    PVDDCB          pSysVMCB,
    PCRS_32         pCRS
) ;

BOOL CDECL VDD_VMMem_InitComplete
(
    DWORD           hSysVM,
    PSTR            pCmdTail,
    PCRS_32         pCRS
) ;

//---
// vddint.c
//---

VOID VDD_Int_SysCriticalExit() ;

//---
// vddinta.asm
//---

VOID PASCAL VDD_Int_MungeInt10h
(
    DWORD           hSysVM
) ;

//---
// vddoem.c
//---

DWORD VDD_OEM_PreRestoreAttr
(
    DWORD           hVM,
    PVDDCB          pMemCVMCB,
    PVDDCB          pCRTCVMCB,
    WORD            wPort,
    DWORD           dwMx
) ;

VOID VDD_OEM_PostRestoreAttr
(
    DWORD           hVM,
    PVDDCB          pMemCVMCB,
    PVDDCB          pCRTCVMCB,
    WORD            wPort
) ;

BOOL VDD_OEM_AdjustCursor
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PUINT           puCurStart,
    PUINT           puCurEnd
) ;

//---
// vddphmem.c:
//---

BOOL VDD_PHMem_GetVisiblePages( DWORD, PVDDCB, PUINT, PUINT ) ;
BOOL VDD_PHMem_GetPageOwner( DWORD, PVDDCB, UINT, PDWORD, PDWORD ) ;
VOID VDD_PHMem_UpdateLRU( PVDDCB, DWORD ) ;
BOOL VDD_PHMem_SharedPage( DWORD, PVDDCB, UINT, UINT ) ;
BOOL VDD_PHMem_FreePage( DWORD ) ;

BOOL VDD_PHMem_PageAccessible( DWORD, PVDDCB, DWORD ) ;

typedef BOOL (*PAGEPROC)( DWORD, PVDDCB, DWORD, DWORD, DWORD ) ;
BOOL VDD_PHMem_ChangePage( DWORD, PVDDCB, UINT, DWORD, PAGEPROC ) ;
DWORD VDD_PHMem_AllocVideoPage( DWORD, PVDDCB, UINT, PAGEPROC ) ;

//---
// vddproc.c:
//---

VOID VDD_VMIdle
(
    DWORD           hVM
) ;

VOID CDECL VDD_VMTimeOut
(
    DWORD           hVM,
    PCRS_32         pCRS,
    DWORD           dwRefData,
    DWORD           dwExtra
) ;

VOID VDD_UpdateWindow
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_SetVMTimeOut
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_ClrVMTimeOut
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID CDECL VDD_Update_CallBack
(
    PCRS_32         pCRS,
    DWORD           dwRefData
) ;

VOID CDECL VDD_UpdateEvent
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           dwRefData
) ;

VOID VDD_VMDA_Grab
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Grab_Event
(
    PVDDCB          pVMCB
) ;

VOID CDECL VDD_Grab_CallBack
(
    PCRS_32         pCRS,
    DWORD           dwRefData
) ;

VOID VDD_Proc_VMResume
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Proc_VMSuspend
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID CDECL VDD_Proc_BGrndINT2F
(
    DWORD           hVM,
    DWORD           dwRefData
) ;

VOID CDECL VDD_Proc_FGrndINT2F
(
    DWORD           hVM,
    DWORD           dwRefData
) ;

VOID VDD_Proc_BGrndNotify
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Proc_NotifyBkgndVMs
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Proc_ChkBkgndUsage
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Proc_SuspendBkgndVMs
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Proc_ResumeBkgndVMs
(
    DWORD           hVM
) ;

VOID VDD_Proc_ForceCurVMFgrnd
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

//---
// vddstate.c:
//---

VOID VDD_State_SysCriticalInit
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

WORD VDD_State_GetCRTCIndexPort() ;

VOID VDD_State_ClearMemCOwner() ;

VOID VDD_State_SetMemCOwner
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_State_CreateVM
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_State_GetState
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PBYTE           pDst,
    DWORD           dwFlags
) ;

#define GETSTATEF_VMCOPYSTATE  0x0000
#define GETSTATEF_MAINSTATE    0x0001

//---
// vddsvc.c:
//---

#define VDD_Num_Services   0x0E

typedef VOID (*GRBSVCPROC)
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID CDECL VDD_Get_Version
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Get_GrbVersion
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Begin_Msg_Mode
(
    PREGS           pRegs
) ;

VOID CDECL VDD_End_Msg_Mode
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Msg_ClrScrn
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Msg_ForColor
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Msg_BakColor
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Msg_TextOut
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Msg_SetCursPos
(
    PREGS           pRegs
) ;

VOID CDECL VDD_PIF_State
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Set_HCurTrk
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Get_GrabRtn
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Grab
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Hide_Cursor
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Query_Access
(
    PREGS           pRegs
) ;


VOID CDECL VDD_Set_VMType
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Check_Update_Soon
(
    PREGS           pRegs
) ;

VOID CDECL VDD_Get_ModTime
(
    PREGS           pRegs
) ;

VOID CDECL VDD_PM_API
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

VOID VDD_Grb_Version
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Get_State
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Get_GrbState
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Get_Mod
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Clear_Mod
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Get_Mem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Get_GrbMem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

BOOL VGMm_Set_Struc
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS
) ;

VOID VDD_Suspend_NormSch
(
    DWORD           hSysVM
) ;

VOID VDD_Free_Grab
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Free_Mem
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_Unlock_App
(
    DWORD           hSysVM,
    PCRS_32         pCRS,
    DWORD           hVM,
    PVDDCB          pVMCB,
    BOOL            fVMValid
) ;

VOID VDD_SVC_DspVersion
(
    PCRS_32         pCRS
) ;

VOID VDD_SVC_SetAddresses
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

VOID VDD_Init_DspDrvRing0
(
    DWORD           hVM,
    PCRS_32         pCRS
) ;

VOID CDECL VDD_Error_CvtFullScrn
(
    DWORD           hVM,
    DWORD           dwRefData,
    PCRS_32         pCRS
) ;

VOID VDD_Error_NoPagesAvail
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Error_NoFGrnd
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_Error_NoMainMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_ErrorMessage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PSTR            pMsg,
    UINT            uEFlagBit
) ;

VOID VDD_Message
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uFlags,
    PSTR            pMsg,
    UINT            uEFlagBit
) ;

//---
// vddsvca.asm:
//---

VOID VDD_Jump_To_Ring0() ;

// vddtio.c:

typedef UINT (*IODISPPROC)( DWORD, PVDDCB, UINT, UINT, PCRS_32, UINT, UINT ) ;

#define IODISP_OWNERF_NONE   0x0000
#define IODISP_OWNERF_CRTC   0x0001

VOID VDD_TIO_EnableLocal
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;


VOID VDD_TIO_DisableLocal
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_TIO_SetTrap
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

UINT CDECL VIOT_Disp
(
    DWORD                   hVM,
    UINT                    uIOType,
    UINT                    uPort,
    PCRS_32                 pCRS,
    UINT                    uData
) ;

UINT VIOT_CRTC
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uIOType,
    UINT            uPort,
    PCRS_32         pCRS,
    UINT            uData,
    UINT            fuOwner
) ;

UINT VIOT_Ignore
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uIOType,
    UINT            uPort,
    PCRS_32         pCRS,
    UINT            uData,
    UINT            fuOwner
) ;

UINT VIOT_3DB
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uIOType,
    UINT            uPort,
    PCRS_32         pCRS,
    UINT            uData,
    UINT            fuOwner
) ;

// vddvga.c

VOID VDD_VGA_SaveRegs() ;

VOID VDD_VGA_SaveAttr
(
    PVDDCB          pVMCB,
    WORD            wPort
) ;

VOID VDD_VGA_SaveMisc
(
    PVDDCB          pVMCB,
    WORD            wPort
) ;

VOID VDD_VGA_SaveCRTC
(
    PVDDCB          pVMCB,
    WORD            wPort
) ;

VOID VDD_VGA_SaveGrp
(
    PVDDCB          pVMCB
) ;

VOID VDD_VGA_SaveSeq
(
    PVDDCB          pVMCB
) ;

VOID VDD_VGA_SaveDAC
(
    PVDDCB          pVMCB
) ;

VOID VDD_VGA_RestoreDAC
(
    PVDDCB          pVMCB
) ;

// vddvmmem.c:

BOOL VDD_VMMem_AllocPageMaps
(
    PVDDCB          pVMCB,
    DWORD           dwVidPages
) ;

BOOL VDD_VMMem_GrowPageMaps
(
    PVDDCB          pVMCB,
    DWORD           dwVidPages
) ;

VOID VDD_VMMem_VMCriticalInit
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_AllocVisiblePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
) ;

BOOL VDD_VMMem_ChkVideoPages
(
    PVDDCB          pVMCB,
    UINT            uFirstPage,
    UINT            uPages
) ;

VOID VDD_VMMem_SetDeviceFocus
(
    DWORD           hVM
) ;

BOOL VDD_VMMem_CreateVM
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_DestroyVM
(
    DWORD           hVM
) ;

BOOL VDD_VMMem_VMResume
(
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_VMSuspend
(
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_UnlockSaveMem
(
    PVDDCB          pVMCB,
    UINT            uPlanes
) ;

VOID VDD_VMMem_DeAlloc
(
    PMEMSTATE       pMS
) ;

VOID VDD_VMMem_DMain
(
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_SetPlaneSizes
(
    PVDDCB              pVMCB,
    PMEMSTATE           pMS,
    WORD                acpPlanes[]
) ;

VOID VDD_VMMem_AdjustPlaneSizes
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS,
    BOOL            fCopy
) ;

BOOL VDD_VMMem_ReAllocPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI
) ;

BOOL VDD_VMMem_ReAlloc
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    PMEMSTATE       pMS
) ;

BOOL VDD_VMMem_FontPage
(
    PVDDCB              pVMCB,
    UINT                uPageID
) ;

VOID VDD_VMMem_EndModeChg
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_ChangeSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_AdjustVisiblePages
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

BOOL VDD_VMMem_CA_CheckPlane0
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uAllocNeeded
) ;

BOOL VDD_VMMem_CA_CheckPlanes
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uAllocNeeded
) ;

BOOL VDD_VMMem_ChkAlloc
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
) ;

VOID VDD_VMMem_MemC_AboutToChange
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_Disable
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

BOOL VDD_VMMem_VMCanUseSaveMem
(
    PVDDCB         pVMCB
) ;

VOID CDECL VDD_PFault
(
    DWORD           hVM,
    UINT            uPageNum
) ;

BOOL VDD_VMMem_MapSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID,
    UINT            uPageNum
) ;

VOID VDD_VMMem_UnmapSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageID
) ;

VOID VDD_VMMem_MapNull
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    UINT            uPageNum
) ;

VOID VDD_VMMem_DisableNull
(
    DWORD               hVM,
    PVDDCB              pVMCB
) ;

BOOL VDD_VMMem_PageHandler
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           dwNotify,
    DWORD           hPage,
    DWORD           hNewPage
) ;

BOOL VDD_VMMem_Msg_PageHandler
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           dwNotify,
    DWORD           hPage,
    DWORD           hNewPage
) ;

VOID VDD_VMMem_UnmapPage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage
) ;

DWORD VDD_VMMem_GetPTE
(
    DWORD           hVM,
    UINT            uLinPageNum
) ;

DWORD VDD_VMMem_IsPageDirty
(
    DWORD           hVM,
    DWORD           hPage
) ;

BOOL VDD_VMMem_ChkSaveMem
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_ChkMemCState
(
    PVDDCB          pVMCB,
    BYTE            bMxPlane
) ;

BOOL VDD_VMMem_SavePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage
) ;

BOOL VDD_VMMem_RestorePage
(
    DWORD           hVM,
    PVDDCB          pVMCB,
    DWORD           hPage,
    PBYTE           pPageFlags
) ;

VOID VDD_VMMem_SelectPlane
(
    BYTE            bPlaneNum,
    BYTE            bPlaneWMask
) ;

BOOL VDD_VMMem_ChkPlaneSize
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    UINT            uPgIndx
) ;

BOOL VDD_VMMem_LockSaveMemPage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    UINT            uPgIndx
) ;

VOID VDD_VMMem_CopyFromPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
) ;

VOID VDD_VMMem_SaveAttributePage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
) ;

VOID VDD_VMMem_RestoreAttributePage
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
) ;

VOID VDD_VMMem_CopyToPlane
(
    DWORD           hVM,
    PMSSPLANEINFO   pMPI,
    DWORD           hPage,
    UINT            uPhysPgNum
) ;

VOID VDD_VMMem_ZeroPage
(
    UINT            uPhysPgNum
) ;

VOID VDD_VMMem_InitText
(
    UINT            uPhysPage,
    WORD            wValue
) ;

VOID VDD_VMMem_AllocMsgModeMem() ;

VOID VDD_VMMem_SaveDirtyState
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_FreePages
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

VOID VDD_VMMem_RecordVRAMMod
(
    PVDDCB          pVMCB,
    UINT            uPageID
) ;

VOID VDD_VMMem_DebugDumpPages
(
    DWORD           hVM,
    PVDDCB          pVMCB
) ;

// vxdthunk.asm:

UINT PASCAL VDD2_GetVersion( VOID ) ;

//---------------------------------------------------------------------------
//  End of File: vddvga.h
//---------------------------------------------------------------------------
