//---------------------------------------------------------------------------
//
//  Module: vmdavga.h
//
//  Purpose:
//     Definitions for VMDOSAPP interface to VDD
//
//     Based on original Windows 3.1 VDD sources as supplied by
//     the Microsoft Windows 3.1 Device Driver Development Kit.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//             10/25/93   BryanW      Wrote it.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

// VDD virtual mode services for the grabber

#define GRB_Get_Version    0x0000
#define GRB_Get_Mem        0x0001
#define GRB_Get_State      0x0002
#define GRB_Get_Mod        0x0003
#define GRB_Clear_Mod      0x0004
#define GRB_Free_Mem       0x0005
#define GRB_Get_GrbMem     0x0006
#define GRB_Free_Grab      0x0007
#define GRB_Get_GrbState   0x0008
#define GRB_Unlock_APP     0x0009

// New APIs for 3.1 display drivers

#define Private_DspDrvr_1  0x000A
#define DspDrvr_Version    0x000B
#define DspDrvr_Addresses  0x000C

#define Grabber_VerNum     0x030A   // version 3.10

//
// controller state structure
//

typedef struct tagVDAEGASTATE
{
   BYTE  VDA_EGA_Mode ;          // Display mode(0-A)
   BYTE  VDA_EGA_Colr ;          // Border color
   BYTE  VDA_EGA_Flags ;         // Flags
   BYTE  VDA_EGA_Rows ;          // Height of display in lines (text mode)
   WORD  VDA_EGA_CurX ;          // Cursor horizontal position
   WORD  VDA_EGA_CurY ;          // Cursor vertical position
   WORD  VDA_EGA_CurBeg ;        // Cursor start scan line
   WORD  VDA_EGA_CurEnd ;        // Cursor end scan line
   BYTE  VDA_EGA_Pal[ 16 ] ;     // Palettes (6 bits of color mapping)

} VDAEGASTATE, *PVDAEGASTATE ;

// controller flags

#define fVDA_V_ScOff        0x0001  // screen off
#define fVDA_V_ScOffBit     0
#define fVDA_V_HCurTrk      0x0002  // horizontal cursor track
#define fVDA_V_HCurTrkBit   1
#define fVDA_V_InvDAC       0x0004  // set if DAC 0 is non-zero
#define fVDA_V_InvDACBit    2

//=========================================================================
//
// EGA/VGA VDD modifications state structure ::= <flag>,<count>,<list>
//
// <flag>  indicates type, list valid and controller change and
//         cursor change.
//
// <count> is count of entries(not bytes or words) in list.
//
// <list>  is type dependent as follows:
//
//         Type 0 - list is modified pages bit map, 1 bit per displayed
//                  4k page, count is page count. Bit map is one byte per
//                  32k linear address space, where byte 0's bits are
//                  <page 0 bit>,<page 1 bit>,...,<page 7 bit>.
//                  byte 1's bits are <page 8 bit>,<page 9 bit>.
//
//                  A set bit indicates a change in that page.
//
//                  This type is not used for text modes.
//
//                  Modes D and E display is less than two pages long
//                  and so the page count will be two. Mode 10 page
//                  count is 7. Modes 11 and 12 page count is 10.
//
//         Type 1 - modified enclosing rectangles, count is rectangle
//                  count, (less than four) , rectangle is two points:
//                  <left>,<top> and <right>,<bottom>. Each point is
//                  relative to left, upper corner which is 0,0 and
//                  each value is stored in one word.
//
//                  This type may be used for all modes.
//
//                  For text modes, modes 0-3, there may be a special,
//                  scroll rectangle.  If the high bit of R_Left is set
//                  in one of these RECTs, then it is not a rect but a
//                  scroll specification.  The structure then becomes:
//                  <scroll count>,<scroll flags>,<right column>,
//                  <left column>,<bottom line>,<top line>,<fill character>,
//                  <fill attribute>.  Note that all items are byte
//                  values and that the scroll event specification
//                  takes the same space as an update rectangle.
//
//=========================================================================

#define ModListMax 64

typedef struct tagVDDMODSTATE
{
   WORD  VDD_Mod_Flag ;               // modified state flag
   WORD  VDD_Mod_Count ;              // count of entries in list
   BYTE  VDD_Mod_List[ ModListMax ] ; // modification list (type dependant)

} VDDMODSTATE, *PVDDMODSTATE ;

// masks and bit definitions for VDD_Mod_Flag

#define fVDD_M_Type       0x0007
#define fVDD_M_TypeBit    0
#define fVDD_M_Ctlr       0x0008
#define fVDD_M_CtlrBit    3
#define fVDD_M_VRAM       0x0010
#define fVDD_M_VRAMBit    4
#define fVDD_M_Curs       0x0020
#define fVDD_M_CursBit    5
#define fVDD_M_ScOff      0x0040
#define fVDD_M_ScOffBit   6
#define fVDD_M_Err        0x8000
#define fVDD_M_ErrBit     15

// values for fVDD_M_TypeMsk field

#define fVDD_M_Type_Page 0x0000   // bit field of modified pages
#define fVDD_M_Type_Rect 0x0001   // array of rect structures

// structure for accessing RECTs

typedef struct tagRECT
{
   WORD  R_Left ;
   WORD  R_Top ;
   WORD  R_Right ;
   WORD  R_Botm ;
} RECT, *PRECT ;

typedef struct tagVDAMEMSTATE
{
   DWORD  VDA_Mem_Addr_P0 ;   // Addr of plane 0 video memory
   DWORD  VDA_Mem_Addr_P1 ;   // Addr of plane 1 video memory
   DWORD  VDA_Mem_Addr_P2 ;   // Addr of plane 2 video memory
   DWORD  VDA_Mem_Addr_P3 ;   // Addr of plane 3 video memory
   DWORD  VDA_Mem_DPagOff ;   // 32 bit offset to start of display
   DWORD  VDA_Mem_Size_P0 ;   // Size of plane 0 video memory
   DWORD  VDA_Mem_Size_P1 ;   // Size of plane 1 video memory
   DWORD  VDA_Mem_Size_P2 ;   // Size of plane 2 video memory
   DWORD  VDA_Mem_Size_P3 ;   // Size of plane 3 video memory
   DWORD  VDA_Mem_DACOff ;    // address of DAC

} VDAMEMSTATE, *PVDAMEMSTATE ;

#define VDA_Mem_Addr_Win_State (FIELDOFFSET( VDAMEMSTATE, VDA_Mem_Addr_P1 ))
#define VDA_Mem_Size_Win_State (FIELDOFFSET( VDAMEMSTATE, VDA_Mem_Size_P1 ))

//---------------------------------------------------------------------------
//  End of File: vmdavga.h
//---------------------------------------------------------------------------

