//---------------------------------------------------------------------------
//
//  Module: vxdinit.c
//
//  Purpose:
//     Init code and data for VMEMTRAP.
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//              3/14/93   BryanW      Wrote it.
//
//---------------------------------------------------------------------------
//
//               Copyright (c) 1993 Woodruff Software Systems.
//                          All Rights Reserved.
//
//---------------------------------------------------------------------------

#include <vmm.h>
#include "vmemtrap.h"

#pragma data_seg( "_IDATA", "ICODE" ) ;

//=============================================================================
//                             I C O D E
//=============================================================================

//------------------------------------------------------------------------
//  BOOL VMEMTRAP_Device_Init
//
//  Description:
//     Allocates the VM control block area.
//
//  Parameters:
//     hVM
//        System VM handle
//
//     pCmdTail
//        pointer to WIN.COM's command tail
//
//     pCRS
//        pointer to System VM client register structure
//
//  History:   Date       Author      Comment
//              3/ 9/93   BryanW      Wrote it.
//
//------------------------------------------------------------------------

BOOL CDECL VMEMTRAP_Device_Init
(
   DWORD     hVM,
   PSTR      pCmdTail,
   PCRS_32   pCRS
)
{
   DWORD   aGlobalVMPages[ 9 ] ;
   WORD    wPage ;
   PVMCB   pSysVMCB ;

   UNUSED_PARAM( hVM ) ;
   UNUSED_PARAM( pCmdTail ) ;
   UNUSED_PARAM( pCRS ) ;

   vmmTraceOut( "VMEMTRAP_Device_Init\r\n" ) ;

   if (!(guCBOffset = vmmAllocateDeviceCBArea( sizeof( VMCB ), NULL )))
   {
      vmmDebugOut( "Failed to allocate control block!\r\n" ) ;
      return ( FALSE ) ;
   }

   pSysVMCB = (PVMCB) (hVM + guCBOffset) ;

   //
   // Search for a page starting from a page above normal
   // ROM (video and fonts, etc.)
   //
   // Pages below the last VM page can not be hooked with the
   // HookV86Page service.  This code is somewhat evil, but
   // I don't know of a better way to demonstrate this technique,
   // other than forcing a SYSTEM.INI switch.
   //

   vmmGetDeviceV86PagesArray( NULL, &aGlobalVMPages, NULL ) ;

   for (wPage = 0xDD; wPage < 0x110; wPage++)
   {
      if (0 == (aGlobalVMPages[ wPage / 32 ] & (1 << (wPage % 32))))
         break ;
   }
   if (wPage == 0x110)
   {
      vmmDebugOut( "VMEMTRAP: Unable to locate unassigned page.\r\n" ) ;
      return ( FALSE ) ;
   }

   vmmTraceOutParam( "VMEMTRAP: Attempting to assign page #EAX\r\n",
                     wPage ) ;

   if (!vmmAssignDeviceV86Pages( wPage, 1, NULL, NULL ))
   {
      vmmDebugOut( "VMEMTRAP: could not assign page.\r\n" ) ;
      return ( FALSE ) ;
   }

   if (pVMEMTRAP_PFault = vmmwrapThunkV86PHProc( VMEMTRAP_PFault ))
      vmmHookV86Page( wPage, pVMEMTRAP_PFault ) ;
   else
   {
      vmmDebugOut( "VMEMTRAP: Could not allocate Page_Fault thunk!\r\n" ) ;
      return ( FALSE ) ;
   }

   // Allocate a page to map into this space

   vmmPageAllocate( &pSysVMCB -> hPage, 1, PG_HOOKED, hVM,
                    NULL, NULL, NULL, NULL, PageZeroInit ) ;

   // Map memory into this space in SysVM

   vmmMapIntoV86( pSysVMCB -> hPage, hVM, wPage, 1, 0, NULL ) ;

   ghOwnerVM = hVM ;
   gwHookedPage = wPage ;

   return ( TRUE ) ;

} // end of VMEMTRAP_Device_Init

//---------------------------------------------------------------------------
//  End of File: vxdinit.c
//---------------------------------------------------------------------------

